/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// Copyright (C) 2001-2003 pat@aftermoon.net for modif flanked by <serverping>

#include <ctype.h>
#ifdef _WIN32
#include <io.h>
#endif
#include "client.h"
#include "../client/qmenu.h"

static int	m_main_cursor;

#define NUM_CURSOR_FRAMES 15

#define MOUSEBUTTON1 0
#define MOUSEBUTTON2 1

static char *menu_in_sound		= "misc/menu1.wav";
static char *menu_move_sound	= "misc/menu2.wav";
static char *menu_out_sound		= "misc/menu3.wav";

void M_Menu_Main_f (void);
	void M_Menu_Game_f (void);
		void M_Menu_LoadGame_f (void);
		void M_Menu_SaveGame_f (void);
		void M_Menu_PlayerConfig_f (void);
			void M_Menu_DownloadOptions_f (void);
		void M_Menu_Credits_f( void );
	void M_Menu_Multiplayer_f( void );
		void M_Menu_JoinServer_f (void);
			void M_Menu_AddressBook_f( void );
		void M_Menu_StartServer_f (void);
			void M_Menu_DMOptions_f (void);
	void M_Menu_Video_f (void);
	void M_Menu_Options_f (void);
		void M_Menu_Keys_f (void);
	void M_Menu_Quit_f (void);

	void M_Menu_Credits( void );

qboolean	m_entersound;		// play after drawing a frame, so caching
								// won't disrupt the sound

cvar_t	*options_menu;
extern cvar_t	*alt_text_color;

void	(*m_drawfunc) (void);
const char *(*m_keyfunc) (int key);

//=============================================================================
/* Support Routines */

#define	MAX_MENU_DEPTH	8


typedef struct
{
	void	(*draw) (void);
	const char *(*key) (int k);
} menulayer_t;

menulayer_t	m_layers[MAX_MENU_DEPTH];
int		m_menudepth;

// Knghtmare- added Psychospaz's menu scaling stuff
float scaledVideo (float param)
{
	return param*menuScale.avg;
}

float videoScale (void)
{
	return menuScale.avg;
}
// end Knightmare

void M_Banner( char *name )
{
	//float ratio;
	int w, h;

	re.DrawGetPicSize (&w, &h, name );
	/*if (w)
	{
		ratio = 45.0/(float)h;
		h = 45;
		w *= ratio;
	}*/
	re.DrawStretchPic (
		viddef.width / 2 - scaledVideo(w)/2,
		viddef.height/ 2 - scaledVideo(150),
		scaledVideo(w), scaledVideo(h), name, 1.0);
	//re.DrawPic( viddef.width / 2 - w / 2, viddef.height / 2 - 110, name );
}

// Knightmare- added Psychospaz's mouse support
void refreshCursorButtons(void)
{
	cursor.buttonused[MOUSEBUTTON2] = true;
	cursor.buttonclicks[MOUSEBUTTON2] = 0;
	cursor.buttonused[MOUSEBUTTON1] = true;
	cursor.buttonclicks[MOUSEBUTTON1] = 0;
}

void M_PushMenu ( void (*draw) (void), const char *(*key) (int k) )
{
	int		i;

	if (Cvar_VariableValue ("maxclients") == 1 && Com_ServerState () && !cls.consoleActive) // Knightmare added
		Cvar_Set ("paused", "1");

	// Knightmare- if just opened menu, and ingame and not DM, grab screen first
	if (cls.key_dest != key_menu && !Cvar_VariableValue("deathmatch")
		&& Com_ServerState() == 2) //ss_game
		//&& !cl.cinematictime && Com_ServerState())
		re.MenuGrabScreen();

	// if this menu is already present, drop back to that level
	// to avoid stacking menus by hotkeys
	for (i=0 ; i<m_menudepth ; i++)
		if (m_layers[i].draw == draw &&
			m_layers[i].key == key)
		{
			m_menudepth = i;
		}

	if (i == m_menudepth)
	{
		if (m_menudepth >= MAX_MENU_DEPTH)
			Com_Error (ERR_FATAL, "M_PushMenu: MAX_MENU_DEPTH");
		m_layers[m_menudepth].draw = m_drawfunc;
		m_layers[m_menudepth].key = m_keyfunc;
		m_menudepth++;
	}

	m_drawfunc = draw;
	m_keyfunc = key;

	m_entersound = true;

	// Knightmare- added Psychospaz's mouse support
	refreshCursorLink();
	refreshCursorButtons();

	cls.key_dest = key_menu;
}

void M_ForceMenuOff (void)
{
	// Knightmare- added Psychospaz's mouse support
	refreshCursorLink();
	m_drawfunc = 0;
	m_keyfunc = 0;
	cls.key_dest = key_game;
	m_menudepth = 0;
	Key_ClearStates ();
	if (!cls.consoleActive) // Knightmare added
		Cvar_Set ("paused", "0");
}

void M_PopMenu (void)
{
	S_StartLocalSound( menu_out_sound );
	if (m_menudepth < 1)
		Com_Error (ERR_FATAL, "M_PopMenu: depth < 1");
	m_menudepth--;

	m_drawfunc = m_layers[m_menudepth].draw;
	m_keyfunc = m_layers[m_menudepth].key;

	// Knightmare- added Psychospaz's mouse support
	refreshCursorLink();
	refreshCursorButtons();

	if (!m_menudepth)
		M_ForceMenuOff ();
}

void Options_MenuDraw (void);
const char *Default_MenuKey( menuframework_s *m, int key )
{
	const char *sound = NULL;
	menucommon_s *item;

	if ( m )
	{
		if ( ( item = Menu_ItemAtCursor( m ) ) != 0 )
		{
			if ( item->type == MTYPE_FIELD )
			{
				if ( Field_Key( ( menufield_s * ) item, key ) )
					return NULL;
			}
		}
	}

	switch ( key )
	{
	case K_ESCAPE:
		if (m_drawfunc==Options_MenuDraw && options_menu->value>0)
		{
			refreshCursorLink();
			Cvar_SetValue( "options_menu", 0 );
		}
		else
			M_PopMenu();
		return menu_out_sound;
	case K_KP_UPARROW:
	case K_UPARROW:
		if ( m )
		{
			m->cursor--;
			// Knightmare- added Psychospaz's mouse support
			refreshCursorLink();

			Menu_AdjustCursor( m, -1 );
			sound = menu_move_sound;
		}
		break;
	case K_TAB:
	case K_KP_DOWNARROW:
	case K_DOWNARROW:
		if ( m )
		{
			m->cursor++;
			// Knightmare- added Psychospaz's mouse support

			refreshCursorLink();
			Menu_AdjustCursor( m, 1 );
			sound = menu_move_sound;
		}
		break;
	case K_KP_LEFTARROW:
	case K_LEFTARROW:
		if ( m )
		{
			Menu_SlideItem( m, -1 );
			sound = menu_move_sound;
		}
		break;
	case K_KP_RIGHTARROW:
	case K_RIGHTARROW:
		if ( m )
		{
			Menu_SlideItem( m, 1 );
			sound = menu_move_sound;
		}
		break;

	/*case K_MOUSE1:
	case K_MOUSE2:
	case K_MOUSE3:
	//Knightmare 12/22/2001
	case K_MOUSE4:
	case K_MOUSE5:
	case K_MOUSE6:*/
	//end Knightmare
	case K_JOY1:
	case K_JOY2:
	case K_JOY3:
	case K_JOY4:
	case K_AUX1:
	case K_AUX2:
	case K_AUX3:
	case K_AUX4:
	case K_AUX5:
	case K_AUX6:
	case K_AUX7:
	case K_AUX8:
	case K_AUX9:
	case K_AUX10:
	case K_AUX11:
	case K_AUX12:
	case K_AUX13:
	case K_AUX14:
	case K_AUX15:
	case K_AUX16:
	case K_AUX17:
	case K_AUX18:
	case K_AUX19:
	case K_AUX20:
	case K_AUX21:
	case K_AUX22:
	case K_AUX23:
	case K_AUX24:
	case K_AUX25:
	case K_AUX26:
	case K_AUX27:
	case K_AUX28:
	case K_AUX29:
	case K_AUX30:
	case K_AUX31:
	case K_AUX32:
		
	case K_KP_ENTER:
	case K_ENTER:
		if ( m )
			Menu_SelectItem( m );
		sound = menu_move_sound;
		break;
	}

	return sound;
}

//=============================================================================

/*
================
M_DrawCharacter

Draws one solid graphics character
cx and cy are in 320*240 coordinates, and will be centered on
higher res screens.
================
*/
void M_DrawCharacter (int cx, int cy, int num,
		int red, int green, int blue, int alpha, qboolean italic)
{
	if (videoScale() > 1)
		re.DrawScaledChar ( scaledVideo( cx + ((MENU_STATIC_WIDTH-320)/2) ),
							scaledVideo( cy + ((MENU_STATIC_HEIGHT- 240)/2) ),
							num, videoScale(), red, green, blue, alpha, italic);
	else
		re.DrawScaledChar ( cx + ((viddef.width - 320)>>1), cy + ((viddef.height - 240)>>1 ),
							num, videoScale(), red, green, blue, alpha, italic);
	//re.DrawChar ( cx + ((viddef.width - 320)>>1), cy + ((viddef.height - 240)>>1), num, 255);
}

void M_Print (int cx, int cy, char *str)
{
	int red=0, green=255, blue=0;
	TextColor ((int)alt_text_color->value, &red, &green, &blue);
	//DrawAltString (cx, cy, str, 255);
	while (*str)
	{
		M_DrawCharacter (cx, cy, (*str)+128, red,green,blue,255, false);
		str++;
		cx += MENU_FONT_SIZE;
	}
}

void M_PrintWhite (int cx, int cy, char *str)
{
	//DrawString (cx, cy, str, 255);
	while (*str)
	{
		M_DrawCharacter (cx, cy, *str, 255,255,255,255, false);
		str++;
		cx += MENU_FONT_SIZE;
	}
}

void M_DrawPic (int x, int y, char *pic)
{
	re.DrawPic (x + ((viddef.width - 320)>>1), y + ((viddef.height - 240)>>1), pic);
}


/*
=============
M_DrawCursor

Draws an animating cursor with the point at
x,y.  The pic will extend to the left of x,
and both above and below y.
=============
*/
void M_DrawCursor( int x, int y, int f )
{
	char	cursorname[80];
	static qboolean cached;
	int w,h;

	if ( !cached )
	{
		int i;

		for ( i = 0; i < NUM_CURSOR_FRAMES; i++ )
		{
			Com_sprintf( cursorname, sizeof( cursorname ), "m_cursor%d", i );

			re.RegisterPic( cursorname );
		}
		cached = true;
	}

	Com_sprintf( cursorname, sizeof(cursorname), "m_cursor%d", f );
	re.DrawGetPicSize( &w, &h, cursorname );
	re.DrawStretchPic (	x,	y,	scaledVideo(w), scaledVideo(h), cursorname, 1.0);
	//re.DrawPic( x, y, cursorname );
}

void M_DrawTextBox (int x, int y, int width, int lines)
{
	int		cx, cy;
	int		n;

	// draw left side
	cx = x;
	cy = y;
	M_DrawCharacter (cx, cy, 1, 255,255,255,255, false);
	for (n = 0; n < lines; n++)
	{
		cy += MENU_FONT_SIZE;
		M_DrawCharacter (cx, cy, 4, 255,255,255,255, false);
	}
	M_DrawCharacter (cx, cy+MENU_FONT_SIZE, 7, 255,255,255,255, false);

	// draw middle
	cx += MENU_FONT_SIZE;
	while (width > 0)
	{
		cy = y;
		M_DrawCharacter (cx, cy, 2, 255,255,255,255, false);
		for (n = 0; n < lines; n++)
		{
			cy += MENU_FONT_SIZE;
			M_DrawCharacter (cx, cy, 5, 255,255,255,255, false);
		}
		M_DrawCharacter (cx, cy+MENU_FONT_SIZE, 8, 255,255,255,255, false);
		width -= 1;
		cx += MENU_FONT_SIZE;
	}

	// draw right side
	cy = y;
	M_DrawCharacter (cx, cy, 3, 255,255,255,255, false);
	for (n = 0; n < lines; n++)
	{
		cy += MENU_FONT_SIZE;
		M_DrawCharacter (cx, cy, 6, 255,255,255,255, false);
	}
	M_DrawCharacter (cx, cy+MENU_FONT_SIZE, 9, 255,255,255,255, false);
}

		
/*
=======================================================================

MAIN MENU

=======================================================================
*/
#define	MAIN_ITEMS	5

char *main_names[] =
{
	"m_main_game",
	"m_main_multiplayer",
	"m_main_options",
	"m_main_video",
	"m_main_quit",
	0
};

void findMenuCoords (int *xoffset, int *ystart, int *totalheight, int *widest)
{
	//float ratio;
	int w, h, i;

	*totalheight = 0;
	*widest = -1;

	for ( i = 0; main_names[i] != 0; i++ )
	{
		re.DrawGetPicSize( &w, &h, main_names[i] );

		/*if (w)
		{
			ratio = 32.0/(float)h;
			h = 32;
			w *= ratio;
		}*/

		if ( w > *widest )
			*widest = w;
		*totalheight += ( h + 12 );
	}

	*ystart = ( viddef.height / 2 - scaledVideo(110) );
	*xoffset = ( viddef.width - scaledVideo(*widest) + scaledVideo(70) ) / 2;
//	*xoffset = ( viddef.width  - scaledVideo(*widest) )/ 2 + scaledVideo(30);
}

void M_Main_Draw (void)
{
	int i;
	int w, h, last_h;
	int ystart;
	int	xoffset;
	int widest = -1;
	int totalheight = 0;
	char litname[80];

	findMenuCoords(&xoffset, &ystart, &totalheight, &widest);

	for ( i = 0; main_names[i] != 0; i++ )
	{
		if ( i != m_main_cursor )
		{
			re.DrawGetPicSize( &w, &h, main_names[i] );
			re.DrawStretchPic ( xoffset,
				ystart + scaledVideo(i * 40 + 13),
				scaledVideo(w), scaledVideo(h), main_names[i], 1.0);
			//re.DrawPic( xoffset, ystart + i * 40 + 13, main_names[i] );
		}
	}
	strcpy( litname, main_names[m_main_cursor] );
	strcat( litname, "_sel" );
	re.DrawGetPicSize( &w, &h, litname );
	re.DrawStretchPic (
		xoffset,
		ystart + scaledVideo(m_main_cursor * 40 + 13),
		scaledVideo(w+2), scaledVideo(h+2), litname, 1.0);
	//re.DrawPic( xoffset, ystart + m_main_cursor * 40 + 13, litname );

	M_DrawCursor(	xoffset - scaledVideo(25),
					ystart + scaledVideo(m_main_cursor * 40 + 11),
					(int)(cls.realtime / 100)%NUM_CURSOR_FRAMES );

	re.DrawGetPicSize( &w, &h, "m_main_plaque" );
	re.DrawStretchPic (
		xoffset - scaledVideo(w/2 + 50),
		ystart + scaledVideo(10),
		scaledVideo(w), scaledVideo(h), "m_main_plaque", 1.0);
	//re.DrawPic( xoffset - 30 - w, ystart, "m_main_plaque" );
	last_h = h;

	re.DrawGetPicSize( &w, &h, "m_main_logo" );
	re.DrawStretchPic (
		xoffset - scaledVideo(w/2 + 50),
		ystart + scaledVideo(last_h) + scaledVideo(30),
		scaledVideo(w), scaledVideo(h), "m_main_logo", 1.0);
	//re.DrawPic( xoffset - 30 - w, ystart + h + 5, "m_main_logo" );
}

typedef struct
{
	int	min[2];
	int max[2];

	void (*OpenMenu)(void);
} mainmenuobject_t;

void addButton (mainmenuobject_t *thisObj, int index, int x, int y)
{
	float ratio;
	//char *name;
	int w, h;

	re.DrawGetPicSize( &w, &h, main_names[index]);
	
	if (w)
	{
		ratio = 32.0/(float)h;
		h = 32;
		w *= ratio;
	}

	thisObj->min[0] = x; thisObj->max[0] = x + scaledVideo(w);
	thisObj->min[1] = y; thisObj->max[1] = y + scaledVideo(h);

	switch (index)
	{
	case 0:
		thisObj->OpenMenu = M_Menu_Game_f;
	case 1:
		thisObj->OpenMenu = M_Menu_Multiplayer_f;
	case 2:
		thisObj->OpenMenu = M_Menu_Options_f;
	case 3:
		thisObj->OpenMenu = M_Menu_Video_f;
	case 4:
		thisObj->OpenMenu = M_Menu_Quit_f;
	}
}

void openMenuFromMain (void)
{
	switch (m_main_cursor)
	{
		case 0:
			M_Menu_Game_f ();
			break;

		case 1:
			M_Menu_Multiplayer_f();
			break;

		case 2:
			M_Menu_Options_f ();
			break;

		case 3:
			M_Menu_Video_f ();
			break;

		case 4:
			M_Menu_Quit_f ();
			break;
	}
}

int MainMenuMouseHover;
void CheckMainMenuMouse (void)
{
	int ystart;
	int	xoffset;
	int widest;
	int totalheight;
	int i, oldhover;
	char *sound = NULL;
	mainmenuobject_t buttons[MAIN_ITEMS];

	oldhover = MainMenuMouseHover;
	MainMenuMouseHover = 0;

	findMenuCoords(&xoffset, &ystart, &totalheight, &widest);
	for ( i = 0; main_names[i] != 0; i++ )
		addButton (&buttons[i], i, xoffset, ystart + scaledVideo(i * 40 + 13));

	//Exit with double click 2nd mouse button
	if (!cursor.buttonused[MOUSEBUTTON2] && cursor.buttonclicks[MOUSEBUTTON2]==2)
	{
		M_PopMenu();
		sound = menu_out_sound;
		cursor.buttonused[MOUSEBUTTON2] = true;
		cursor.buttonclicks[MOUSEBUTTON2] = 0;
	}

	for (i=MAIN_ITEMS-1;i>=0;i--)
	{
		if (cursor.x>=buttons[i].min[0] && cursor.x<=buttons[i].max[0] &&
			cursor.y>=buttons[i].min[1] && cursor.y<=buttons[i].max[1])
		{
			if (cursor.mouseaction)
				m_main_cursor = i;

			MainMenuMouseHover = 1 + i;

			if (oldhover == MainMenuMouseHover && MainMenuMouseHover-1 == m_main_cursor &&
				!cursor.buttonused[MOUSEBUTTON1] && cursor.buttonclicks[MOUSEBUTTON1]==1)
			{
				openMenuFromMain();
				sound = menu_move_sound;
				cursor.buttonused[MOUSEBUTTON1] = true;
				cursor.buttonclicks[MOUSEBUTTON1] = 0;
			}
			break;
		}
	}

	if (!MainMenuMouseHover)
	{
		cursor.buttonused[MOUSEBUTTON1] = false;
		cursor.buttonclicks[MOUSEBUTTON1] = 0;
		cursor.buttontime[MOUSEBUTTON1] = 0;
	}

	cursor.mouseaction = false;

	if ( sound )
		S_StartLocalSound( sound );
}

const char *M_Main_Key (int key)
{
	const char *sound = menu_move_sound;

	switch (key)
	{
	case K_ESCAPE:
		M_PopMenu ();
		break;

	case K_KP_DOWNARROW:
	case K_DOWNARROW:
		if (++m_main_cursor >= MAIN_ITEMS)
			m_main_cursor = 0;
		return sound;

	case K_KP_UPARROW:
	case K_UPARROW:
		if (--m_main_cursor < 0)
			m_main_cursor = MAIN_ITEMS - 1;
		return sound;

	case K_KP_ENTER:
	case K_ENTER:
		m_entersound = true;

		switch (m_main_cursor)
		{
		case 0:
			M_Menu_Game_f ();
			break;

		case 1:
			M_Menu_Multiplayer_f();
			break;

		case 2:
			M_Menu_Options_f ();
			break;

		case 3:
			M_Menu_Video_f ();
			break;

		case 4:
			M_Menu_Quit_f ();
			break;
		}
	}

	return NULL;
}


void M_Menu_Main_f (void)
{
	M_PushMenu (M_Main_Draw, M_Main_Key);
}

/*
=======================================================================

MULTIPLAYER MENU

=======================================================================
*/
static menuframework_s	s_multiplayer_menu;
static menuaction_s		s_join_network_server_action;
static menuaction_s		s_start_network_server_action;
static menuaction_s		s_player_setup_action;

static void Multiplayer_MenuDraw (void)
{
	M_Banner( "m_banner_multiplayer" );

	Menu_AdjustCursor( &s_multiplayer_menu, 1 );
	Menu_Draw( &s_multiplayer_menu );
}

static void PlayerSetupFunc( void *unused )
{
	M_Menu_PlayerConfig_f();
}

static void JoinNetworkServerFunc( void *unused )
{
	M_Menu_JoinServer_f();
}

static void StartNetworkServerFunc( void *unused )
{
	M_Menu_StartServer_f ();
}

void Multiplayer_MenuInit( void )
{
	s_multiplayer_menu.x = viddef.width * 0.50 - scaledVideo(8*MENU_FONT_SIZE); //64
	s_multiplayer_menu.nitems = 0;

	s_join_network_server_action.generic.type	= MTYPE_ACTION;
	s_join_network_server_action.generic.flags  = QMF_LEFT_JUSTIFY;
	s_join_network_server_action.generic.x		= 0;
	s_join_network_server_action.generic.y		= 0;
	s_join_network_server_action.generic.name	= " join network server";
	s_join_network_server_action.generic.callback = JoinNetworkServerFunc;

	s_start_network_server_action.generic.type	= MTYPE_ACTION;
	s_start_network_server_action.generic.flags  = QMF_LEFT_JUSTIFY;
	s_start_network_server_action.generic.x		= 0;
	s_start_network_server_action.generic.y		= 2*MENU_FONT_SIZE;	// was 10
	s_start_network_server_action.generic.name	= " start network server";
	s_start_network_server_action.generic.callback = StartNetworkServerFunc;

	s_player_setup_action.generic.type	= MTYPE_ACTION;
	s_player_setup_action.generic.flags  = QMF_LEFT_JUSTIFY;
	s_player_setup_action.generic.x		= 0;
	s_player_setup_action.generic.y		= 4*MENU_FONT_SIZE;	// was 20
	s_player_setup_action.generic.name	= " player setup";
	s_player_setup_action.generic.callback = PlayerSetupFunc;

	Menu_AddItem( &s_multiplayer_menu, ( void * ) &s_join_network_server_action );
	Menu_AddItem( &s_multiplayer_menu, ( void * ) &s_start_network_server_action );
	Menu_AddItem( &s_multiplayer_menu, ( void * ) &s_player_setup_action );

	Menu_SetStatusBar( &s_multiplayer_menu, NULL );

	Menu_Center( &s_multiplayer_menu );
}

const char *Multiplayer_MenuKey( int key )
{
	return Default_MenuKey( &s_multiplayer_menu, key );
}

void M_Menu_Multiplayer_f( void )
{
	Multiplayer_MenuInit();
	M_PushMenu( Multiplayer_MenuDraw, Multiplayer_MenuKey );
}

/*
=======================================================================

KEYS MENU

=======================================================================
*/
char *bindnames[][2] =
{
{"+attack", 		"attack"},

// Knightmare added
{"+attack2", 		"alternate attack"},
{"+use", 			"activate"},
// end Knightmare

{"weapnext", 		"next weapon"},
{"+forward", 		"walk forward"},
{"+back", 			"backpedal"},
{"+left", 			"turn left"},
{"+right", 			"turn right"},
{"+speed", 			"run"},
{"+moveleft", 		"step left"},
{"+moveright", 		"step right"},
{"+strafe", 		"sidestep"},
{"+lookup", 		"look up"},
{"+lookdown", 		"look down"},
{"centerview", 		"center view"},
{"+mlook", 			"mouse look"},
{"+klook", 			"keyboard look"},
{"+moveup",			"up / jump"},
{"+movedown",		"down / crouch"},
{"inven",			"inventory"},
{"invuse",			"use item"},
{"invdrop",			"drop item"},
{"invprev",			"prev item"},
{"invnext",			"next item"},
{"cmd help", 		"help computer" }, 
{ 0, 0 }
};

int				keys_cursor;
static int		bind_grab;

static menuframework_s	s_keys_menu;
static menuaction_s s_keys_binds[64];

static void M_UnbindCommand (char *command)
{
	int		j;
	int		l;
	char	*b;

	l = strlen(command);

	for (j=0 ; j<256 ; j++)
	{
		b = keybindings[j];
		if (!b)
			continue;
		if (!strncmp (b, command, l) )
			Key_SetBinding (j, "");
	}
}

static void M_FindKeysForCommand (char *command, int *twokeys)
{
	int		count;
	int		j;
	int	 	l;
	char	*b;

	twokeys[0] = twokeys[1] = -1;
	l = strlen(command);
	count = 0;

	for (j=0 ; j<256 ; j++)
	{
		b = keybindings[j];
		if (!b)
			continue;
		// Knightmare- fix bug with key bound to +attack2 being confused with +attack
		if (!strncmp (b, "+attack2", 8) && strncmp (command, "+attack2", 8))
			continue;
		if (!strncmp (b, command, l))
		{
			twokeys[count] = j;
			count++;
			if (count == 2)
				break;
		}
	}
}

static void KeyCursorDrawFunc( menuframework_s *menu )
{
	if ( bind_grab )
		re.DrawScaledChar ( menu->x, menu->y + scaledVideo(menu->cursor * (MENU_FONT_SIZE+2) ),
							'=', videoScale(), 255,255,255,255, false);
		/*re.DrawChar(
		menu->x, menu->y + scaledVideo(menu->cursor * 9),
		'=', 255 );*/
	else
		re.DrawScaledChar ( menu->x, menu->y + scaledVideo(menu->cursor * (MENU_FONT_SIZE+2) ),
							12 + ( ( int ) ( Sys_Milliseconds() / 250 ) & 1 ),
							videoScale(), 255,255,255,255, false);
		/*re.DrawChar(
		menu->x, menu->y + scaledVideo(menu->cursor * 9),
		12 + ( ( int ) ( Sys_Milliseconds() / 250 ) & 1 ), 255 );*/
}

static void DrawKeyBindingFunc( void *self )
{
	int keys[2];
	menuaction_s *a = ( menuaction_s * ) self;

	M_FindKeysForCommand( bindnames[a->generic.localdata[0]][0], keys);
		
	if (keys[0] == -1)
	{
		Menu_DrawString(
			scaledVideo(a->generic.x + 16) + a->generic.parent->x,
			scaledVideo(a->generic.y) + a->generic.parent->y,
			"???", 255 );
	}
	else
	{
		int x;
		const char *name;

		name = Key_KeynumToString (keys[0]);

		Menu_DrawString(
			scaledVideo(a->generic.x + 16) + a->generic.parent->x,
			scaledVideo(a->generic.y) + a->generic.parent->y,
			name , 255);

		x = scaledVideo(strlen(name) * MENU_FONT_SIZE);

		if (keys[1] != -1)
		{
			Menu_DrawString(
				scaledVideo(a->generic.x + MENU_FONT_SIZE*3 + x) + a->generic.parent->x,
				scaledVideo(a->generic.y) + a->generic.parent->y,
				"or", 255 );
			Menu_DrawString(
				scaledVideo(a->generic.x + MENU_FONT_SIZE*6 + x) + a->generic.parent->x,
				scaledVideo(a->generic.y) + a->generic.parent->y,
				Key_KeynumToString (keys[1]), 255 );
		}
	}
}

static void KeyBindingFunc( void *self )
{
	menuaction_s *a = ( menuaction_s * ) self;
	int keys[2];

	M_FindKeysForCommand( bindnames[a->generic.localdata[0]][0], keys );

	if (keys[1] != -1)
		M_UnbindCommand( bindnames[a->generic.localdata[0]][0]);

	bind_grab = true;

	Menu_SetStatusBar( &s_keys_menu, "press a key or button for this action" );
}

int listSize (char* list[][2])
{
	int i=0;
	while (list[i][1])
		i++;

	return i;	
}

void addBindOption (int i, char* list[][2])
{		
	s_keys_binds[i].generic.type	= MTYPE_ACTION;
	s_keys_binds[i].generic.flags  = QMF_GRAYED;
	s_keys_binds[i].generic.x		= 0;
	s_keys_binds[i].generic.y		= i*(MENU_FONT_SIZE+2);
	s_keys_binds[i].generic.ownerdraw = DrawKeyBindingFunc;
	s_keys_binds[i].generic.localdata[0] = i;
	s_keys_binds[i].generic.name	= list[s_keys_binds[i].generic.localdata[0]][1];
	s_keys_binds[i].generic.callback = KeyBindingFunc;

	if (strstr ("MENUSPACE", list[i][0]))
		s_keys_binds[i].generic.type	= MTYPE_SEPARATOR;
}

static void Keys_MenuInit( void )
{
	int BINDS_MAX;
	int i = 0;

	s_keys_menu.x = viddef.width * 0.50;
	s_keys_menu.y = viddef.height * 0.50 - scaledVideo(7.25*MENU_FONT_SIZE);
	s_keys_menu.nitems = 0;
	s_keys_menu.cursordraw = KeyCursorDrawFunc;

	BINDS_MAX = listSize(bindnames);
	for (i=0;i<BINDS_MAX;i++)
		addBindOption(i, bindnames);

	for (i=0;i<BINDS_MAX;i++)
		Menu_AddItem( &s_keys_menu, ( void * ) &s_keys_binds[i] );
	
	Menu_SetStatusBar( &s_keys_menu, "enter or mouse1 to change, backspace to clear" );
	// Don't center it- it's too large
	//Menu_Center( &s_keys_menu );
}

static void Keys_MenuDraw (void)
{
	M_Banner( "m_banner_customize" ); // Knightmare added
	Menu_AdjustCursor( &s_keys_menu, 1 );
	Menu_Draw( &s_keys_menu );
}

static const char *Keys_MenuKey( int key )
{
	menuaction_s *item = ( menuaction_s * ) Menu_ItemAtCursor( &s_keys_menu );

	//pressing mouse1 to pick a new bind wont force bind/unbind itself - spaz
	if ( bind_grab && !(cursor.buttonused[MOUSEBUTTON1]&&key==K_MOUSE1))
	{	
		if ( key != K_ESCAPE && key != '`' )
		{
			char cmd[1024];

			Com_sprintf (cmd, sizeof(cmd), "bind \"%s\" \"%s\"\n", Key_KeynumToString(key), bindnames[item->generic.localdata[0]][0]);
			Cbuf_InsertText (cmd);
		}
		
		// Knightmare- added Psychospaz's mouse support
		//dont let selecting with mouse buttons screw everything up
		refreshCursorButtons();
		if (key==K_MOUSE1)
			cursor.buttonclicks[MOUSEBUTTON1] = -1;

		Menu_SetStatusBar( &s_keys_menu, "enter to change, backspace to clear" );
		bind_grab = false;
		return menu_out_sound;
	}

	switch ( key )
	{
	case K_KP_ENTER:
	case K_ENTER:
		KeyBindingFunc( item );
		return menu_in_sound;
	case K_BACKSPACE:		// delete bindings
	case K_DEL:				// delete bindings
	case K_KP_DEL:
		M_UnbindCommand( bindnames[item->generic.localdata[0]][0] );
		return menu_out_sound;
	default:
		return Default_MenuKey( &s_keys_menu, key );
	}
}

void M_Menu_Keys_f (void)
{
	Keys_MenuInit();
	M_PushMenu( Keys_MenuDraw, Keys_MenuKey );
}


/*
=======================================================================

CONTROLS MENU

=======================================================================
*/
static cvar_t *win_noalttab;
extern cvar_t *in_joystick;

static menuframework_s	s_options_menu;
static menuframework_s	s_options_sound_menu;
static menuframework_s	s_options_controls_menu;
static menuframework_s	s_options_interface_menu;
static menuframework_s	s_options_ingame_menu;

static menuaction_s		s_options_sound_section;
static menuaction_s		s_options_controls_section;
static menuaction_s		s_options_interface_section;
static menuaction_s		s_options_ingame_section;

static menuseparator_s		s_options_sound_header;
static menuseparator_s		s_options_controls_header;
static menuseparator_s		s_options_ingame_header;
static menuseparator_s		s_options_interface_header;

static menulist_s		s_options_noalttab_box;
static menuaction_s		s_options_defaults_action;
static menulist_s		s_options_console_action;
static menulist_s		s_options_backmain_action;

//sound
static menuslider_s		s_options_sfxvolume_slider;
static menulist_s		s_options_cdvolume_box;
static menulist_s		s_options_quality_list;
static menulist_s		s_options_compatibility_list;

// controls
static menuslider_s		s_options_sensitivity_slider;
static menulist_s		s_options_alwaysrun_box;
// Knightmare- added Psychospaz's chaseam
static menulist_s		s_options_thirdperson_box;
static menuslider_s		s_options_thirdperson_distance_slider;
static menuslider_s		s_options_thirdperson_angle_slider;
//end Knightmare
static menulist_s		s_options_invertmouse_box;
static menulist_s		s_options_lookspring_box;
static menulist_s		s_options_lookstrafe_box;
static menulist_s		s_options_freelook_box;
static menulist_s		s_options_joystick_box;
static menuaction_s		s_options_customize_options_action;

//ingame
static menulist_s		s_options_blood_box;
static menulist_s		s_options_crosshair_box;
// Knightmare- Psychospaz's changeable size crosshair
static menuslider_s		s_options_crosshairscale_slider;
static menuslider_s		s_options_hudscale_slider;
static menuslider_s		s_options_hudalpha_slider;
// Knightmare- particle effect complexity option
static menuslider_s		s_options_particle_comp_slider;
static menulist_s		s_options_fps_box;
// Knightmare- foostep override option
static menulist_s		s_options_footstep_box;
// Knightmare- Psychospaz's changeable rail trail
static menulist_s		s_options_railtrail_box;
static menuslider_s		s_options_railcolor_box[3];
static menuaction_s		s_options_ingame_defaults_action;
static menulist_s		s_options_ingame_console_action;
static menulist_s		s_options_ingame_backmain_action;

// interface
static menuslider_s		s_options_conalpha_slider;
static menuslider_s		s_options_conheight_slider;
static menuslider_s		s_options_menumouse_slider;
static menulist_s		s_options_menuscale_box;
static menulist_s		s_options_font_box;
static menuslider_s		s_options_fontsize_slider;
static menulist_s		s_options_alt_text_color_box;

static void BloodFunc( void *unused )
{
	Cvar_SetValue( "cl_blood", s_options_blood_box.curvalue );
}

/*static void OldCrosshairFunc( void *unused )
{
	Cvar_SetValue( "crosshair", s_options_crosshair_box.curvalue );
}*/

//Knightmare- Psychospaz's changeable size crosshair
static void CrosshairSizeFunc( void *unused )
{
	Cvar_SetValue( "crosshair_scale", s_options_crosshairscale_slider.curvalue*0.25);
}

// Knightmare- hud scaling option
static void HudScaleFunc( void *unused )
{
	Cvar_SetValue( "hud_scale", s_options_hudscale_slider.curvalue-1);
}

// Knightmare- hud trans option
static void HudAlphaFunc( void *unused )
{
	Cvar_SetValue( "hud_alpha", (s_options_hudalpha_slider.curvalue-1)/10);
}

// Knightmare- particle effect complexity option
static void ParticleCompFunc( void *unused )
{
	Cvar_SetValue( "cl_particle_scale", (s_options_particle_comp_slider.curvalue-3)*-1+3);
}

// Knightmare- FPS counter option
static void FPSFunc( void *unused )
{
	Cvar_SetValue( "cl_drawfps", s_options_fps_box.curvalue);
}

// Knightmare- foostep override option
static void FootStepFunc( void *unused )
{
	Cvar_SetValue( "cl_footstep_override", s_options_footstep_box.curvalue );
}

//Knightmare- Psychospaz's changeable rail trail
static void RailTrailFunc( void *unused )
{
	Cvar_SetValue( "cl_railtype", s_options_railtrail_box.curvalue );
}
static void RailColorRedFunc( void *unused )
{
	Cvar_SetValue( "cl_railred", s_options_railcolor_box[0].curvalue*16 );
}
static void RailColorGreenFunc( void *unused )
{
	Cvar_SetValue( "cl_railgreen", s_options_railcolor_box[1].curvalue*16 );
}
static void RailColorBlueFunc( void *unused )
{
	Cvar_SetValue( "cl_railblue", s_options_railcolor_box[2].curvalue*16 );
}
//end Knightmare

static void JoystickFunc( void *unused )
{
	Cvar_SetValue( "in_joystick", s_options_joystick_box.curvalue );
}

static void CustomizeControlsFunc( void *unused )
{
	M_Menu_Keys_f();
}

static void AlwaysRunFunc( void *unused )
{
	Cvar_SetValue( "cl_run", s_options_alwaysrun_box.curvalue );
}

// Knigntmare- added Psychospaz's chaseam
static void ThirdPersonFunc( void *unused )
{
	Cvar_SetValue( "cl_3dcam", s_options_thirdperson_box.curvalue );
}

static void ThirdPersonDistFunc( void *unused )
{
	Cvar_SetValue( "cl_3dcam_dist", (int)(s_options_thirdperson_distance_slider.curvalue * 25) );
}

static void ThirdPersonAngleFunc( void *unused )
{
	Cvar_SetValue( "cl_3dcam_angle", (int)(s_options_thirdperson_angle_slider.curvalue * 10) );
}

//end Knightmare

static void FreeLookFunc( void *unused )
{
	Cvar_SetValue( "freelook", s_options_freelook_box.curvalue );
}

static void MouseSpeedFunc( void *unused )
{
	Cvar_SetValue( "sensitivity", s_options_sensitivity_slider.curvalue / 2.0F );
}

// Knightmare- Psychospaz's menu mouse support
static void MouseMenuFunc( void *unused )
{
	Cvar_SetValue( "menu_sensitivity", s_options_menumouse_slider.curvalue / 4.0F );
}

// Knightmare- menu scaling option
static void MenuScaleFunc( void *unused )
{
	Cvar_SetValue( "menu_scale", s_options_menuscale_box.curvalue);
}

static void AltTextColorFunc( void *unused )
{
	Cvar_SetValue( "alt_text_color", s_options_alt_text_color_box.curvalue);
}

// Knightmare- Psychospaz's transparent console
static void ConAlphaFunc( void *unused )
{
	Cvar_SetValue( "con_alpha", s_options_conalpha_slider.curvalue * 0.05 );
}
// Kngihtmare- variable console height
static void ConHeightFunc( void *unused )
{
	Cvar_SetValue( "con_height", 0.25 + (s_options_conheight_slider.curvalue * 0.05) );
}
// end Knightmare

static void NoAltTabFunc( void *unused )
{
	Cvar_SetValue( "win_noalttab", s_options_noalttab_box.curvalue );
}

static float ClampCvar( float min, float max, float value )
{
	if ( value < min ) return min;
	if ( value > max ) return max;
	return value;
}

/*
=======================================================================
Font loading
=======================================================================
*/

static void FreeFileList( char **list, int n )
{
	int i;

	for ( i = 0; i < n; i++ )
	{
		if ( list && list[i] )
		{
			free( list[i] );
			list[i] = 0;
		}
	}
	free( list );
}

cvar_t *con_font;
#define MAX_FONTS 32
char **font_names;
int	numfonts;

static void FontSizeFunc( void *unused )
{
	Cvar_SetValue( "con_font_size", s_options_fontsize_slider.curvalue * 4 );
}

static void FontFunc( void *unused )
{
	Cvar_Set( "con_font", font_names[s_options_font_box.curvalue] );
}

void SetFontCursor (void)
{
	int i;
	s_options_font_box.curvalue = 0;

	if (!con_font)
		con_font = Cvar_Get ("con_font", "default", CVAR_ARCHIVE);

	if (numfonts>1)
		for (i=0; font_names[i]; i++)
		{
			if (!Q_strcasecmp(con_font->string, font_names[i]))
			{
				s_options_font_box.curvalue = i;
				return;
			}
		}
}

qboolean fontInList (char *check, int num, char **list)
{
	int i;
	for (i=0;i<num;i++)
		if (!Q_strcasecmp(check, list[i]))
			return true;
	return false;
}

void insertFont (char ** list, char *insert, int len )
{
	int i, j;

	if (!list)
		return;

	//i=1 so default stays first!
	for (i=1; i<len; i++)
	{
		if (!list[i])
			break;

		if (strcmp( list[i], insert ))
		{
			for (j=len; j>i ;j--)
				list[j] = list[j-1];

			list[i] = strdup(insert);

			return;
		}
	}

	list[len] = strdup(insert);
}

char **SetFontNames (void)
{
	char *curFont;
	char **list = 0, *p;//, *s;
	char findname[1024];
	int nfonts = 0, nfontnames;
	char **fontfiles;
	char *path = NULL;
	int i;//, j;
	extern char **FS_ListFiles( char *, int *, unsigned, unsigned );

	list = malloc( sizeof( char * ) * MAX_FONTS );
	memset( list, 0, sizeof( char * ) * MAX_FONTS );

	list[0] = strdup("default");

	nfontnames = 1;

	path = FS_NextPath( path );
	while (path) 
	{
		Com_sprintf( findname, sizeof(findname), "%s/fonts/*.*", path );
		fontfiles = FS_ListFiles( findname, &nfonts, 0, SFF_SUBDIR | SFF_HIDDEN | SFF_SYSTEM );

		for (i=0;i<nfonts && nfontnames<MAX_FONTS;i++)
		{
			int num;

			if (!fontfiles || !fontfiles[i])	// Knightmare added array base check
				continue;

			p = strstr(fontfiles[i], "/fonts/"); p++;
			p = strstr(p, "/"); p++;

			if (	!strstr(p, ".tga")
				&&	!strstr(p, ".jpg")
				&&	!strstr(p, ".pcx")
				)
				continue;

			num = strlen(p)-4;
			p[num] = 0;//NULL;

			curFont = p;

			if (!fontInList(curFont, nfontnames, list))
			{
				insertFont(list, strdup(curFont),nfontnames);
				nfontnames++;
			}
			
			//set back so whole string get deleted.
			p[num] = '.';
		}
		if (nfonts)
			FreeFileList( fontfiles, nfonts );
		
		path = FS_NextPath( path );
	}

	//check pak after
	if (fontfiles = FS_ListPak("fonts/", &nfonts))
	{
		for (i=0;i<nfonts && nfontnames<MAX_FONTS;i++)
		{

			int num;

			if (!fontfiles || !fontfiles[i])	// Knightmare added array base check
				continue;

			p = strstr(fontfiles[i], "/"); p++;

			if (	!strstr(p, ".tga")
				&&	!strstr(p, ".jpg")
				&&	!strstr(p, ".pcx")
				)
				continue;

			num = strlen(p)-4;
			p[num] = 0; //NULL;

			curFont = p;

			if (!fontInList(curFont, nfontnames, list))
			{
				insertFont(list, strdup(curFont),nfontnames);
				nfontnames++;
			}
			
			//set back so whole string get deleted.
			p[num] = '.';

		}
	}
	if (nfonts)
		FreeFileList( fontfiles, nfonts );

	numfonts = nfontnames;

	return list;		
}

/*
=======================================================================
Crosshair loading
=======================================================================
*/

#define MAX_CROSSHAIRS 100
char **crosshair_names;
int	numcrosshairs;

static void CrosshairFunc( void *unused )
{
	if (s_options_crosshair_box.curvalue == 0) {
		Cvar_SetValue( "crosshair", 0); return; }
	else
		Cvar_SetValue( "crosshair", atoi(strdup(crosshair_names[s_options_crosshair_box.curvalue]+2)) );
	//Cvar_SetValue( "crosshair", s_options_crosshair_box.curvalue );
}

void SetCrosshairCursor (void)
{
	int i;
	s_options_crosshair_box.curvalue = 0;

	if (numcrosshairs > 1)
		for (i=0; crosshair_names[i]; i++)
		{
			if (!Q_strcasecmp(va("ch%i", (int)crosshair->value), crosshair_names[i]))
			{
				s_options_crosshair_box.curvalue = i;
				return;
			}
		}
}

qboolean crosshairInList (char *check, int num, char **list)
{
	int i;
	for (i=0; i<num; i++)
		if (!Q_strcasecmp(check, list[i]))
			return true;
	return false;
}

void insertCrosshair (char ** list, char *insert, int len )
{
	int i;
	if (!list) return;

	//i=1 so none stays first!
	for (i=1; i<len; i++)
	{
		if (!list[i])
		{
			list[i] = strdup(insert);
			return;
		}
	}
	list[len] = strdup(insert);
}

void sortCrosshairs (char ** list, int len )
{
	int i, j;
	char *temp;
	qboolean moved;

	if (!list || len < 2)
		return;

	for (i=(len-1); i>0; i--)
	{
		moved = false;
		for (j=0; j<i; j++)
		{
			if (!list[j]) break;
			if ( atoi(strdup(list[j]+2)) > atoi(strdup(list[j+1]+2)) )
			{
				temp = list[j];
				list[j] = list[j+1];
				list[j+1] = temp;
				moved = true;
			}
		}
		if (!moved) break; //done sorting
	}
}

qboolean isNumeric (char ch)
{
	if (ch >= '0' && ch <= '9')
		return true;
	else return false;
}

char **SetCrosshairNames (void)
{
	char *curCrosshair;
	char **list = 0, *p;
	char findname[1024];
	int ncrosshairs = 0, ncrosshairnames;
	char **crosshairfiles;
	char *path = NULL;
	int i;
	extern char **FS_ListFiles( char *, int *, unsigned, unsigned );

	list = malloc( sizeof( char * ) * MAX_CROSSHAIRS+1 );
	memset( list, 0, sizeof( char * ) * MAX_CROSSHAIRS+1 );

	list[0] = strdup("none"); //was default
	ncrosshairnames = 1;

	path = FS_NextPath( path );
	while (path) 
	{
		Com_sprintf( findname, sizeof(findname), "%s/pics/ch*.*", path );
		crosshairfiles = FS_ListFiles( findname, &ncrosshairs, 0, SFF_SUBDIR | SFF_HIDDEN | SFF_SYSTEM );

		for (i=0; i < ncrosshairs && ncrosshairnames < MAX_CROSSHAIRS; i++)
		{
			int num, namelen;

			if (!crosshairfiles || !crosshairfiles[i])
				continue;

			p = strstr(crosshairfiles[i], "/pics/"); p++;
			p = strstr(p, "/"); p++;

			if (	!strstr(p, ".tga")
				&&	!strstr(p, ".jpg")
				&&	!strstr(p, ".pcx")
				)
				continue;

			// filename must be chxxx
			if (strncmp(p, "ch", 2)) 
				continue;
			namelen = strlen(strdup(p));
			if (namelen < 7 || namelen > 9)
				continue;
			if (!isNumeric(p[2]))
				continue;
			if (namelen >= 8 && !isNumeric(p[3]))
				continue;
			// ch100 is only valid 5-char name
			if (namelen == 9 && (p[2] != '1' || p[3] != '0' || p[4] != '0'))
				continue;

			num = strlen(p)-4;
			p[num] = 0; //NULL;

			curCrosshair = p;

			if (!fontInList(curCrosshair, ncrosshairnames, list))
			{
				insertCrosshair(list, strdup(curCrosshair), ncrosshairnames);
				ncrosshairnames++;
			}
			
			//set back so whole string get deleted.
			p[num] = '.';
		}
		if (ncrosshairs)
			FreeFileList( crosshairfiles, ncrosshairs );
		
		path = FS_NextPath( path );
	}

	//check pak after
	if (crosshairfiles = FS_ListPak("pics/", &ncrosshairs))
	{
		for (i=0; i<ncrosshairs && ncrosshairnames < MAX_CROSSHAIRS; i++)
		{
			int num, namelen;

			if (!crosshairfiles || !crosshairfiles[i])
				continue;

			p = strstr(crosshairfiles[i], "/"); p++;

			if (	!strstr(p, ".tga")
				&&	!strstr(p, ".jpg")
				&&	!strstr(p, ".pcx")
				)
				continue;

			// filename must be chxxx
			if (strncmp(p, "ch", 2))
				continue;
			namelen = strlen(strdup(p));
			if (namelen < 7 || namelen > 9)
				continue;
			if (!isNumeric(p[2]))
				continue;
			if (namelen >= 8 && !isNumeric(p[3]))
				continue;
			// ch100 is only valid 5-char name
			if (namelen == 9 && (p[2] != '1' || p[3] != '0' || p[4] != '0'))
				continue;

			num = strlen(p)-4;
			p[num] = 0; //NULL;

			curCrosshair = p;

			if (!crosshairInList(curCrosshair, ncrosshairnames, list))
			{
				insertCrosshair(list, strdup(curCrosshair), ncrosshairnames);
				ncrosshairnames++;
			}
			
			//set back so whole string get deleted.
			p[num] = '.';
		}
	}
	// sort the list
	sortCrosshairs (list, ncrosshairnames);

	if (ncrosshairs)
		FreeFileList( crosshairfiles, ncrosshairs );

	numcrosshairs = ncrosshairnames;

	return list;		
}

//=======================================================================

static void ControlsSetMenuItemValues( void )
{
	s_options_sfxvolume_slider.curvalue		= Cvar_VariableValue( "s_volume" ) * 10;
	s_options_cdvolume_box.curvalue 		= !Cvar_VariableValue("cd_nocd");

	//s_options_quality_list.curvalue			= !Cvar_VariableValue( "s_loadas8bit" );
	//**  DMP convert setting into index for option display text
	switch((int)Cvar_VariableValue("s_khz"))
	{
	case 48:  s_options_quality_list.curvalue = 3;  break;
	case 44:  s_options_quality_list.curvalue = 2;  break;
	case 22:  s_options_quality_list.curvalue = 1;  break;
	default:  s_options_quality_list.curvalue = 0;  break;
	}
	//** DMP end sound menu changes

	SetFontCursor();
	SetCrosshairCursor();

	s_options_sensitivity_slider.curvalue	= ( sensitivity->value ) * 2;

	// Knightmare- Psychospaz's menu mouse support
	s_options_menumouse_slider.curvalue		= ( menu_sensitivity->value ) * 4;
	// Knightmare- menu scaling option
	s_options_menuscale_box.curvalue		= menu_scale->value;
	// font size option
	s_options_fontsize_slider.curvalue		= ( con_font_size->value ) * 0.25;
	// Knightmare- alt text color option
	Cvar_SetValue( "alt_text_color", ClampCvar( 0, 9, alt_text_color->value ) );
	s_options_alt_text_color_box.curvalue	= alt_text_color->value;

	// Knightmare- Psychospaz's transparent console cvar
	s_options_conalpha_slider.curvalue		= ( con_alpha->value ) * 20;
	// Knightmare- variable console height
	Cvar_SetValue( "con_height", ClampCvar( 0.25, 0.75, con_height->value ) );
	s_options_conheight_slider.curvalue		= 20 * (con_height->value - 0.25);

	Cvar_SetValue( "cl_run", ClampCvar( 0, 1, cl_run->value ) );
	s_options_alwaysrun_box.curvalue		= cl_run->value;

	// Knightmare- added Psychospaz's chasecam
	Cvar_SetValue( "cl_3dcam", ClampCvar( 0, 1, cl_3dcam->value ) );
	s_options_thirdperson_box.curvalue		= cl_3dcam->value;
	s_options_thirdperson_distance_slider.curvalue	= cl_3dcam_dist->value / 25;
	s_options_thirdperson_angle_slider.curvalue	= cl_3dcam_angle->value / 10;
	// end Knightmare

	s_options_invertmouse_box.curvalue		= m_pitch->value < 0;

	Cvar_SetValue( "lookspring", ClampCvar( 0, 1, lookspring->value ) );
	s_options_lookspring_box.curvalue		= lookspring->value;

	Cvar_SetValue( "lookstrafe", ClampCvar( 0, 1, lookstrafe->value ) );
	s_options_lookstrafe_box.curvalue		= lookstrafe->value;

	Cvar_SetValue( "freelook", ClampCvar( 0, 1, freelook->value ) );
	s_options_freelook_box.curvalue			= freelook->value;

	Cvar_SetValue( "cl_blood", ClampCvar( 0, 3, cl_blood->value ) );
	s_options_blood_box.curvalue = cl_blood->value;

	Cvar_SetValue( "crosshair", ClampCvar( 0, 100, crosshair->value ) );
	//s_options_crosshair_box.curvalue			= crosshair->value;
	s_options_crosshairscale_slider.curvalue	= crosshair_scale->value*4;

	Cvar_SetValue( "hud_scale", ClampCvar( 0, 7, hud_scale->value ) );
	s_options_hudscale_slider.curvalue		= hud_scale->value+1;

	Cvar_SetValue( "hud_alpha", ClampCvar( 0, 7, hud_alpha->value ) );
	s_options_hudalpha_slider.curvalue		= hud_alpha->value*10 + 1;

	Cvar_SetValue( "cl_particle_scale", ClampCvar( 0, 5, cl_particle_scale->value ) );
	s_options_particle_comp_slider.curvalue			= (cl_particle_scale->value-3)*-1+3;

	Cvar_SetValue( "cl_drawfps", ClampCvar( 0, 1, cl_drawfps->value ) );
	s_options_fps_box.curvalue				= cl_drawfps->value;

	// Knightmare- foostep override option
	Cvar_SetValue( "cl_footstep_override", ClampCvar( 0, 1, cl_footstep_override->value ) );
	s_options_footstep_box.curvalue			= cl_footstep_override->value;

	//Knightmare- Psychospaz's changeable rail trail
	Cvar_SetValue( "cl_railtype", ClampCvar( 0, 2, cl_railtype->value ) );
	s_options_railtrail_box.curvalue		= cl_railtype->value;
	s_options_railcolor_box[0].curvalue		= cl_railred->value/16;
	s_options_railcolor_box[1].curvalue		= cl_railgreen->value/16;
	s_options_railcolor_box[2].curvalue		= cl_railblue->value/16;
	//end Knightmare

	Cvar_SetValue( "in_joystick", ClampCvar( 0, 1, in_joystick->value ) );
	s_options_joystick_box.curvalue		= in_joystick->value;

	s_options_noalttab_box.curvalue			= win_noalttab->value;
}

static void ControlsResetDefaultsFunc( void *unused )
{
	Cbuf_AddText ("exec default.cfg\n");
	Cbuf_Execute();

	ControlsSetMenuItemValues();
}

static void InvertMouseFunc( void *unused )
{
	Cvar_SetValue( "m_pitch", -m_pitch->value );
}

static void LookspringFunc( void *unused )
{
	Cvar_SetValue( "lookspring", !lookspring->value );
}

static void LookstrafeFunc( void *unused )
{
	Cvar_SetValue( "lookstrafe", !lookstrafe->value );
}

static void UpdateVolumeFunc( void *unused )
{
	Cvar_SetValue( "s_volume", s_options_sfxvolume_slider.curvalue / 10 );
}

static void UpdateCDVolumeFunc( void *unused )
{
	Cvar_SetValue( "cd_nocd", !s_options_cdvolume_box.curvalue );
}

//Knightmare- Psychospaz's menu select box
static void setMenuOptionsMain ( void *unused )
{
	Cvar_SetValue( "options_menu", 0 );
	refreshCursorLink();
}

static void setMenuSound ( void *unused )
{
	Cvar_SetValue( "options_menu", 1 );
	refreshCursorLink();
	// release mouse button 1 to not inadvertently affect sliders
	cursor.buttondown[MOUSEBUTTON1] = false;
}

static void setMenuControls ( void *unused )
{
	Cvar_SetValue( "options_menu", 2 );
	refreshCursorLink();
	// release mouse button 1 to not inadvertently affect sliders
	cursor.buttondown[MOUSEBUTTON1] = false; 
}

static void setMenuIngame ( void *unused )
{
	Cvar_SetValue( "options_menu", 3 );
	refreshCursorLink();
	// release mouse button 1 to not inadvertently affect sliders
	cursor.buttondown[MOUSEBUTTON1] = false; 
}

static void setMenuInterface ( void *unused )
{
	Cvar_SetValue( "options_menu", 4 );
	refreshCursorLink();
	// release mouse button 1 to not inadvertently affect sliders
	cursor.buttondown[MOUSEBUTTON1] = false; 
}
//end Knightmare

static void ConsoleFunc( void *unused )
{
	/*
	** the proper way to do this is probably to have ToggleConsole_f accept a parameter
	*/
	extern void Key_ClearTyping( void );

	if ( cl.attractloop )
	{
		Cbuf_AddText ("killserver\n");
		return;
	}

	Key_ClearTyping ();
	Con_ClearNotify ();

	//M_ForceMenuOff ();
	cls.consoleActive = true; // Knigthmare changed
	//cls.key_dest = key_console;
}

static void UpdateSoundQualityFunc( void *unused )
{
	//Knightmare- added DMP's 44/48 KHz sound support
	//** DMP check the newly added sound quality menu options
	switch (s_options_quality_list.curvalue)
	{
	case 1:
		Cvar_SetValue( "s_khz", 22 );
		Cvar_SetValue( "s_loadas8bit", false );
		break;
	case 2:
		Cvar_SetValue( "s_khz", 44 );
		Cvar_SetValue( "s_loadas8bit", false );
		break;
	case 3:
		Cvar_SetValue( "s_khz", 48 );
		Cvar_SetValue( "s_loadas8bit", false );
		break;
	default:
		Cvar_SetValue( "s_khz", 11 );
		Cvar_SetValue( "s_loadas8bit", true );
		break;
	}
	//** DMP end sound menu changes

	Cvar_SetValue( "s_primary", s_options_compatibility_list.curvalue );

	//viddef.width/2 - scaledVideo(38*MENU_FONT_SIZE)/2,

	M_DrawTextBox( 8, 120 - 48, 36, 3 );
	M_Print( 28, 120 - 48 + MENU_FONT_SIZE,  "Restarting the sound system. This" );
	M_Print( 28, 120 - 48 + MENU_FONT_SIZE*2, "could take up to a minute, so" );
	M_Print( 28, 120 - 48 + MENU_FONT_SIZE*3, "please be patient." );

	// the text box won't show up unless we do a buffer swap
	re.EndFrame();

	CL_Snd_Restart_f();
}

void Options_MenuInit( void )
{
	//Knightmare- Psychospaz's menu select box
	static const char *menu_names[] =
	{
		"sound",
		"controls",
		"ingame",
		"interface",
		0
	};

	static const char *cd_music_items[] =
	{
		"disabled",
		"enabled",
		0
	};

	static const char *quality_items[] =
	{
		"low (11KHz/8-bit)",			//** DMP - changed text
		"normal (22KHz/16-bit)",		//** DMP - changed text
		"high (44KHz/16-bit)",			//** DMP - added 44 Khz menu item
		"highest (48KHz/16-bit)",		//** DMP - added 48 Khz menu item
		0
	};

	static const char *compatibility_items[] =
	{
		"max compatibility", "max performance", 0
	};

	static const char *yesno_names[] =
	{
		"no",
		"yes",
		0
	};

	static const char *textcolor_names[] =
	{
		"gray",
		"red",
		"green",
		"yellow",
		"blue",
		"teal",
		"pink",
		"white",
		"black",
		"orange",
		0
	};

	/*static const char *old_crosshair_names[] =
	{
		"none",
		"cross 1",
		"dot",
		"angle",
		"cross 2",
		"cross 3",
		"cross 4",
		"trans disc",
		"trans circle",
		0
	};*/

	//Knightmare- Psychospaz's changeable rail trail
	static const char *railtrail_names[] =
	{
		"colored spiral",
		"colored beam", //laser
		"colored devrail",
		0
	};

	static const char *blood_names[] =
	{
		"puff",
		"splat",
		"bleed",
		"gore",
		0
	};

	int y;

	if (!options_menu)
		options_menu = Cvar_Get ("options_menu", "0", 0);

	win_noalttab = Cvar_Get( "win_noalttab", "0", CVAR_ARCHIVE);

	/*
	** configure controls menu and menu items
	*/
	/*
	s_options_menu.x = viddef.width / 2;
	s_options_menu.y = viddef.height / 2 - 58;
	s_options_menu.nitems = 0;
	*/
	s_options_menu.x = viddef.width * 0.50;
	s_options_ingame_menu.x = s_options_menu.x;
	s_options_sound_menu.x = s_options_menu.x;
	s_options_controls_menu.x = s_options_menu.x;
	s_options_interface_menu.x = s_options_menu.x;

	s_options_menu.y = viddef.height * 0.50 - scaledVideo(7.25*MENU_FONT_SIZE); //58
	s_options_ingame_menu.y = s_options_menu.y;
	s_options_sound_menu.y = s_options_menu.y;
	s_options_controls_menu.y = s_options_menu.y;
	s_options_interface_menu.y = s_options_menu.y;

	s_options_menu.nitems = 0;
	s_options_sound_menu.nitems = 0;
	s_options_controls_menu.nitems = 0;
	s_options_ingame_menu.nitems = 0;
	s_options_interface_menu.nitems = 0;

	//SECTION SELECTORS
	{
		s_options_sound_section.generic.type	= MTYPE_ACTION;
		//s_options_sound_section.generic.flags  = QMF_LEFT_JUSTIFY;
		s_options_sound_section.generic.name	= "sound";
		s_options_sound_section.generic.x		= MENU_FONT_SIZE*0.5*strlen(s_options_sound_section.generic.name);
		s_options_sound_section.generic.y		= MENU_FONT_SIZE * 2;
		s_options_sound_section.generic.callback = setMenuSound;
		s_options_sound_section.generic.cursor_offset = -(MENU_FONT_SIZE*10);
		
		s_options_controls_section.generic.type	= MTYPE_ACTION;
		//s_options_controls_section.generic.flags  = QMF_LEFT_JUSTIFY;
		s_options_controls_section.generic.name	= "controls";
		s_options_controls_section.generic.x		= MENU_FONT_SIZE*0.5*strlen(s_options_controls_section.generic.name);
		s_options_controls_section.generic.y		= MENU_FONT_SIZE * 4;
		s_options_controls_section.generic.callback = setMenuControls;
		s_options_controls_section.generic.cursor_offset = -(MENU_FONT_SIZE*10);
		
		s_options_ingame_section.generic.type	= MTYPE_ACTION;
		//s_options_ingame_section.generic.flags  = QMF_LEFT_JUSTIFY;
		s_options_ingame_section.generic.name	= "ingame";
		s_options_ingame_section.generic.x		= MENU_FONT_SIZE*0.5*strlen(s_options_ingame_section.generic.name);
		s_options_ingame_section.generic.y		= MENU_FONT_SIZE * 6;
		s_options_ingame_section.generic.callback = setMenuIngame;
		s_options_ingame_section.generic.cursor_offset = -(MENU_FONT_SIZE*10);

		s_options_interface_section.generic.type	= MTYPE_ACTION;
		//s_options_interface_section.generic.flags  = QMF_LEFT_JUSTIFY;
		s_options_interface_section.generic.name	= "interface";
		s_options_interface_section.generic.x		= MENU_FONT_SIZE*0.5*strlen(s_options_interface_section.generic.name);
		s_options_interface_section.generic.y		= MENU_FONT_SIZE * 8;
		s_options_interface_section.generic.callback = setMenuInterface;
		s_options_interface_section.generic.cursor_offset = -(MENU_FONT_SIZE*10);
	}
	//SECTION HEADERS
	{
		s_options_sound_header.generic.type	= MTYPE_SEPARATOR;
		s_options_sound_header.generic.name	= "sound";
		s_options_sound_header.generic.x	= MENU_FONT_SIZE/2 * strlen(s_options_sound_header.generic.name);
		s_options_sound_header.generic.y	= 0;

		s_options_controls_header.generic.type	= MTYPE_SEPARATOR;
		s_options_controls_header.generic.name	= "controls";
		s_options_controls_header.generic.x	= MENU_FONT_SIZE/2 * strlen(s_options_controls_header.generic.name);
		s_options_controls_header.generic.y	= 0;

		s_options_ingame_header.generic.type	= MTYPE_SEPARATOR;
		s_options_ingame_header.generic.name	= "ingame";
		s_options_ingame_header.generic.x	= MENU_FONT_SIZE/2 * strlen(s_options_ingame_header.generic.name);
		s_options_ingame_header.generic.y	= 0;

		s_options_interface_header.generic.type	= MTYPE_SEPARATOR;
		s_options_interface_header.generic.name	= "interface";
		s_options_interface_header.generic.x	= MENU_FONT_SIZE/2 * strlen(s_options_interface_header.generic.name);
		s_options_interface_header.generic.y	= 0;
	}
	//BEGIN SOUND OPTIONS
	{
		y = 3*(MENU_FONT_SIZE+2);

		s_options_sfxvolume_slider.generic.type	= MTYPE_SLIDER;
		s_options_sfxvolume_slider.generic.x	= 0;
		s_options_sfxvolume_slider.generic.y	= y; //30
		s_options_sfxvolume_slider.generic.name	= "effects volume";
		s_options_sfxvolume_slider.generic.callback	= UpdateVolumeFunc;
		s_options_sfxvolume_slider.minvalue		= 0;
		s_options_sfxvolume_slider.maxvalue		= 10;
		s_options_sfxvolume_slider.curvalue		= Cvar_VariableValue( "s_volume" ) * 10;

		s_options_cdvolume_box.generic.type	= MTYPE_SPINCONTROL;
		s_options_cdvolume_box.generic.x		= 0;
		s_options_cdvolume_box.generic.y		= y+=(MENU_FONT_SIZE+2); //40
		s_options_cdvolume_box.generic.name	= "CD music";
		s_options_cdvolume_box.generic.callback	= UpdateCDVolumeFunc;
		s_options_cdvolume_box.itemnames		= cd_music_items;
		s_options_cdvolume_box.curvalue 		= !Cvar_VariableValue("cd_nocd");

		s_options_quality_list.generic.type	= MTYPE_SPINCONTROL;
		s_options_quality_list.generic.x		= 0;
		s_options_quality_list.generic.y		= y+=(MENU_FONT_SIZE+2); //50
		s_options_quality_list.generic.name		= "sound quality";
		s_options_quality_list.generic.callback = UpdateSoundQualityFunc;
		s_options_quality_list.itemnames		= quality_items;
		s_options_quality_list.curvalue			= !Cvar_VariableValue( "s_loadas8bit" );

		s_options_compatibility_list.generic.type	= MTYPE_SPINCONTROL;
		s_options_compatibility_list.generic.x		= 0;
		s_options_compatibility_list.generic.y		= y+=(MENU_FONT_SIZE+2); //60
		s_options_compatibility_list.generic.name	= "sound compatibility";
		s_options_compatibility_list.generic.callback = UpdateSoundQualityFunc;
		s_options_compatibility_list.itemnames		= compatibility_items;
		s_options_compatibility_list.curvalue		= Cvar_VariableValue( "s_primary" );
	}
	//BEGIN CONTROL OPTIONS
	{
		y = 3*(MENU_FONT_SIZE+2);

		s_options_sensitivity_slider.generic.type	= MTYPE_SLIDER;
		s_options_sensitivity_slider.generic.x		= 0;
		s_options_sensitivity_slider.generic.y		= y; //30
		s_options_sensitivity_slider.generic.name	= "mouse speed";
		s_options_sensitivity_slider.generic.callback = MouseSpeedFunc;
		s_options_sensitivity_slider.minvalue		= 2;
		s_options_sensitivity_slider.maxvalue		= 22;

		s_options_invertmouse_box.generic.type = MTYPE_SPINCONTROL;
		s_options_invertmouse_box.generic.x	= 0;
		s_options_invertmouse_box.generic.y	= y+=(MENU_FONT_SIZE+2); //40
		s_options_invertmouse_box.generic.name	= "invert mouse";
		s_options_invertmouse_box.generic.callback = InvertMouseFunc;
		s_options_invertmouse_box.itemnames = yesno_names;

		s_options_thirdperson_box.generic.type = MTYPE_SPINCONTROL;
		s_options_thirdperson_box.generic.x	= 0;
		s_options_thirdperson_box.generic.y	= y+=2*(MENU_FONT_SIZE+2); //60
		s_options_thirdperson_box.generic.name	= "third person";
		s_options_thirdperson_box.generic.callback = ThirdPersonFunc;
		s_options_thirdperson_box.itemnames = yesno_names;

		s_options_thirdperson_distance_slider.generic.type	= MTYPE_SLIDER;
		s_options_thirdperson_distance_slider.generic.x		= 0;
		s_options_thirdperson_distance_slider.generic.y		= y+=(MENU_FONT_SIZE+2); //70
		s_options_thirdperson_distance_slider.generic.name	= "camera distance";
		s_options_thirdperson_distance_slider.generic.callback = ThirdPersonDistFunc;
		s_options_thirdperson_distance_slider.minvalue		= 1;
		s_options_thirdperson_distance_slider.maxvalue		= 5;

		s_options_thirdperson_angle_slider.generic.type	= MTYPE_SLIDER;
		s_options_thirdperson_angle_slider.generic.x		= 0;
		s_options_thirdperson_angle_slider.generic.y		= y+=(MENU_FONT_SIZE+2); //80
		s_options_thirdperson_angle_slider.generic.name	= "camera angle";
		s_options_thirdperson_angle_slider.generic.callback = ThirdPersonAngleFunc;
		s_options_thirdperson_angle_slider.minvalue		= 0;
		s_options_thirdperson_angle_slider.maxvalue		= 4;

		s_options_alwaysrun_box.generic.type = MTYPE_SPINCONTROL;
		s_options_alwaysrun_box.generic.x	= 0;
		s_options_alwaysrun_box.generic.y	= y+=2*(MENU_FONT_SIZE+2); //100
		s_options_alwaysrun_box.generic.name	= "always run";
		s_options_alwaysrun_box.generic.callback = AlwaysRunFunc;
		s_options_alwaysrun_box.itemnames = yesno_names;

		s_options_lookspring_box.generic.type = MTYPE_SPINCONTROL;
		s_options_lookspring_box.generic.x	= 0;
		s_options_lookspring_box.generic.y	= y+=(MENU_FONT_SIZE+2); //110
		s_options_lookspring_box.generic.name	= "lookspring";
		s_options_lookspring_box.generic.callback = LookspringFunc;
		s_options_lookspring_box.itemnames = yesno_names;

		s_options_lookstrafe_box.generic.type = MTYPE_SPINCONTROL;
		s_options_lookstrafe_box.generic.x	= 0;
		s_options_lookstrafe_box.generic.y	= y+=(MENU_FONT_SIZE+2); //120
		s_options_lookstrafe_box.generic.name	= "lookstrafe";
		s_options_lookstrafe_box.generic.callback = LookstrafeFunc;
		s_options_lookstrafe_box.itemnames = yesno_names;

		s_options_freelook_box.generic.type = MTYPE_SPINCONTROL;
		s_options_freelook_box.generic.x	= 0;
		s_options_freelook_box.generic.y	= y+=(MENU_FONT_SIZE+2); //130
		s_options_freelook_box.generic.name	= "free look";
		s_options_freelook_box.generic.callback = FreeLookFunc;
		s_options_freelook_box.itemnames = yesno_names;

		s_options_joystick_box.generic.type = MTYPE_SPINCONTROL;
		s_options_joystick_box.generic.x	= 0;
		s_options_joystick_box.generic.y	= y+=(MENU_FONT_SIZE+2); //140
		s_options_joystick_box.generic.name	= "use joystick";
		s_options_joystick_box.generic.callback = JoystickFunc;
		s_options_joystick_box.itemnames = yesno_names;

		s_options_customize_options_action.generic.type	= MTYPE_ACTION;
		s_options_customize_options_action.generic.x		= MENU_FONT_SIZE;
		s_options_customize_options_action.generic.y		= y+=2*(MENU_FONT_SIZE+2); //160
		s_options_customize_options_action.generic.name	= "customize controls";
		s_options_customize_options_action.generic.callback = CustomizeControlsFunc;
	}
	//BEGIN INGAME OPTIONS
	{
		y = 3*(MENU_FONT_SIZE+2);

		crosshair_names = SetCrosshairNames ();
		s_options_crosshair_box.generic.type = MTYPE_SPINCONTROL;
		s_options_crosshair_box.generic.x	= 0;
		s_options_crosshair_box.generic.y	= y; //30
		s_options_crosshair_box.generic.name	= "crosshair";
		s_options_crosshair_box.generic.callback = CrosshairFunc;
		s_options_crosshair_box.itemnames = crosshair_names;

		//Knightmare- Psychospaz's changeable size crosshair
		s_options_crosshairscale_slider.generic.type	= MTYPE_SLIDER;
		s_options_crosshairscale_slider.generic.x		= 0;
		s_options_crosshairscale_slider.generic.y		= y+=5*(MENU_FONT_SIZE+2); //80
		s_options_crosshairscale_slider.generic.name	= "crosshair scale";
		s_options_crosshairscale_slider.generic.callback = CrosshairSizeFunc;
		s_options_crosshairscale_slider.minvalue		= 1;
		s_options_crosshairscale_slider.maxvalue		= 12;

		// Knightmare- hud scaling option
		s_options_hudscale_slider.generic.type			= MTYPE_SLIDER;
		s_options_hudscale_slider.generic.x				= 0;
		s_options_hudscale_slider.generic.y				= y+=(MENU_FONT_SIZE+2); //90
		s_options_hudscale_slider.generic.name			= "status bar scale";
		s_options_hudscale_slider.generic.callback		= HudScaleFunc;
		s_options_hudscale_slider.minvalue				= 1;
		s_options_hudscale_slider.maxvalue				= 7;

		// Knightmare- hud trans option
		s_options_hudalpha_slider.generic.type			= MTYPE_SLIDER;
		s_options_hudalpha_slider.generic.x				= 0;
		s_options_hudalpha_slider.generic.y				= y+=(MENU_FONT_SIZE+2); //100
		s_options_hudalpha_slider.generic.name			= "status bar transparency";
		s_options_hudalpha_slider.generic.callback		= HudAlphaFunc;
		s_options_hudalpha_slider.minvalue				= 1;
		s_options_hudalpha_slider.maxvalue				= 11;

		// Knightmare- particle effect complexity option
		s_options_particle_comp_slider.generic.type			= MTYPE_SLIDER;
		s_options_particle_comp_slider.generic.x			= 0;
		s_options_particle_comp_slider.generic.y			= y+=(MENU_FONT_SIZE+2); //110
		s_options_particle_comp_slider.generic.name			= "particle effect complexity";
		s_options_particle_comp_slider.generic.callback		= ParticleCompFunc;
		s_options_particle_comp_slider.generic.statusbar	= "lower = faster performance";
		s_options_particle_comp_slider.minvalue				= 1;
		s_options_particle_comp_slider.maxvalue				= 5;

		s_options_fps_box.generic.type = MTYPE_SPINCONTROL;
		s_options_fps_box.generic.x	= 0;
		s_options_fps_box.generic.y	= y +=(MENU_FONT_SIZE+2); //120
		s_options_fps_box.generic.name	= "FPS counter";
		s_options_fps_box.generic.callback = FPSFunc;
		s_options_fps_box.itemnames = yesno_names;

		// Knightmare- foostep override option
		s_options_footstep_box.generic.type = MTYPE_SPINCONTROL;
		s_options_footstep_box.generic.x	= 0;
		s_options_footstep_box.generic.y	= y += 2*(MENU_FONT_SIZE+2); //140
		s_options_footstep_box.generic.name	= "override footstep sounds";
		s_options_footstep_box.generic.callback = FootStepFunc;
		s_options_footstep_box.generic.statusbar = "sets footstep sounds with definitions in texsurfs.txt";
		s_options_footstep_box.itemnames = yesno_names;

		s_options_blood_box.generic.type = MTYPE_SPINCONTROL;
		s_options_blood_box.generic.x	= 0;
		s_options_blood_box.generic.y	= y += 2*(MENU_FONT_SIZE+2); //160
		s_options_blood_box.generic.name	= "blood";
		s_options_blood_box.generic.callback = BloodFunc;
		s_options_blood_box.itemnames = blood_names;

		//Knightmare- Psychospaz's changeable rail trail
		s_options_railtrail_box.generic.type = MTYPE_SPINCONTROL;
		s_options_railtrail_box.generic.x	= 0;
		s_options_railtrail_box.generic.y	= y += (MENU_FONT_SIZE+2); //170
		s_options_railtrail_box.generic.name	= "railtrail type";
		s_options_railtrail_box.generic.callback = RailTrailFunc;
		s_options_railtrail_box.itemnames = railtrail_names;

		s_options_railcolor_box[0].generic.type	= MTYPE_SLIDER;
		s_options_railcolor_box[0].generic.x		= 0;
		s_options_railcolor_box[0].generic.y		= y+=(MENU_FONT_SIZE+2); //180
		s_options_railcolor_box[0].generic.name	= "red - railtrail";
		s_options_railcolor_box[0].generic.callback = RailColorRedFunc;
		s_options_railcolor_box[0].minvalue		= 0;
		s_options_railcolor_box[0].maxvalue		= 16;

		s_options_railcolor_box[1].generic.type	= MTYPE_SLIDER;
		s_options_railcolor_box[1].generic.x		= 0;
		s_options_railcolor_box[1].generic.y		= y+=(MENU_FONT_SIZE+2); //190
		s_options_railcolor_box[1].generic.name	= "green - railtrail";
		s_options_railcolor_box[1].generic.callback = RailColorGreenFunc;
		s_options_railcolor_box[1].minvalue		= 0;
		s_options_railcolor_box[1].maxvalue		= 16;

		s_options_railcolor_box[2].generic.type	= MTYPE_SLIDER;
		s_options_railcolor_box[2].generic.x		= 0;
		s_options_railcolor_box[2].generic.y		= y+=(MENU_FONT_SIZE+2); //200
		s_options_railcolor_box[2].generic.name	= "blue - railtrail";
		s_options_railcolor_box[2].generic.callback = RailColorBlueFunc;
		s_options_railcolor_box[2].minvalue		= 0;
		s_options_railcolor_box[2].maxvalue		= 16;

		s_options_ingame_defaults_action.generic.type	= MTYPE_ACTION;
		s_options_ingame_defaults_action.generic.x		= MENU_FONT_SIZE;
		s_options_ingame_defaults_action.generic.y		= y+=2*(MENU_FONT_SIZE+2); //220;
		s_options_ingame_defaults_action.generic.name	= "reset defaults";
		s_options_ingame_defaults_action.generic.callback = ControlsResetDefaultsFunc;

		s_options_ingame_console_action.generic.type	= MTYPE_ACTION;
		s_options_ingame_console_action.generic.x		= MENU_FONT_SIZE;
		s_options_ingame_console_action.generic.y		= y+=(MENU_FONT_SIZE+2); //230;
		s_options_ingame_console_action.generic.name	= "go to console";
		s_options_ingame_console_action.generic.callback = ConsoleFunc;

		s_options_ingame_backmain_action.generic.type	= MTYPE_ACTION;
		s_options_ingame_backmain_action.generic.x		= MENU_FONT_SIZE;
		s_options_ingame_backmain_action.generic.y		= y+=(MENU_FONT_SIZE+2); //240;
		s_options_ingame_backmain_action.generic.name	= "back to options";
		s_options_ingame_backmain_action.generic.callback = setMenuOptionsMain;
	}
	//BEGIN INTERFACE
	{	
		y = 3*(MENU_FONT_SIZE+2);
		// Knightmare- Psychospaz's menu mouse support
		s_options_menumouse_slider.generic.type		= MTYPE_SLIDER;
		s_options_menumouse_slider.generic.x		= 0;
		s_options_menumouse_slider.generic.y		= y; //30
		s_options_menumouse_slider.generic.name		= "mouse speed";
		s_options_menumouse_slider.generic.callback = MouseMenuFunc;
		s_options_menumouse_slider.minvalue			= 1;
		s_options_menumouse_slider.maxvalue			= 8;

		// Knightmare- menu scaling option
		s_options_menuscale_box.generic.type		= MTYPE_SPINCONTROL;
		s_options_menuscale_box.generic.x			= 0;
		s_options_menuscale_box.generic.y			= y+=(MENU_FONT_SIZE+2); //40
		s_options_menuscale_box.generic.name		= "scaled menus";
		s_options_menuscale_box.generic.callback	= MenuScaleFunc;
		s_options_menuscale_box.itemnames			= yesno_names;

		font_names = SetFontNames ();
		s_options_font_box.generic.type = MTYPE_SPINCONTROL;
		s_options_font_box.generic.x	= 0;
		s_options_font_box.generic.y	= y+=2*(MENU_FONT_SIZE+2); //60
		s_options_font_box.generic.name	= "font";
		s_options_font_box.generic.callback = FontFunc;
		s_options_font_box.itemnames = font_names;

		s_options_fontsize_slider.generic.type		= MTYPE_SLIDER;
		s_options_fontsize_slider.generic.x			= 0;
		s_options_fontsize_slider.generic.y			= y+=(MENU_FONT_SIZE+2); //70
		s_options_fontsize_slider.generic.name		= "console font size";
		s_options_fontsize_slider.generic.callback	= FontSizeFunc;
		s_options_fontsize_slider.minvalue			= 2;
		s_options_fontsize_slider.maxvalue			= 6;

		// Knightmare- text color option
		s_options_alt_text_color_box.generic.type = MTYPE_SPINCONTROL;
		s_options_alt_text_color_box.generic.x	= 0;
		s_options_alt_text_color_box.generic.y	= y+=(MENU_FONT_SIZE+2); //80
		s_options_alt_text_color_box.generic.name	= "alt text color";
		s_options_alt_text_color_box.generic.callback = AltTextColorFunc;
		s_options_alt_text_color_box.itemnames = textcolor_names;

		// Knightmare- Psychospaz's transparent console cvar
		s_options_conalpha_slider.generic.type	= MTYPE_SLIDER;
		s_options_conalpha_slider.generic.x		= 0;
		s_options_conalpha_slider.generic.y		= y+=2*(MENU_FONT_SIZE+2); //100
		s_options_conalpha_slider.generic.name	= "console opacity";
		s_options_conalpha_slider.generic.callback = ConAlphaFunc;
		s_options_conalpha_slider.minvalue		= 0;
		s_options_conalpha_slider.maxvalue		= 20;

		// Knightmare- variable console height
		s_options_conheight_slider.generic.type	= MTYPE_SLIDER;
		s_options_conheight_slider.generic.x		= 0;
		s_options_conheight_slider.generic.y		= y+=(MENU_FONT_SIZE+2); //110
		s_options_conheight_slider.generic.name	= "console height";
		s_options_conheight_slider.generic.callback = ConHeightFunc;
		s_options_conheight_slider.minvalue		= 0;
		s_options_conheight_slider.maxvalue		= 10;

		s_options_noalttab_box.generic.type = MTYPE_SPINCONTROL;
		s_options_noalttab_box.generic.x	= 0;
		s_options_noalttab_box.generic.y	= y+=2*(MENU_FONT_SIZE+2); //130;
		s_options_noalttab_box.generic.name	= "disable alt-tab";
		s_options_noalttab_box.generic.callback = NoAltTabFunc;
		s_options_noalttab_box.itemnames = yesno_names;

	} // end Knightmare

	s_options_defaults_action.generic.type	= MTYPE_ACTION;
	s_options_defaults_action.generic.x		= MENU_FONT_SIZE;
	s_options_defaults_action.generic.y		= 17*(MENU_FONT_SIZE+2); //170;
	s_options_defaults_action.generic.name	= "reset defaults";
	s_options_defaults_action.generic.callback = ControlsResetDefaultsFunc;

	s_options_console_action.generic.type	= MTYPE_ACTION;
	s_options_console_action.generic.x		= MENU_FONT_SIZE;
	s_options_console_action.generic.y		= 18*(MENU_FONT_SIZE+2); //180;
	s_options_console_action.generic.name	= "go to console";
	s_options_console_action.generic.callback = ConsoleFunc;

	s_options_backmain_action.generic.type	= MTYPE_ACTION;
	s_options_backmain_action.generic.x		= MENU_FONT_SIZE;
	s_options_backmain_action.generic.y		= 19*(MENU_FONT_SIZE+2); //190;
	s_options_backmain_action.generic.name	= "back to options";
	s_options_backmain_action.generic.callback = setMenuOptionsMain;

	ControlsSetMenuItemValues();

	//Knightmare- Psychospaz's menu select box
		Menu_AddItem( &s_options_menu,			( void * ) &s_options_sound_section );
		Menu_AddItem( &s_options_menu,			( void * ) &s_options_controls_section );
		Menu_AddItem( &s_options_menu,			( void * ) &s_options_ingame_section );
		Menu_AddItem( &s_options_menu,			( void * ) &s_options_interface_section );

	//sound
		Menu_AddItem( &s_options_sound_menu, ( void * ) &s_options_sound_header );
		Menu_AddItem( &s_options_sound_menu, ( void * ) &s_options_sfxvolume_slider );
		Menu_AddItem( &s_options_sound_menu, ( void * ) &s_options_cdvolume_box );
		Menu_AddItem( &s_options_sound_menu, ( void * ) &s_options_quality_list );
		Menu_AddItem( &s_options_sound_menu, ( void * ) &s_options_compatibility_list );

	//controls
		Menu_AddItem( &s_options_controls_menu, ( void * ) &s_options_controls_header );
		Menu_AddItem( &s_options_controls_menu, ( void * ) &s_options_sensitivity_slider );
		Menu_AddItem( &s_options_controls_menu, ( void * ) &s_options_invertmouse_box );
		// Knightmare- added Psychospaz's chasecam
		Menu_AddItem( &s_options_controls_menu, ( void * ) &s_options_thirdperson_box );
		Menu_AddItem( &s_options_controls_menu, ( void * ) &s_options_thirdperson_distance_slider );
		Menu_AddItem( &s_options_controls_menu, ( void * ) &s_options_thirdperson_angle_slider );
		// end Knightmare
		Menu_AddItem( &s_options_controls_menu, ( void * ) &s_options_alwaysrun_box );
		Menu_AddItem( &s_options_controls_menu, ( void * ) &s_options_lookspring_box );
		Menu_AddItem( &s_options_controls_menu, ( void * ) &s_options_lookstrafe_box );
		Menu_AddItem( &s_options_controls_menu, ( void * ) &s_options_freelook_box );
		Menu_AddItem( &s_options_controls_menu, ( void * ) &s_options_joystick_box );
		Menu_AddItem( &s_options_controls_menu, ( void * ) &s_options_customize_options_action );

	//ingame
		Menu_AddItem( &s_options_ingame_menu, ( void * ) &s_options_ingame_header );
		Menu_AddItem( &s_options_ingame_menu, ( void * ) &s_options_crosshair_box );
		Menu_AddItem( &s_options_ingame_menu, ( void * ) &s_options_crosshairscale_slider );
		Menu_AddItem( &s_options_ingame_menu, ( void * ) &s_options_hudscale_slider );
		Menu_AddItem( &s_options_ingame_menu, ( void * ) &s_options_hudalpha_slider );
		Menu_AddItem( &s_options_ingame_menu, ( void * ) &s_options_particle_comp_slider );
		Menu_AddItem( &s_options_ingame_menu, ( void * ) &s_options_fps_box );

		// Knightmare- foostep override option
		Menu_AddItem( &s_options_ingame_menu, ( void * ) &s_options_footstep_box );

		Menu_AddItem( &s_options_ingame_menu, ( void * ) &s_options_blood_box );

		//Knightmare- Psychospaz's changeable rail trail
		Menu_AddItem( &s_options_ingame_menu, ( void * ) &s_options_railtrail_box );
		Menu_AddItem( &s_options_ingame_menu, ( void * ) &s_options_railcolor_box[0] );
		Menu_AddItem( &s_options_ingame_menu, ( void * ) &s_options_railcolor_box[1] );
		Menu_AddItem( &s_options_ingame_menu, ( void * ) &s_options_railcolor_box[2] );

	//interface
		Menu_AddItem( &s_options_interface_menu, ( void * ) &s_options_interface_header );
		Menu_AddItem( &s_options_interface_menu, ( void * ) &s_options_menumouse_slider );
		Menu_AddItem( &s_options_interface_menu, ( void * ) &s_options_menuscale_box );
		Menu_AddItem( &s_options_interface_menu, ( void * ) &s_options_font_box );
		Menu_AddItem( &s_options_interface_menu, ( void * ) &s_options_fontsize_slider );
		Menu_AddItem( &s_options_interface_menu, ( void * ) &s_options_alt_text_color_box );
		Menu_AddItem( &s_options_interface_menu, ( void * ) &s_options_conalpha_slider );
		Menu_AddItem( &s_options_interface_menu, ( void * ) &s_options_conheight_slider );
		Menu_AddItem( &s_options_interface_menu, ( void * ) &s_options_noalttab_box );

	Menu_AddItem( &s_options_sound_menu, ( void * ) &s_options_defaults_action );
	Menu_AddItem( &s_options_sound_menu, ( void * ) &s_options_console_action );
	Menu_AddItem( &s_options_sound_menu, ( void * ) &s_options_backmain_action );

	Menu_AddItem( &s_options_controls_menu, ( void * ) &s_options_defaults_action );
	Menu_AddItem( &s_options_controls_menu, ( void * ) &s_options_backmain_action );
	Menu_AddItem( &s_options_controls_menu, ( void * ) &s_options_console_action );

	Menu_AddItem( &s_options_ingame_menu, ( void * ) &s_options_ingame_defaults_action );
	Menu_AddItem( &s_options_ingame_menu, ( void * ) &s_options_ingame_console_action );
	Menu_AddItem( &s_options_ingame_menu, ( void * ) &s_options_ingame_backmain_action );

	Menu_AddItem( &s_options_interface_menu, ( void * ) &s_options_defaults_action );
	Menu_AddItem( &s_options_interface_menu, ( void * ) &s_options_console_action );
	Menu_AddItem( &s_options_interface_menu, ( void * ) &s_options_backmain_action );
}

void DrawMenuCrosshair (void)
{

	re.DrawFill (
		viddef.width/2 - scaledVideo(32)/2 + scaledVideo(MENU_FONT_SIZE/2-2),
		s_options_ingame_menu.y + scaledVideo(42),
		scaledVideo(36), scaledVideo(36), 3);

	re.DrawFill (
		viddef.width/2 - scaledVideo(32)/2 + scaledVideo(MENU_FONT_SIZE/2-1),
		s_options_ingame_menu.y + scaledVideo(43),
		scaledVideo(34), scaledVideo(34), 0);

	if (s_options_crosshair_box.curvalue < 1)
		return;

	re.DrawStretchPic (
		viddef.width/2 - scaledVideo(32)/2 + scaledVideo(MENU_FONT_SIZE/2),
		s_options_ingame_menu.y + scaledVideo(44),
		scaledVideo(32), scaledVideo(32), crosshair_names[s_options_crosshair_box.curvalue], 1.0);
}

void Options_MenuDraw (void)
{
	M_Banner( "m_banner_options" );

	if (options_menu->value==1)
	{
		Menu_AdjustCursor( &s_options_sound_menu, 1 );
		Menu_Draw( &s_options_sound_menu );
	}
	else if (options_menu->value==2)
	{
		Menu_AdjustCursor( &s_options_controls_menu, 1 );
		Menu_Draw( &s_options_controls_menu );
	}
	else if (options_menu->value==3)
	{
		Menu_AdjustCursor( &s_options_ingame_menu, 1 );
		Menu_Draw( &s_options_ingame_menu );
		DrawMenuCrosshair();
	}
	else if (options_menu->value==4)
	{
		Menu_AdjustCursor( &s_options_interface_menu, 1 );
		Menu_Draw( &s_options_interface_menu );
	}
	else
	{
		Menu_AdjustCursor( &s_options_menu, 1 );
		Menu_Draw( &s_options_menu );
	}

}

const char *Options_MenuKey( int key )
{
	if (options_menu->value==1)
		return Default_MenuKey( &s_options_sound_menu, key );
	else if (options_menu->value==2)
		return Default_MenuKey( &s_options_controls_menu, key );
	else if  (options_menu->value==3)
		return Default_MenuKey( &s_options_ingame_menu, key );
	else if  (options_menu->value==4)
		return Default_MenuKey( &s_options_interface_menu, key );
	else
		return Default_MenuKey( &s_options_menu, key );
}

void M_Menu_Options_f (void)
{
	Options_MenuInit();
	M_PushMenu ( Options_MenuDraw, Options_MenuKey );
}

/*
=======================================================================

VIDEO MENU

=======================================================================
*/

void M_Menu_Video_f (void)
{
	VID_MenuInit();
	M_PushMenu( VID_MenuDraw, VID_MenuKey );
}

/*
=============================================================================

END GAME MENU

=============================================================================
*/
static int credits_start_time;
// Knigthtmare added- allow credits to scroll past top of screen
static int credits_start_line;
static const char **credits;
static char *creditsIndex[256];
static char *creditsBuffer;
static const char *idcredits[] =
{
	"^b^s^aQUAKE II BY ID SOFTWARE",
	"",
	"^b^s^aPROGRAMMING",
	"John Carmack",
	"John Cash",
	"Brian Hook",
	"",
	"^b^s^aART",
	"Adrian Carmack",
	"Kevin Cloud",
	"Paul Steed",
	"",
	"^b^s^aLEVEL DESIGN",
	"Tim Willits",
	"American McGee",
	"Christian Antkow",
	"Paul Jaquays",
	"Brandon James",
	"",
	"^b^s^aBIZ",
	"Todd Hollenshead",
	"Barrett (Bear) Alexander",
	"Donna Jackson",
	"",
	"",
	"^b^s^aSPECIAL THANKS",
	"Ben Donges for beta testing",
	"",
	"",
	"",
	"",
	"",
	"",
	"^b^s^aADDITIONAL SUPPORT",
	"",
	"^b^s^aLINUX PORT AND CTF",
	"Dave \"Zoid\" Kirsch",
	"",
	"^b^s^aCINEMATIC SEQUENCES",
	"Ending Cinematic by Blur Studio - ",
	"Venice, CA",
	"",
	"Environment models for Introduction",
	"Cinematic by Karl Dolgener",
	"",
	"Assistance with environment design",
	"by Cliff Iwai",
	"",
	"^b^s^aSOUND EFFECTS AND MUSIC",
	"Sound Design by Soundelux Media Labs.",
	"Music Composed and Produced by",
	"Soundelux Media Labs.  Special thanks",
	"to Bill Brown, Tom Ozanich, Brian",
	"Celano, Jeff Eisner, and The Soundelux",
	"Players.",
	"",
	"\"Level Music\" by Sonic Mayhem",
	"www.sonicmayhem.com",
	"",
	"\"Quake II Theme Song\"",
	"(C) 1997 Rob Zombie. All Rights",
	"Reserved.",
	"",
	"Track 10 (\"Climb\") by Jer Sypult",
	"",
	"Voice of computers by",
	"Carly Staehlin-Taylor",
	"",
	"^b^s^aTHANKS TO ACTIVISION",
	"^b^s^aIN PARTICULAR:",
	"",
	"John Tam",
	"Steve Rosenthal",
	"Marty Stratton",
	"Henk Hartong",
	"",
	"Quake II(tm) (C)1997 Id Software, Inc.",
	"All Rights Reserved.  Distributed by",
	"Activision, Inc. under license.",
	"Quake II(tm), the Id Software name,",
	"the \"Q II\"(tm) logo and id(tm)",
	"logo are trademarks of Id Software,",
	"Inc. Activision(R) is a registered",
	"trademark of Activision, Inc. All",
	"other trademarks and trade names are",
	"properties of their respective owners.",
	0
};

static const char *xatcredits[] =
{
	"^b^s^aQUAKE II MISSION PACK: THE RECKONING",
	"^b^s^aBY",
	"^b^s^aXATRIX ENTERTAINMENT, INC.",
	"",
	"^b^s^aDESIGN AND DIRECTION",
	"Drew Markham",
	"",
	"^b^s^aPRODUCED BY",
	"Greg Goodrich",
	"",
	"^b^s^aPROGRAMMING",
	"Rafael Paiz",
	"",
	"^b^s^aLEVEL DESIGN / ADDITIONAL GAME DESIGN",
	"Alex Mayberry",
	"",
	"^b^s^aLEVEL DESIGN",
	"Mal Blackwell",
	"Dan Koppel",
	"",
	"^b^s^aART DIRECTION",
	"Michael \"Maxx\" Kaufman",
	"",
	"^b^s^aCOMPUTER GRAPHICS SUPERVISOR AND",
	"^b^s^aCHARACTER ANIMATION DIRECTION",
	"Barry Dempsey",
	"",
	"^b^s^aSENIOR ANIMATOR AND MODELER",
	"Jason Hoover",
	"",
	"^b^s^aCHARACTER ANIMATION AND",
	"^b^s^aMOTION CAPTURE SPECIALIST",
	"Amit Doron",
	"",
	"^b^s^aART",
	"Claire Praderie-Markham",
	"Viktor Antonov",
	"Corky Lehmkuhl",
	"",
	"^b^s^aINTRODUCTION ANIMATION",
	"Dominique Drozdz",
	"",
	"^b^s^aADDITIONAL LEVEL DESIGN",
	"Aaron Barber",
	"Rhett Baldwin",
	"",
	"^b^s^a3D CHARACTER ANIMATION TOOLS",
	"Gerry Tyra, SA Technology",
	"",
	"^b^s^aADDITIONAL EDITOR TOOL PROGRAMMING",
	"Robert Duffy",
	"",
	"^b^s^aADDITIONAL PROGRAMMING",
	"Ryan Feltrin",
	"",
	"^b^s^aPRODUCTION COORDINATOR",
	"Victoria Sylvester",
	"",
	"^b^s^aSOUND DESIGN",
	"Gary Bradfield",
	"",
	"^b^s^aMUSIC BY",
	"Sonic Mayhem",
	"",
	"",
	"",
	"^b^s^aSPECIAL THANKS",
	"^b^s^aTO",
	"^b^s^aOUR FRIENDS AT ID SOFTWARE",
	"",
	"John Carmack",
	"John Cash",
	"Brian Hook",
	"Adrian Carmack",
	"Kevin Cloud",
	"Paul Steed",
	"Tim Willits",
	"Christian Antkow",
	"Paul Jaquays",
	"Brandon James",
	"Todd Hollenshead",
	"Barrett (Bear) Alexander",
	"Dave \"Zoid\" Kirsch",
	"Donna Jackson",
	"",
	"",
	"",
	"^b^s^aTHANKS TO ACTIVISION",
	"^b^s^aIN PARTICULAR:",
	"",
	"Marty Stratton",
	"Henk \"The Original Ripper\" Hartong",
	"Kevin Kraff",
	"Jamey Gottlieb",
	"Chris Hepburn",
	"",
	"^b^s^aAND THE GAME TESTERS",
	"",
	"Tim Vanlaw",
	"Doug Jacobs",
	"Steven Rosenthal",
	"David Baker",
	"Chris Campbell",
	"Aaron Casillas",
	"Steve Elwell",
	"Derek Johnstone",
	"Igor Krinitskiy",
	"Samantha Lee",
	"Michael Spann",
	"Chris Toft",
	"Juan Valdes",
	"",
	"^b^s^aTHANKS TO INTERGRAPH COMPUTER SYTEMS",
	"^b^s^aIN PARTICULAR:",
	"",
	"Michael T. Nicolaou",
	"",
	"",
	"Quake II Mission Pack: The Reckoning",
	"(tm) (C)1998 Id Software, Inc. All",
	"Rights Reserved. Developed by Xatrix",
	"Entertainment, Inc. for Id Software,",
	"Inc. Distributed by Activision Inc.",
	"under license. Quake(R) is a",
	"registered trademark of Id Software,",
	"Inc. Quake II Mission Pack: The",
	"Reckoning(tm), Quake II(tm), the Id",
	"Software name, the \"Q II\"(tm) logo",
	"and id(tm) logo are trademarks of Id",
	"Software, Inc. Activision(R) is a",
	"registered trademark of Activision,",
	"Inc. Xatrix(R) is a registered",
	"trademark of Xatrix Entertainment,",
	"Inc. All other trademarks and trade",
	"names are properties of their",
	"respective owners.",
	0
};

static const char *roguecredits[] =
{
	"^b^s^aQUAKE II MISSION PACK 2: GROUND ZERO",
	"^b^s^aBY",
	"^b^s^aROGUE ENTERTAINMENT, INC.",
	"",
	"^b^s^aPRODUCED BY",
	"Jim Molinets",
	"",
	"^b^s^aPROGRAMMING",
	"Peter Mack",
	"Patrick Magruder",
	"",
	"^b^s^aLEVEL DESIGN",
	"Jim Molinets",
	"Cameron Lamprecht",
	"Berenger Fish",
	"Robert Selitto",
	"Steve Tietze",
	"Steve Thoms",
	"",
	"^b^s^aART DIRECTION",
	"Rich Fleider",
	"",
	"^b^s^aART",
	"Rich Fleider",
	"Steve Maines",
	"Won Choi",
	"",
	"^b^s^aANIMATION SEQUENCES",
	"Creat Studios",
	"Steve Maines",
	"",
	"^b^s^aADDITIONAL LEVEL DESIGN",
	"Rich Fleider",
	"Steve Maines",
	"Peter Mack",
	"",
	"^b^s^aSOUND",
	"James Grunke",
	"",
	"^b^s^aGROUND ZERO THEME",
	"^b^s^aAND",
	"^b^s^aMUSIC BY",
	"Sonic Mayhem",
	"",
	"^b^s^aVWEP MODELS",
	"Brent \"Hentai\" Dill",
	"",
	"",
	"",
	"^b^s^aSPECIAL THANKS",
	"^b^s^aTO",
	"^b^s^aOUR FRIENDS AT ID SOFTWARE",
	"",
	"John Carmack",
	"John Cash",
	"Brian Hook",
	"Adrian Carmack",
	"Kevin Cloud",
	"Paul Steed",
	"Tim Willits",
	"Christian Antkow",
	"Paul Jaquays",
	"Brandon James",
	"Todd Hollenshead",
	"Barrett (Bear) Alexander",
	"Katherine Anna Kang",
	"Donna Jackson",
	"Dave \"Zoid\" Kirsch",
	"",
	"",
	"",
	"^b^s^aTHANKS TO ACTIVISION",
	"^b^s^aIN PARTICULAR:",
	"",
	"Marty Stratton",
	"Henk Hartong",
	"Mitch Lasky",
	"Steve Rosenthal",
	"Steve Elwell",
	"",
	"^b^s^aAND THE GAME TESTERS",
	"",
	"The Ranger Clan",
	"Dave \"Zoid\" Kirsch",
	"Nihilistic Software",
	"Robert Duffy",
	"",
	"And Countless Others",
	"",
	"",
	"",
	"Quake II Mission Pack 2: Ground Zero",
	"(tm) (C)1998 Id Software, Inc. All",
	"Rights Reserved. Developed by Rogue",
	"Entertainment, Inc. for Id Software,",
	"Inc. Distributed by Activision Inc.",
	"under license. Quake(R) is a",
	"registered trademark of Id Software,",
	"Inc. Quake II Mission Pack 2: Ground",
	"Zero(tm), Quake II(tm), the Id",
	"Software name, the \"Q II\"(tm) logo",
	"and id(tm) logo are trademarks of Id",
	"Software, Inc. Activision(R) is a",
	"registered trademark of Activision,",
	"Inc. Rogue(R) is a registered",
	"trademark of Rogue Entertainment,",
	"Inc. All other trademarks and trade",
	"names are properties of their",
	"respective owners.",
	0
};

int stringLengthExtra ( const char *string);
void M_Credits_MenuDraw( void )
{
	float alpha, time = (cls.realtime - credits_start_time) * 0.05;
	int i, y, x, len;


	if ((viddef.height - ((cls.realtime - credits_start_time)/40.0F) + credits_start_line * scaledVideo(MENU_FONT_SIZE+2)) < 0)
	{
		credits_start_line++;
		if (!credits[credits_start_line])
		{
			credits_start_line = 0;
			credits_start_time = cls.realtime;
		}
	}

	/*
	** draw the credits
	*/
	for (i=credits_start_line, y=viddef.height - ((cls.realtime - credits_start_time)/40.0F) + credits_start_line * scaledVideo(MENU_FONT_SIZE+2);
		credits[i] && y < viddef.height; y += scaledVideo(MENU_FONT_SIZE+2), i++) // was 10
	{
		//int j;
		int	stringoffset = 0;
		int bold = false;

		if ( y <= -MENU_FONT_SIZE )
			continue;
		if ( y > viddef.height )
			continue;

		if ( credits[i][0] == '+' )
		{
			bold = true;
			stringoffset = 1;
		}
		else
		{
			bold = false;
			stringoffset = 0;
		}

		if (y > 7*viddef.height/8)
		{
			float y_test, h_test;
			y_test = y - (7.0/8.0)*viddef.height;
			h_test = viddef.height/8;

			alpha = 1 - (y_test/h_test);

			if (alpha>1) alpha=1; if (alpha<0) alpha=0;
		}
		else if (y < viddef.height/8)
		{
			float y_test, h_test;
			y_test = y;
			h_test = viddef.height/8;

			alpha = y_test/h_test;

			if (alpha>1) alpha=1; if (alpha<0) alpha=0;
		}
		else
			alpha = 1;

		len = strlen(credits[i]) - stringLengthExtra(credits[i]);

		x = ( viddef.width - len * scaledVideo(MENU_FONT_SIZE) - stringoffset * scaledVideo(MENU_FONT_SIZE) ) / 2 + ( stringoffset ) * scaledVideo(MENU_FONT_SIZE);
		Menu_DrawString(x, y, credits[i], alpha*255);

		//x = ( viddef.width - len * MENU_FONT_SIZE - stringoffset * MENU_FONT_SIZE ) / 2 + ( stringoffset ) * MENU_FONT_SIZE;
		//DrawString (x, y, credits[i], (int)(alpha*255));
		/*for ( j = 0; credits[i][j+stringoffset]; j++ )
		{
			x = (viddef.width - strlen(credits[i]) * scaledVideo(MENU_FONT_SIZE) - stringoffset * scaledVideo(MENU_FONT_SIZE) ) / 2
				+ (j + stringoffset) * scaledVideo(MENU_FONT_SIZE);
			if ( bold )
				re.DrawChar( x, y, credits[i][j+stringoffset] + 128, alpha*255);
			else
				re.DrawChar( x, y, credits[i][j+stringoffset], alpha*255);
		}*/
	}
	/*if (y < 0)
		credits_start_time = cls.realtime;*/
}

const char *M_Credits_Key( int key )
{
	char *sound = NULL;

	switch (key)
	{
	case K_ESCAPE:
	//case K_MOUSE2
		if (creditsBuffer)
			FS_FreeFile (creditsBuffer);
		M_PopMenu ();
		sound = menu_out_sound;
		break;
	}

	return sound;
}


void M_Menu_Credits_f( void )
{
	int		n;
	int		count;
	char	*p;

	creditsBuffer = NULL;
	count = FS_LoadFile ("credits", &creditsBuffer);
	if (count != -1)
	{
		p = creditsBuffer;
		for (n = 0; n < 255; n++)
		{
			creditsIndex[n] = p;
			while (*p != '\r' && *p != '\n')
			{
				p++;
				if (--count == 0)
					break;
			}
			if (*p == '\r')
			{
				*p++ = 0;
				if (--count == 0)
					break;
			}
			*p++ = 0;
			if (--count == 0)
				break;
		}
		creditsIndex[++n] = 0;
		credits = creditsIndex;
	}
	else
	{	
		if (modType("xatrix"))			// xatrix
			credits = xatcredits;
		else if (modType("rogue"))		// ROGUE
			credits = roguecredits;
		else
			credits = idcredits;	
	}

	credits_start_time = cls.realtime;
	credits_start_line = 0; // Knightmare- allow credits to scroll past top of screen
	M_PushMenu( M_Credits_MenuDraw, M_Credits_Key);
}

/*
=============================================================================

GAME MENU

=============================================================================
*/

static int		m_game_cursor;

static menuframework_s	s_game_menu;
static menuaction_s		s_easy_game_action;
static menuaction_s		s_medium_game_action;
static menuaction_s		s_hard_game_action;
static menuaction_s		s_nitemare_game_action;
static menuaction_s		s_load_game_action;
static menuaction_s		s_save_game_action;
static menuaction_s		s_credits_action;
static menuseparator_s	s_blankline;

static void StartGame( void )
{
	// disable updates and start the cinematic going
	cl.servercount = -1;
	M_ForceMenuOff ();
	Cvar_SetValue( "deathmatch", 0 );
	Cvar_SetValue( "coop", 0 );
	Cvar_SetValue( "gamerules", 0 );		//PGM

	Cbuf_AddText ("loading ; killserver ; wait ; newgame\n");
	cls.key_dest = key_game;
}

static void EasyGameFunc( void *data )
{
	Cvar_ForceSet( "skill", "0" );
	StartGame();
}

static void MediumGameFunc( void *data )
{
	Cvar_ForceSet( "skill", "1" );
	StartGame();
}

static void HardGameFunc( void *data )
{
	Cvar_ForceSet( "skill", "2" );
	StartGame();
}

static void NitemareGameFunc( void *data )
{
	Cvar_ForceSet( "skill", "3" );
	StartGame();
}

static void LoadGameFunc( void *unused )
{
	M_Menu_LoadGame_f ();
}

static void SaveGameFunc( void *unused )
{
	M_Menu_SaveGame_f();
}

static void CreditsFunc( void *unused )
{
	M_Menu_Credits_f();
}

void Game_MenuInit( void )
{
	static const char *difficulty_names[] =
	{
		"easy",
		"medium",
		"hard",
		0
	};
	int y = 0;

	s_game_menu.x = viddef.width * 0.50;
	s_game_menu.nitems = 0;

	s_easy_game_action.generic.type	= MTYPE_ACTION;
	s_easy_game_action.generic.flags  = QMF_LEFT_JUSTIFY;
	s_easy_game_action.generic.x		= 0;
	s_easy_game_action.generic.y		= y; // 0
	s_easy_game_action.generic.name	= "easy";
	s_easy_game_action.generic.callback = EasyGameFunc;

	s_medium_game_action.generic.type	= MTYPE_ACTION;
	s_medium_game_action.generic.flags  = QMF_LEFT_JUSTIFY;
	s_medium_game_action.generic.x		= 0;
	s_medium_game_action.generic.y		= y += (MENU_FONT_SIZE+2); //10
	s_medium_game_action.generic.name	= "medium";
	s_medium_game_action.generic.callback = MediumGameFunc;

	s_hard_game_action.generic.type	= MTYPE_ACTION;
	s_hard_game_action.generic.flags  = QMF_LEFT_JUSTIFY;
	s_hard_game_action.generic.x		= 0;
	s_hard_game_action.generic.y		= y += (MENU_FONT_SIZE+2); //20
	s_hard_game_action.generic.name	= "hard";
	s_hard_game_action.generic.callback = HardGameFunc;

	s_nitemare_game_action.generic.type	= MTYPE_ACTION;
	s_nitemare_game_action.generic.flags  = QMF_LEFT_JUSTIFY;
	s_nitemare_game_action.generic.x		= 0;
	s_nitemare_game_action.generic.y		= y += (MENU_FONT_SIZE+2); //30
	s_nitemare_game_action.generic.name	= "nightmare";
	s_nitemare_game_action.generic.callback = NitemareGameFunc;


	s_blankline.generic.type = MTYPE_SEPARATOR;

	s_load_game_action.generic.type	= MTYPE_ACTION;
	s_load_game_action.generic.flags  = QMF_LEFT_JUSTIFY;
	s_load_game_action.generic.x		= 0;
	s_load_game_action.generic.y		= y += 2*(MENU_FONT_SIZE+2); //50
	s_load_game_action.generic.name	= "load game";
	s_load_game_action.generic.callback = LoadGameFunc;

	s_save_game_action.generic.type	= MTYPE_ACTION;
	s_save_game_action.generic.flags  = QMF_LEFT_JUSTIFY;
	s_save_game_action.generic.x		= 0;
	s_save_game_action.generic.y		= y += (MENU_FONT_SIZE+2); //60
	s_save_game_action.generic.name	= "save game";
	s_save_game_action.generic.callback = SaveGameFunc;

	s_credits_action.generic.type	= MTYPE_ACTION;
	s_credits_action.generic.flags  = QMF_LEFT_JUSTIFY;
	s_credits_action.generic.x		= 0;
	s_credits_action.generic.y		= y += (MENU_FONT_SIZE+2); //70
	s_credits_action.generic.name	= "credits";
	s_credits_action.generic.callback = CreditsFunc;

	Menu_AddItem( &s_game_menu, ( void * ) &s_easy_game_action );
	Menu_AddItem( &s_game_menu, ( void * ) &s_medium_game_action );
	Menu_AddItem( &s_game_menu, ( void * ) &s_hard_game_action );
	Menu_AddItem( &s_game_menu, ( void * ) &s_nitemare_game_action );

	Menu_AddItem( &s_game_menu, ( void * ) &s_blankline );

	Menu_AddItem( &s_game_menu, ( void * ) &s_load_game_action );
	Menu_AddItem( &s_game_menu, ( void * ) &s_save_game_action );

	Menu_AddItem( &s_game_menu, ( void * ) &s_blankline );
	Menu_AddItem( &s_game_menu, ( void * ) &s_credits_action );

	Menu_Center( &s_game_menu );
}

void Game_MenuDraw( void )
{
	M_Banner( "m_banner_game" );
	Menu_AdjustCursor( &s_game_menu, 1 );
	Menu_Draw( &s_game_menu );
}

const char *Game_MenuKey( int key )
{
	return Default_MenuKey( &s_game_menu, key );
}

void M_Menu_Game_f (void)
{
	Game_MenuInit();
	M_PushMenu( Game_MenuDraw, Game_MenuKey );
	m_game_cursor = 1;
}

/*
=============================================================================

LOADGAME MENU

=============================================================================
*/

#define	MAX_SAVEGAMES	15

static menuframework_s	s_savegame_menu;

static menuframework_s	s_loadgame_menu;
static menuaction_s		s_loadgame_actions[MAX_SAVEGAMES];

char		m_savestrings[MAX_SAVEGAMES][32];
qboolean	m_savevalid[MAX_SAVEGAMES];
qboolean	m_saveshotvalid[MAX_SAVEGAMES+1];

void Create_Savestrings (void)
{
	int		i;
	FILE	*f;
	char	name[MAX_OSPATH];

	for (i=0 ; i<MAX_SAVEGAMES ; i++)
	{
		Com_sprintf (name, sizeof(name), "%s/save/save%i/server.ssv", FS_Gamedir(), i);
		f = fopen (name, "rb");
		if (!f)
		{
			strcpy (m_savestrings[i], "<EMPTY>");
			m_savevalid[i] = false;
		}
		else
		{
			FS_Read (m_savestrings[i], sizeof(m_savestrings[i]), f);
			fclose (f);
			m_savevalid[i] = true;
		}
	}
}

char *load_saveshot;
char loadshotname[MAX_QPATH];
void LoadGameCallback( void *self )
{
	menuaction_s *a = ( menuaction_s * ) self;

	// set saveshot name here
	if ( m_saveshotvalid[ a->generic.localdata[0] ] ) {
		Com_sprintf(loadshotname, sizeof(loadshotname), "/save/save%i/shot.jpg", a->generic.localdata[0]);
		load_saveshot = loadshotname; }
	else
		load_saveshot = NULL;

	if ( m_savevalid[ a->generic.localdata[0] ] )
		Cbuf_AddText (va("load save%i\n",  a->generic.localdata[0] ) );
	M_ForceMenuOff ();
}

void InitLoadGameSaveshots (void)
{
	int i;
	char shotname [MAX_QPATH];

	for ( i = 0; i < MAX_SAVEGAMES; i++ )
	{	// free previously loaded shots
		Com_sprintf(shotname, sizeof(shotname), "save/save%i/shot.jpg", i);
		re.UnregisterPic(shotname);
		Com_sprintf(shotname, sizeof(shotname), "/save/save%i/shot.jpg", i);
		if (re.RegisterPic(shotname))
			m_saveshotvalid[i] = true;
		else
			m_saveshotvalid[i] = false;
	}
	// register null saveshot
	if (re.RegisterPic("/gfx/noscreen.pcx"))
		m_saveshotvalid[MAX_SAVEGAMES] = true;
	else
		m_saveshotvalid[MAX_SAVEGAMES] = false;
}

void LoadGame_MenuInit ( void )
{
	int i;

	s_loadgame_menu.x = viddef.width / 2 - scaledVideo(240);
	//s_loadgame_menu.x = viddef.width / 2 - scaledVideo(15*MENU_FONT_SIZE);	// was 120
	s_loadgame_menu.y = viddef.height / 2 - scaledVideo(7.25*MENU_FONT_SIZE); // was 58
	s_loadgame_menu.nitems = 0;

	Create_Savestrings();

	for ( i = 0; i < MAX_SAVEGAMES; i++ )
	{
		s_loadgame_actions[i].generic.name			= m_savestrings[i];
		s_loadgame_actions[i].generic.flags			= QMF_LEFT_JUSTIFY;
		s_loadgame_actions[i].generic.localdata[0]	= i;
		s_loadgame_actions[i].generic.callback		= LoadGameCallback;

		s_loadgame_actions[i].generic.x = 0;
		s_loadgame_actions[i].generic.y = (i) * (MENU_FONT_SIZE+2); //10
		if (i>0)	// separate from autosave
			s_loadgame_actions[i].generic.y += 10;

		s_loadgame_actions[i].generic.type = MTYPE_ACTION;

		Menu_AddItem( &s_loadgame_menu, &s_loadgame_actions[i] );
	}
	InitLoadGameSaveshots(); // register saveshots
}

void DrawLoadGameSaveshot (void)
{
	char shotname [MAX_QPATH];
	int i = s_loadgame_actions[s_loadgame_menu.cursor].generic.localdata[0];

	re.DrawFill (
		viddef.width/2 + scaledVideo(44),
		viddef.height/2 - scaledVideo(60),
		scaledVideo(244), scaledVideo(184), 3);

	if ( m_savevalid[i] && m_saveshotvalid[i] )
	{
		Com_sprintf(shotname, sizeof(shotname), "/save/save%i/shot.jpg", i);

		re.DrawStretchPic (
			viddef.width/2 + scaledVideo(46),
			viddef.height/2 - scaledVideo(58),
			scaledVideo(240), scaledVideo(180), shotname, 1.0);
	}
	else if (m_saveshotvalid[MAX_SAVEGAMES])
		re.DrawStretchPic (
			viddef.width/2 + scaledVideo(46),
			viddef.height/2 - scaledVideo(58),
			scaledVideo(240), scaledVideo(180), "/gfx/noscreen.pcx", 1.0);
	else
		re.DrawFill (
			viddef.width/2 + scaledVideo(46),
			viddef.height/2 - scaledVideo(58),
			scaledVideo(240), scaledVideo(180), 0);
}

void LoadGame_MenuDraw( void )
{
	M_Banner( "m_banner_load_game" );
//	Menu_AdjustCursor( &s_loadgame_menu, 1 );
	Menu_Draw( &s_loadgame_menu );
	DrawLoadGameSaveshot();
}

const char *LoadGame_MenuKey( int key )
{
	if ( key == K_ESCAPE || key == K_ENTER )
	{
		s_savegame_menu.cursor = s_loadgame_menu.cursor - 1;
		if ( s_savegame_menu.cursor < 0 )
			s_savegame_menu.cursor = 0;
	}
	return Default_MenuKey( &s_loadgame_menu, key );
}

void M_Menu_LoadGame_f (void)
{
	LoadGame_MenuInit();
	M_PushMenu( LoadGame_MenuDraw, LoadGame_MenuKey );
}


/*
=============================================================================

SAVEGAME MENU

=============================================================================
*/
static menuframework_s	s_savegame_menu;
static menuaction_s		s_savegame_actions[MAX_SAVEGAMES];

void SaveGameCallback( void *self )
{
	menuaction_s *a = ( menuaction_s * ) self;

	Cbuf_AddText (va("save save%i\n", a->generic.localdata[0] ));
	M_ForceMenuOff ();
}

void SaveGame_MenuDraw( void )
{
	M_Banner( "m_banner_save_game" );
	Menu_AdjustCursor( &s_savegame_menu, 1 );
	Menu_Draw( &s_savegame_menu );
}

void SaveGame_MenuInit( void )
{
	int i;

	s_savegame_menu.x = viddef.width / 2 - scaledVideo(15*MENU_FONT_SIZE);	// was 120
	s_savegame_menu.y = viddef.height / 2 - scaledVideo(7.25*MENU_FONT_SIZE);//was 58
	s_savegame_menu.nitems = 0;

	Create_Savestrings();

	// don't include the autosave slot
	for ( i = 0; i < MAX_SAVEGAMES-1; i++ )
	{
		s_savegame_actions[i].generic.name = m_savestrings[i+1];
		s_savegame_actions[i].generic.localdata[0] = i+1;
		s_savegame_actions[i].generic.flags = QMF_LEFT_JUSTIFY;
		s_savegame_actions[i].generic.callback = SaveGameCallback;

		s_savegame_actions[i].generic.x = 0;
		s_savegame_actions[i].generic.y = ( i ) * (MENU_FONT_SIZE+2); //10;

		s_savegame_actions[i].generic.type = MTYPE_ACTION;

		Menu_AddItem( &s_savegame_menu, &s_savegame_actions[i] );
	}
}

const char *SaveGame_MenuKey( int key )
{
	if ( key == K_ENTER || key == K_ESCAPE )
	{
		s_loadgame_menu.cursor = s_savegame_menu.cursor - 1;
		if ( s_loadgame_menu.cursor < 0 )
			s_loadgame_menu.cursor = 0;
	}
	return Default_MenuKey( &s_savegame_menu, key );
}

void M_Menu_SaveGame_f (void)
{
	if (!Com_ServerState())
		return;		// not playing a game

	SaveGame_MenuInit();
	M_PushMenu( SaveGame_MenuDraw, SaveGame_MenuKey );
	Create_Savestrings ();
}


/*
=============================================================================

JOIN SERVER MENU

=============================================================================
*/
#define MAX_LOCAL_SERVERS 8

static menuframework_s	s_joinserver_menu;
static menuseparator_s	s_joinserver_server_title;

//Knightmare- client compatibility option
static menulist_s		s_joinserver_compatibility_box;
static menuseparator_s	s_joinserver_compat_title;

static menuaction_s		s_joinserver_search_action;
static menuaction_s		s_joinserver_address_book_action;
static menuaction_s		s_joinserver_server_actions[MAX_LOCAL_SERVERS];

int		m_num_servers;
#define	NO_SERVER_STRING	"<no server>"

// user readable information
static char local_server_names[MAX_LOCAL_SERVERS][80];

// network address
static netadr_t local_server_netadr[MAX_LOCAL_SERVERS];


void SearchLocalGames( void );

//Knightmare- client compatibility option
static void ClientCompatibilityFunc( void *unused )
{
	Cvar_SetValue( "cl_servertrick", s_joinserver_compatibility_box.curvalue );
}


#if 1
//<serverping> Added code for compute ping time of server broadcasted
// The server is displayed like : 
// "protocol ping hostname mapname nb players/max players"
// "udp 100ms Pat  q2dm1 2/8"

int      global_udp_server_time;
int      global_ipx_server_time;
int      global_adr_server_time[16];
netadr_t global_adr_server_netadr[16];

void M_AddToServerList (netadr_t adr, char *info)
{
	int		i;
    int     iPing;
    char    *pszProtocol;

	if (m_num_servers == MAX_LOCAL_SERVERS)
		return;
	while ( *info == ' ' )
		info++;

	// Ignore if duplicated
	for (i=0 ; i<m_num_servers ; i++)
        if ( strncmp(info, &local_server_names[i][11], sizeof(local_server_names[0])-10)==0 )
			return;

    iPing = 0 ;
	for (i=0 ; i<MAX_LOCAL_SERVERS ; i++)
    {
        if (    memcmp(&adr.ip, &global_adr_server_netadr[i].ip, sizeof(adr.ip))==0 
             && adr.port == global_adr_server_netadr[i].port )
        {
            // bookmark server
            iPing = Sys_Milliseconds() - global_adr_server_time[i] ;
            pszProtocol = "bkm" ;
            break;
        }
    }
    if ( i == MAX_LOCAL_SERVERS )
    {
        if ( adr.ip[0] > 0 )    // udp server
        {
            iPing = Sys_Milliseconds() - global_udp_server_time ;
            pszProtocol = "UDP" ;
        }
        else                    // ipx server
        {
            iPing = Sys_Milliseconds() - global_ipx_server_time ;
            pszProtocol = "IPX" ;
        }
    }

    Com_sprintf( local_server_names[m_num_servers], 
                 sizeof(local_server_names[0]), 
                 "%s %4dms %s", pszProtocol, iPing, info ) ;

    local_server_netadr[m_num_servers] = adr;

    m_num_servers++;
}
// </serverping>
#else
void M_AddToServerList (netadr_t adr, char *info)
{
	int		i;

	if (m_num_servers == MAX_LOCAL_SERVERS)
		return;
	while ( *info == ' ' )
		info++;

	// ignore if duplicated
	for (i=0 ; i<m_num_servers ; i++)
		if (!strcmp(info, local_server_names[i]))
			return;

	local_server_netadr[m_num_servers] = adr;
	strncpy (local_server_names[m_num_servers], info, sizeof(local_server_names[0])-1);
	m_num_servers++;
}
#endif

void JoinServerFunc( void *self )
{
	char	buffer[128];
	int		index;

	index = ( menuaction_s * ) self - s_joinserver_server_actions;

	if ( Q_stricmp( local_server_names[index], NO_SERVER_STRING ) == 0 )
		return;

	if (index >= m_num_servers)
		return;

	Com_sprintf (buffer, sizeof(buffer), "connect %s\n", NET_AdrToString (local_server_netadr[index]));
	Cbuf_AddText (buffer);
	M_ForceMenuOff ();
	cls.disable_screen = 1; // Knightmare- show loading screen
}

void AddressBookFunc( void *self )
{
	M_Menu_AddressBook_f();
}

//Knightmare- init client compatibility menu option
static void JoinserverSetMenuItemValues( void )
{
	Cvar_SetValue( "cl_servertrick", ClampCvar( 0, 1, cl_servertrick->value ) );
	s_joinserver_compatibility_box.curvalue		= cl_servertrick->value;
}

void NullCursorDraw( void *self )
{
}

void SearchLocalGames( void )
{
	int		i;

	m_num_servers = 0;
	for (i=0 ; i<MAX_LOCAL_SERVERS ; i++)
		strcpy (local_server_names[i], NO_SERVER_STRING);

	M_DrawTextBox( 8, 120 - 48, 36, 3 );
	M_Print( 28, 120 - 48 + MENU_FONT_SIZE,  "Searching for local servers, this" );
	M_Print( 28, 120 - 48 + MENU_FONT_SIZE*2, "could take up to a minute, so" );
	M_Print( 28, 120 - 48 + MENU_FONT_SIZE*3, "please be patient." );

	// the text box won't show up unless we do a buffer swap
	re.EndFrame();

	// send out info packets
	CL_PingServers_f();
}

void SearchLocalGamesFunc( void *self )
{
	SearchLocalGames();
}


void JoinServer_MenuInit( void )
{
	int i;
	int y = 30;

	static const char *compatibility_names[] =
	{
		"version 35 (KMQuake2)",
		"version 34 (stock Quake2)",
		0
	};

	JoinserverSetMenuItemValues(); // init item values

	s_joinserver_menu.x = viddef.width * 0.50 - scaledVideo(120);
	s_joinserver_menu.nitems = 0;

	// Knightmare- init client compatibility menu option
	s_joinserver_compat_title.generic.type = MTYPE_SEPARATOR;
	s_joinserver_compat_title.generic.name = "client protcol compatibility";
	s_joinserver_compat_title.generic.x    = 200;
	s_joinserver_compat_title.generic.y	   = y; //30

	s_joinserver_compatibility_box.generic.type = MTYPE_SPINCONTROL;
	s_joinserver_compatibility_box.generic.name	= "";
	s_joinserver_compatibility_box.generic.x	= -32;
	s_joinserver_compatibility_box.generic.y	= y += (MENU_FONT_SIZE+2); //40
	s_joinserver_compatibility_box.generic.cursor_offset = -24;
	s_joinserver_compatibility_box.generic.callback = ClientCompatibilityFunc;
	s_joinserver_compatibility_box.generic.statusbar = "set to version 34 to ping non-KMQuake2 servers";
	s_joinserver_compatibility_box.itemnames = compatibility_names;
	// end Knightmare

	s_joinserver_address_book_action.generic.type	= MTYPE_ACTION;
	s_joinserver_address_book_action.generic.name	= "address book";
	s_joinserver_address_book_action.generic.flags	= QMF_LEFT_JUSTIFY;
	s_joinserver_address_book_action.generic.x		= 0;
	s_joinserver_address_book_action.generic.y		= y += 2*(MENU_FONT_SIZE+2); //60
	s_joinserver_address_book_action.generic.callback = AddressBookFunc;

	s_joinserver_search_action.generic.type = MTYPE_ACTION;
	s_joinserver_search_action.generic.name	= "refresh server list";
	s_joinserver_search_action.generic.flags	= QMF_LEFT_JUSTIFY;
	s_joinserver_search_action.generic.x	= 0;
	s_joinserver_search_action.generic.y	= y += (MENU_FONT_SIZE+2); //70
	s_joinserver_search_action.generic.callback = SearchLocalGamesFunc;
	s_joinserver_search_action.generic.statusbar = "search for servers";

	s_joinserver_server_title.generic.type = MTYPE_SEPARATOR;
	s_joinserver_server_title.generic.name = "connect to...";
	s_joinserver_server_title.generic.x    = 80;
	s_joinserver_server_title.generic.y	   = y += 2*(MENU_FONT_SIZE+2); //90

	y += (MENU_FONT_SIZE+2); //100
	for ( i = 0; i < MAX_LOCAL_SERVERS; i++ )
	{
		s_joinserver_server_actions[i].generic.type	= MTYPE_ACTION;
		strcpy (local_server_names[i], NO_SERVER_STRING);
		s_joinserver_server_actions[i].generic.name	= local_server_names[i];
		s_joinserver_server_actions[i].generic.flags	= QMF_LEFT_JUSTIFY;
		s_joinserver_server_actions[i].generic.x		= 0;
		s_joinserver_server_actions[i].generic.y		= y + i*(MENU_FONT_SIZE+2); //100 + i*10;
		s_joinserver_server_actions[i].generic.callback = JoinServerFunc;
		s_joinserver_server_actions[i].generic.statusbar = "press ENTER to connect";
	}

	// Knightmare- init client compatibility menu option
	Menu_AddItem( &s_joinserver_menu, &s_joinserver_compat_title );
	Menu_AddItem( &s_joinserver_menu, &s_joinserver_compatibility_box );

	Menu_AddItem( &s_joinserver_menu, &s_joinserver_address_book_action );
	Menu_AddItem( &s_joinserver_menu, &s_joinserver_server_title );
	Menu_AddItem( &s_joinserver_menu, &s_joinserver_search_action );

	for ( i = 0; i < 8; i++ )
		Menu_AddItem( &s_joinserver_menu, &s_joinserver_server_actions[i] );

	Menu_Center( &s_joinserver_menu );

	// skip over compatibility title
	if (s_joinserver_menu.cursor == 0)
		s_joinserver_menu.cursor = 1;

	SearchLocalGames();
}

void JoinServer_MenuDraw(void)
{
	M_Banner( "m_banner_join_server" );
	Menu_Draw( &s_joinserver_menu );
}


const char *JoinServer_MenuKey( int key )
{
	return Default_MenuKey( &s_joinserver_menu, key );
}

void M_Menu_JoinServer_f (void)
{
	JoinServer_MenuInit();
	M_PushMenu( JoinServer_MenuDraw, JoinServer_MenuKey );
}


/*
=============================================================================

START SERVER MENU

=============================================================================
*/
static menuframework_s s_startserver_menu;
static char **mapnames;
static int	  nummaps;

static menuaction_s	s_startserver_start_action;
static menuaction_s	s_startserver_dmoptions_action;
static menufield_s	s_timelimit_field;
static menufield_s	s_fraglimit_field;
static menufield_s	s_maxclients_field;
static menufield_s	s_hostname_field;
static menulist_s	s_startmap_list;
static menulist_s	s_rules_box;

void DMOptionsFunc( void *self )
{
	if (s_rules_box.curvalue == 1)
		return;
	M_Menu_DMOptions_f();
}

void RulesChangeFunc ( void *self )
{
	if (s_rules_box.curvalue == 0) 	// DM
	{
		s_maxclients_field.generic.statusbar = NULL;
		s_startserver_dmoptions_action.generic.statusbar = NULL;
	}
	else if(s_rules_box.curvalue == 1)		// coop				// PGM
	{
		s_maxclients_field.generic.statusbar = "4 maximum for cooperative";
		if (atoi(s_maxclients_field.buffer) > 4)
			strcpy( s_maxclients_field.buffer, "4" );
		s_startserver_dmoptions_action.generic.statusbar = "N/A for cooperative";
	}
	// ROGUE GAMES
	//Knightmare 12/23/2001
	else if (roguepath() && s_rules_box.curvalue == 2) // tag	
	{
		s_maxclients_field.generic.statusbar = NULL;
		s_startserver_dmoptions_action.generic.statusbar = NULL;
	}
	else if ( (roguepath() && s_rules_box.curvalue == 3) // CTF
		|| (!roguepath() && s_rules_box.curvalue == 2) )
	{
		s_maxclients_field.generic.statusbar = NULL;
		if (atoi(s_maxclients_field.buffer) <= 12) // set default of 8
			strcpy( s_maxclients_field.buffer, "12" );
		s_startserver_dmoptions_action.generic.statusbar = NULL;
	}
	else if ( (roguepath() && s_rules_box.curvalue == 4) // 3Team CTF
		|| (!roguepath() && s_rules_box.curvalue == 3) )
	{
		s_maxclients_field.generic.statusbar = NULL;
		if (atoi(s_maxclients_field.buffer) <= 18) // set default of 8
			strcpy( s_maxclients_field.buffer, "18" );
		s_startserver_dmoptions_action.generic.statusbar = NULL;
	}
}

void StartServerActionFunc( void *self )
{
	char	startmap[1024];
	int		timelimit;
	int		fraglimit;
	int		maxclients;
	char	*spot;

	strcpy( startmap, strchr( mapnames[s_startmap_list.curvalue], '\n' ) + 1 );

	maxclients  = atoi( s_maxclients_field.buffer );
	timelimit	= atoi( s_timelimit_field.buffer );
	fraglimit	= atoi( s_fraglimit_field.buffer );

	Cvar_SetValue( "maxclients", ClampCvar( 0, maxclients, maxclients ) );
	Cvar_SetValue ("timelimit", ClampCvar( 0, timelimit, timelimit ) );
	Cvar_SetValue ("fraglimit", ClampCvar( 0, fraglimit, fraglimit ) );
	Cvar_Set("hostname", s_hostname_field.buffer );

	// Knightmare- redid this
	if (roguepath()) {
		Cvar_SetValue ("deathmatch", s_rules_box.curvalue != 1);
		Cvar_SetValue ("coop", s_rules_box.curvalue == 1);
		Cvar_SetValue ("gamerules", s_rules_box.curvalue == 2 );
		Cvar_SetValue ("ctf", s_rules_box.curvalue >= 3 );
		Cvar_SetValue ("ttctf", s_rules_box.curvalue == 4 );
	} else {
		Cvar_SetValue ("deathmatch", s_rules_box.curvalue != 1);
		Cvar_SetValue ("coop", s_rules_box.curvalue == 1);
		Cvar_SetValue ("gamerules", 0 );
		Cvar_SetValue ("ctf", s_rules_box.curvalue >= 2 );
		Cvar_SetValue ("ttctf", s_rules_box.curvalue == 3 );
	}

	spot = NULL;
	if (s_rules_box.curvalue == 1)		// PGM
	{
 		if(Q_stricmp(startmap, "bunk1") == 0)
  			spot = "start";
 		else if(Q_stricmp(startmap, "mintro") == 0)
  			spot = "start";
 		else if(Q_stricmp(startmap, "fact1") == 0)
  			spot = "start";
 		else if(Q_stricmp(startmap, "power1") == 0)
  			spot = "pstart";
 		else if(Q_stricmp(startmap, "biggun") == 0)
  			spot = "bstart";
 		else if(Q_stricmp(startmap, "hangar1") == 0)
  			spot = "unitstart";
 		else if(Q_stricmp(startmap, "city1") == 0)
  			spot = "unitstart";
 		else if(Q_stricmp(startmap, "boss1") == 0)
			spot = "bosstart";
	}

	if (spot)
	{
		if (Com_ServerState())
			Cbuf_AddText ("disconnect\n");
		Cbuf_AddText (va("gamemap \"*%s$%s\"\n", startmap, spot));
	}
	else
	{
		Cbuf_AddText (va("map %s\n", startmap));
	}

	M_ForceMenuOff ();
}

void StartServer_MenuInit( void )
{
	static const char *dm_coop_names[] =
	{
		"deathmatch",
		"cooperative",
		"CTF",
		"3Team CTF",
		0
	};
//=======
//PGM
	static const char *dm_coop_names_rogue[] =
	{
		"deathmatch",
		"cooperative",
		"tag",
		"CTF",
		"3Team CTF",
//		"deathball",
		0
	};
//PGM
//=======
	char *buffer;
	char  mapsname[1024];
	char *s;
	int length;
	int i;
	int y = 0;
	FILE *fp;

	/*
	** load the list of map names
	*/
	Com_sprintf( mapsname, sizeof( mapsname ), "%s/maps.lst", FS_Gamedir() );
	if ( ( fp = fopen( mapsname, "rb" ) ) == 0 )
	{
		if ( ( length = FS_LoadFile( "maps.lst", ( void ** ) &buffer ) ) == -1 )
			Com_Error( ERR_DROP, "couldn't find maps.lst\n" );
	}
	else
	{
#ifdef _WIN32
		length = filelength( fileno( fp  ) );
#else
		fseek(fp, 0, SEEK_END);
		length = ftell(fp);
		fseek(fp, 0, SEEK_SET);
#endif
		buffer = malloc( length );
		fread( buffer, length, 1, fp );
	}

	s = buffer;

	i = 0;
	while ( i < length )
	{
		if ( s[i] == '\r' )
			nummaps++;
		i++;
	}

	if ( nummaps == 0 )
		Com_Error( ERR_DROP, "no maps in maps.lst\n" );

	mapnames = malloc( sizeof( char * ) * ( nummaps + 1 ) );
	memset( mapnames, 0, sizeof( char * ) * ( nummaps + 1 ) );

	s = buffer;

	for ( i = 0; i < nummaps; i++ )
	{
    char  shortname[MAX_TOKEN_CHARS];
    char  longname[MAX_TOKEN_CHARS];
		char  scratch[200];
		int		j, l;

		strcpy( shortname, COM_Parse( &s ) );
		l = strlen(shortname);
		for (j=0 ; j<l ; j++)
			shortname[j] = toupper(shortname[j]);
		strcpy( longname, COM_Parse( &s ) );
		Com_sprintf( scratch, sizeof( scratch ), "%s\n%s", longname, shortname );

		mapnames[i] = malloc( strlen( scratch ) + 1 );
		strcpy( mapnames[i], scratch );
	}
	mapnames[nummaps] = 0;

	if ( fp != 0 )
	{
		fp = 0;
		free( buffer );
	}
	else
	{
		FS_FreeFile( buffer );
	}

	/*
	** initialize the menu stuff
	*/
	s_startserver_menu.x = viddef.width * 0.50;
	s_startserver_menu.nitems = 0;

	s_startmap_list.generic.type = MTYPE_SPINCONTROL;
	s_startmap_list.generic.x	= 0;
	s_startmap_list.generic.y	= y; //0
	s_startmap_list.generic.name	= "initial map";
	s_startmap_list.itemnames = mapnames;

	s_rules_box.generic.type = MTYPE_SPINCONTROL;
	s_rules_box.generic.x	= 0;
	s_rules_box.generic.y	= y += 2*(MENU_FONT_SIZE+2); //20
	s_rules_box.generic.name	= "rules";
//PGM - rogue games only available with rogue DLL.
	if (roguepath())
		s_rules_box.itemnames = dm_coop_names_rogue;
	else
		s_rules_box.itemnames = dm_coop_names;
//PGM
	if (roguepath()) {
		if (Cvar_VariableValue("ttctf"))
			s_rules_box.curvalue = 4;
		else if (Cvar_VariableValue("ctf"))
			s_rules_box.curvalue = 3;
		else if (Cvar_VariableValue("gamerules"))
			s_rules_box.curvalue = 2;
		else if (Cvar_VariableValue("coop"))
			s_rules_box.curvalue = 1;
		else
			s_rules_box.curvalue = 0;
	} else {
		if (Cvar_VariableValue("ttctf"))
			s_rules_box.curvalue = 3;
		else if (Cvar_VariableValue("ctf"))
			s_rules_box.curvalue = 2;
		else if (Cvar_VariableValue("coop"))
			s_rules_box.curvalue = 1;
		else
			s_rules_box.curvalue = 0;
	}
	s_rules_box.generic.callback = RulesChangeFunc;

	s_timelimit_field.generic.type = MTYPE_FIELD;
	s_timelimit_field.generic.name = "time limit";
	s_timelimit_field.generic.flags = QMF_NUMBERSONLY;
	s_timelimit_field.generic.x	= 0;
	s_timelimit_field.generic.y	= y += 2*MENU_FONT_SIZE; //36
	s_timelimit_field.generic.statusbar = "0 = no limit";
	s_timelimit_field.length = 4;
	s_timelimit_field.visible_length = 4;
	strcpy( s_timelimit_field.buffer, Cvar_VariableString("timelimit") );
	s_timelimit_field.cursor = strlen( s_timelimit_field.buffer );

	s_fraglimit_field.generic.type = MTYPE_FIELD;
	s_fraglimit_field.generic.name = "frag limit";
	s_fraglimit_field.generic.flags = QMF_NUMBERSONLY;
	s_fraglimit_field.generic.x	= 0;
	s_fraglimit_field.generic.y	= y += 2.25*MENU_FONT_SIZE; //54
	s_fraglimit_field.generic.statusbar = "0 = no limit";
	s_fraglimit_field.length = 4;
	s_fraglimit_field.visible_length = 4;
	strcpy( s_fraglimit_field.buffer, Cvar_VariableString("fraglimit") );
	s_fraglimit_field.cursor = strlen( s_fraglimit_field.buffer );

	/*
	** maxclients determines the maximum number of players that can join
	** the game.  If maxclients is only "1" then we should default the menu
	** option to 8 players, otherwise use whatever its current value is. 
	** Clamping will be done when the server is actually started.
	*/
	s_maxclients_field.generic.type = MTYPE_FIELD;
	s_maxclients_field.generic.name = "max players";
	s_maxclients_field.generic.flags = QMF_NUMBERSONLY;
	s_maxclients_field.generic.x	= 0;
	s_maxclients_field.generic.y	= y += 2.25*MENU_FONT_SIZE; //72
	s_maxclients_field.generic.statusbar = NULL;
	s_maxclients_field.length = 3;
	s_maxclients_field.visible_length = 3;
	if ( Cvar_VariableValue( "maxclients" ) == 1 )
		strcpy( s_maxclients_field.buffer, "8" );
	else 
		strcpy( s_maxclients_field.buffer, Cvar_VariableString("maxclients") );
	s_maxclients_field.cursor = strlen( s_maxclients_field.buffer );

	s_hostname_field.generic.type = MTYPE_FIELD;
	s_hostname_field.generic.name = "hostname";
	s_hostname_field.generic.flags = 0;
	s_hostname_field.generic.x	= 0;
	s_hostname_field.generic.y	= y += 2.25*MENU_FONT_SIZE; //90
	s_hostname_field.generic.statusbar = NULL;
	s_hostname_field.length = 12;
	s_hostname_field.visible_length = 12;
	strcpy( s_hostname_field.buffer, Cvar_VariableString("hostname") );
	s_hostname_field.cursor = strlen( s_hostname_field.buffer );

	s_startserver_dmoptions_action.generic.type = MTYPE_ACTION;
	s_startserver_dmoptions_action.generic.name	= " deathmatch flags";
	s_startserver_dmoptions_action.generic.flags= QMF_LEFT_JUSTIFY;
	s_startserver_dmoptions_action.generic.x	= 24;
	s_startserver_dmoptions_action.generic.y	= y += 2.25*MENU_FONT_SIZE; //108
	s_startserver_dmoptions_action.generic.statusbar = NULL;
	s_startserver_dmoptions_action.generic.callback = DMOptionsFunc;

	s_startserver_start_action.generic.type = MTYPE_ACTION;
	s_startserver_start_action.generic.name	= " begin";
	s_startserver_start_action.generic.flags= QMF_LEFT_JUSTIFY;
	s_startserver_start_action.generic.x	= 24;
	s_startserver_start_action.generic.y	= y += 2*(MENU_FONT_SIZE+2); //128
	s_startserver_start_action.generic.callback = StartServerActionFunc;

	Menu_AddItem( &s_startserver_menu, &s_startmap_list );
	Menu_AddItem( &s_startserver_menu, &s_rules_box );
	Menu_AddItem( &s_startserver_menu, &s_timelimit_field );
	Menu_AddItem( &s_startserver_menu, &s_fraglimit_field );
	Menu_AddItem( &s_startserver_menu, &s_maxclients_field );
	Menu_AddItem( &s_startserver_menu, &s_hostname_field );
	Menu_AddItem( &s_startserver_menu, &s_startserver_dmoptions_action );
	Menu_AddItem( &s_startserver_menu, &s_startserver_start_action );

	Menu_Center( &s_startserver_menu );

	// call this now to set proper inital state
	RulesChangeFunc ( NULL );
}

void StartServer_MenuDraw(void)
{
	M_Banner( "m_banner_start_server" ); // Knightmare added
	Menu_Draw( &s_startserver_menu );
}

const char *StartServer_MenuKey( int key )
{
	if ( key == K_ESCAPE )
	{
		if ( mapnames )
		{
			int i;

			for ( i = 0; i < nummaps; i++ )
				free( mapnames[i] );
			free( mapnames );
		}
		mapnames = 0;
		nummaps = 0;
	}

	return Default_MenuKey( &s_startserver_menu, key );
}

void M_Menu_StartServer_f (void)
{
	StartServer_MenuInit();
	M_PushMenu( StartServer_MenuDraw, StartServer_MenuKey );
}

/*
=============================================================================

DMOPTIONS BOOK MENU

=============================================================================
*/
static char dmoptions_statusbar[128];

static menuframework_s s_dmoptions_menu;

static menulist_s	s_friendlyfire_box;
static menulist_s	s_falls_box;
static menulist_s	s_weapons_stay_box;
static menulist_s	s_instant_powerups_box;
static menulist_s	s_powerups_box;
static menulist_s	s_health_box;
static menulist_s	s_spawn_farthest_box;
static menulist_s	s_teamplay_box;
static menulist_s	s_samelevel_box;
static menulist_s	s_force_respawn_box;
static menulist_s	s_armor_box;
static menulist_s	s_allow_exit_box;
static menulist_s	s_infinite_ammo_box;
static menulist_s	s_fixed_fov_box;
static menulist_s	s_quad_drop_box;

//Xatrix
static menulist_s	s_quadfire_drop_box;

//ROGUE
static menulist_s	s_no_mines_box;
static menulist_s	s_no_nukes_box;
static menulist_s	s_stack_double_box;
static menulist_s	s_no_spheres_box;

// CTF
static menulist_s	s_ctf_forceteam_box;
static menulist_s	s_ctf_armor_protect_box;
static menulist_s	s_ctf_notechs_box;


#define DF_CTF_FORCEJOIN	131072	
#define DF_ARMOR_PROTECT	262144
#define DF_CTF_NO_TECH      524288

qboolean CTF_menumode (void)
{
	if ( (roguepath() && s_rules_box.curvalue >= 3)
		|| (!roguepath() && s_rules_box.curvalue >= 2) )
		return true;
	return false;
}

static void DMFlagCallback( void *self )
{
	menulist_s *f = ( menulist_s * ) self;
	int flags;
	int bit = 0;

	flags = Cvar_VariableValue( "dmflags" );

	if ( f == &s_friendlyfire_box )
	{
		if ( f->curvalue )
			flags &= ~DF_NO_FRIENDLY_FIRE;
		else
			flags |= DF_NO_FRIENDLY_FIRE;
		goto setvalue;
	}
	else if ( f == &s_falls_box )
	{
		if ( f->curvalue )
			flags &= ~DF_NO_FALLING;
		else
			flags |= DF_NO_FALLING;
		goto setvalue;
	}
	else if ( f == &s_weapons_stay_box ) 
	{
		bit = DF_WEAPONS_STAY;
	}
	else if ( f == &s_instant_powerups_box )
	{
		bit = DF_INSTANT_ITEMS;
	}
	else if ( f == &s_allow_exit_box )
	{
		bit = DF_ALLOW_EXIT;
	}
	else if ( f == &s_powerups_box )
	{
		if ( f->curvalue )
			flags &= ~DF_NO_ITEMS;
		else
			flags |= DF_NO_ITEMS;
		goto setvalue;
	}
	else if ( f == &s_health_box )
	{
		if ( f->curvalue )
			flags &= ~DF_NO_HEALTH;
		else
			flags |= DF_NO_HEALTH;
		goto setvalue;
	}
	else if ( f == &s_spawn_farthest_box )
	{
		bit = DF_SPAWN_FARTHEST;
	}
	else if ( f == &s_teamplay_box )
	{
		if ( f->curvalue == 1 )
		{
			flags |=  DF_SKINTEAMS;
			flags &= ~DF_MODELTEAMS;
		}
		else if ( f->curvalue == 2 )
		{
			flags |=  DF_MODELTEAMS;
			flags &= ~DF_SKINTEAMS;
		}
		else
		{
			flags &= ~( DF_MODELTEAMS | DF_SKINTEAMS );
		}

		goto setvalue;
	}
	else if ( f == &s_samelevel_box )
	{
		bit = DF_SAME_LEVEL;
	}
	else if ( f == &s_force_respawn_box )
	{
		bit = DF_FORCE_RESPAWN;
	}
	else if ( f == &s_armor_box )
	{
		if ( f->curvalue )
			flags &= ~DF_NO_ARMOR;
		else
			flags |= DF_NO_ARMOR;
		goto setvalue;
	}
	else if ( f == &s_infinite_ammo_box )
	{
		bit = DF_INFINITE_AMMO;
	}
	else if ( f == &s_fixed_fov_box )
	{
		bit = DF_FIXED_FOV;
	}
	else if ( f == &s_quad_drop_box )
	{
		bit = DF_QUAD_DROP;
	}
	// Knightmare added
	else if (modType("xatrix"))
	{
		if ( f == &s_quadfire_drop_box)
		{
			bit = DF_QUADFIRE_DROP;
		}
	}
//=======
//ROGUE
	else if (roguepath())
	{
		if ( f == &s_no_mines_box)
		{
			bit = DF_NO_MINES;
		}
		else if ( f == &s_no_nukes_box)
		{
			bit = DF_NO_NUKES;
		}
		else if ( f == &s_stack_double_box)
		{
			bit = DF_NO_STACK_DOUBLE;
		}
		else if ( f == &s_no_spheres_box)
		{
			bit = DF_NO_SPHERES;
		}
	}
//ROGUE
//=======
	// Knightmare added-  CTF flags
	else if ( CTF_menumode() )
	{
		if ( f == &s_ctf_forceteam_box)
		{
			bit = DF_CTF_FORCEJOIN;
		}
		else if ( f == &s_ctf_armor_protect_box)
		{
			bit = DF_ARMOR_PROTECT;
		}
		else if ( f == &s_ctf_notechs_box)
		{
			bit = DF_CTF_NO_TECH;
		}
	}

	if ( f )
	{
		if ( f->curvalue == 0 )
			flags &= ~bit;
		else
			flags |= bit;
	}

setvalue:
	Cvar_SetValue ("dmflags", flags);

	Com_sprintf( dmoptions_statusbar, sizeof( dmoptions_statusbar ), "dmflags = %d", flags );

}

void DMOptions_MenuInit( void )
{
	static const char *yes_no_names[] =
	{
		"no", "yes", 0
	};
	static const char *teamplay_names[] = 
	{
		"disabled", "by skin", "by model", 0
	};
	int dmflags = Cvar_VariableValue( "dmflags" );
	int y = 0;

	s_dmoptions_menu.x = viddef.width * 0.50;
	s_dmoptions_menu.nitems = 0;

	s_falls_box.generic.type = MTYPE_SPINCONTROL;
	s_falls_box.generic.x	= 0;
	s_falls_box.generic.y	= y;
	s_falls_box.generic.name	= "falling damage";
	s_falls_box.generic.callback = DMFlagCallback;
	s_falls_box.itemnames = yes_no_names;
	s_falls_box.curvalue = ( dmflags & DF_NO_FALLING ) == 0;

	s_weapons_stay_box.generic.type = MTYPE_SPINCONTROL;
	s_weapons_stay_box.generic.x	= 0;
	s_weapons_stay_box.generic.y	= y += (MENU_FONT_SIZE+2); //10
	s_weapons_stay_box.generic.name	= "weapons stay";
	s_weapons_stay_box.generic.callback = DMFlagCallback;
	s_weapons_stay_box.itemnames = yes_no_names;
	s_weapons_stay_box.curvalue = ( dmflags & DF_WEAPONS_STAY ) != 0;

	s_instant_powerups_box.generic.type = MTYPE_SPINCONTROL;
	s_instant_powerups_box.generic.x	= 0;
	s_instant_powerups_box.generic.y	= y += (MENU_FONT_SIZE+2); //20
	s_instant_powerups_box.generic.name	= "instant powerups";
	s_instant_powerups_box.generic.callback = DMFlagCallback;
	s_instant_powerups_box.itemnames = yes_no_names;
	s_instant_powerups_box.curvalue = ( dmflags & DF_INSTANT_ITEMS ) != 0;

	s_powerups_box.generic.type = MTYPE_SPINCONTROL;
	s_powerups_box.generic.x	= 0;
	s_powerups_box.generic.y	= y += (MENU_FONT_SIZE+2); //30
	s_powerups_box.generic.name	= "allow powerups";
	s_powerups_box.generic.callback = DMFlagCallback;
	s_powerups_box.itemnames = yes_no_names;
	s_powerups_box.curvalue = ( dmflags & DF_NO_ITEMS ) == 0;

	s_health_box.generic.type = MTYPE_SPINCONTROL;
	s_health_box.generic.x	= 0;
	s_health_box.generic.y	= y += (MENU_FONT_SIZE+2); //40
	s_health_box.generic.callback = DMFlagCallback;
	s_health_box.generic.name	= "allow health";
	s_health_box.itemnames = yes_no_names;
	s_health_box.curvalue = ( dmflags & DF_NO_HEALTH ) == 0;

	s_armor_box.generic.type = MTYPE_SPINCONTROL;
	s_armor_box.generic.x	= 0;
	s_armor_box.generic.y	= y += (MENU_FONT_SIZE+2); //50
	s_armor_box.generic.name	= "allow armor";
	s_armor_box.generic.callback = DMFlagCallback;
	s_armor_box.itemnames = yes_no_names;
	s_armor_box.curvalue = ( dmflags & DF_NO_ARMOR ) == 0;

	s_spawn_farthest_box.generic.type = MTYPE_SPINCONTROL;
	s_spawn_farthest_box.generic.x	= 0;
	s_spawn_farthest_box.generic.y	= y += (MENU_FONT_SIZE+2); //60
	s_spawn_farthest_box.generic.name	= "spawn farthest";
	s_spawn_farthest_box.generic.callback = DMFlagCallback;
	s_spawn_farthest_box.itemnames = yes_no_names;
	s_spawn_farthest_box.curvalue = ( dmflags & DF_SPAWN_FARTHEST ) != 0;

	s_samelevel_box.generic.type = MTYPE_SPINCONTROL;
	s_samelevel_box.generic.x	= 0;
	s_samelevel_box.generic.y	= y += (MENU_FONT_SIZE+2); //70
	s_samelevel_box.generic.name	= "same map";
	s_samelevel_box.generic.callback = DMFlagCallback;
	s_samelevel_box.itemnames = yes_no_names;
	s_samelevel_box.curvalue = ( dmflags & DF_SAME_LEVEL ) != 0;

	s_force_respawn_box.generic.type = MTYPE_SPINCONTROL;
	s_force_respawn_box.generic.x	= 0;
	s_force_respawn_box.generic.y	= y += (MENU_FONT_SIZE+2); //80
	s_force_respawn_box.generic.name	= "force respawn";
	s_force_respawn_box.generic.callback = DMFlagCallback;
	s_force_respawn_box.itemnames = yes_no_names;
	s_force_respawn_box.curvalue = ( dmflags & DF_FORCE_RESPAWN ) != 0;

	s_teamplay_box.generic.type = MTYPE_SPINCONTROL;
	s_teamplay_box.generic.x	= 0;
	s_teamplay_box.generic.y	= y += (MENU_FONT_SIZE+2); //90
	s_teamplay_box.generic.name	= "teamplay";
	s_teamplay_box.generic.callback = DMFlagCallback;
	s_teamplay_box.itemnames = teamplay_names;

	s_allow_exit_box.generic.type = MTYPE_SPINCONTROL;
	s_allow_exit_box.generic.x	= 0;
	s_allow_exit_box.generic.y	= y += (MENU_FONT_SIZE+2); //100
	s_allow_exit_box.generic.name	= "allow exit";
	s_allow_exit_box.generic.callback = DMFlagCallback;
	s_allow_exit_box.itemnames = yes_no_names;
	s_allow_exit_box.curvalue = ( dmflags & DF_ALLOW_EXIT ) != 0;

	s_infinite_ammo_box.generic.type = MTYPE_SPINCONTROL;
	s_infinite_ammo_box.generic.x	= 0;
	s_infinite_ammo_box.generic.y	= y += (MENU_FONT_SIZE+2); //110
	s_infinite_ammo_box.generic.name	= "infinite ammo";
	s_infinite_ammo_box.generic.callback = DMFlagCallback;
	s_infinite_ammo_box.itemnames = yes_no_names;
	s_infinite_ammo_box.curvalue = ( dmflags & DF_INFINITE_AMMO ) != 0;

	s_fixed_fov_box.generic.type = MTYPE_SPINCONTROL;
	s_fixed_fov_box.generic.x	= 0;
	s_fixed_fov_box.generic.y	= y += (MENU_FONT_SIZE+2); //120
	s_fixed_fov_box.generic.name	= "fixed FOV";
	s_fixed_fov_box.generic.callback = DMFlagCallback;
	s_fixed_fov_box.itemnames = yes_no_names;
	s_fixed_fov_box.curvalue = ( dmflags & DF_FIXED_FOV ) != 0;

	s_quad_drop_box.generic.type = MTYPE_SPINCONTROL;
	s_quad_drop_box.generic.x	= 0;
	s_quad_drop_box.generic.y	= y += (MENU_FONT_SIZE+2); //130
	s_quad_drop_box.generic.name	= "quad drop";
	s_quad_drop_box.generic.callback = DMFlagCallback;
	s_quad_drop_box.itemnames = yes_no_names;
	s_quad_drop_box.curvalue = ( dmflags & DF_QUAD_DROP ) != 0;

	s_friendlyfire_box.generic.type = MTYPE_SPINCONTROL;
	s_friendlyfire_box.generic.x	= 0;
	s_friendlyfire_box.generic.y	= y += (MENU_FONT_SIZE+2); //140
	s_friendlyfire_box.generic.name	= "friendly fire";
	s_friendlyfire_box.generic.callback = DMFlagCallback;
	s_friendlyfire_box.itemnames = yes_no_names;
	s_friendlyfire_box.curvalue = ( dmflags & DF_NO_FRIENDLY_FIRE ) == 0;

	// Knightmare added
	if (modType("xatrix"))
	{
		s_quadfire_drop_box.generic.type = MTYPE_SPINCONTROL;
		s_quadfire_drop_box.generic.x	= 0;
		s_quadfire_drop_box.generic.y	= y += (MENU_FONT_SIZE+2); //150
		s_quadfire_drop_box.generic.name	= "dualfire drop";
		s_quadfire_drop_box.generic.callback = DMFlagCallback;
		s_quadfire_drop_box.itemnames = yes_no_names;
		s_quadfire_drop_box.curvalue = ( dmflags & DF_QUADFIRE_DROP ) != 0;
	}
//============
//ROGUE
	//Knightmare 12/23/2001
	else if (roguepath())
	{
		s_no_mines_box.generic.type = MTYPE_SPINCONTROL;
		s_no_mines_box.generic.x	= 0;
		s_no_mines_box.generic.y	= y += (MENU_FONT_SIZE+2); //150
		s_no_mines_box.generic.name	= "remove mines";
		s_no_mines_box.generic.callback = DMFlagCallback;
		s_no_mines_box.itemnames = yes_no_names;
		s_no_mines_box.curvalue = ( dmflags & DF_NO_MINES ) != 0;

		s_no_nukes_box.generic.type = MTYPE_SPINCONTROL;
		s_no_nukes_box.generic.x	= 0;
		s_no_nukes_box.generic.y	= y += (MENU_FONT_SIZE+2); //160
		s_no_nukes_box.generic.name	= "remove nukes";
		s_no_nukes_box.generic.callback = DMFlagCallback;
		s_no_nukes_box.itemnames = yes_no_names;
		s_no_nukes_box.curvalue = ( dmflags & DF_NO_NUKES ) != 0;

		s_stack_double_box.generic.type = MTYPE_SPINCONTROL;
		s_stack_double_box.generic.x	= 0;
		s_stack_double_box.generic.y	= y += (MENU_FONT_SIZE+2); //170
		s_stack_double_box.generic.name	= "2x/4x stacking off";
		s_stack_double_box.generic.callback = DMFlagCallback;
		s_stack_double_box.itemnames = yes_no_names;
		s_stack_double_box.curvalue = ( dmflags & DF_NO_STACK_DOUBLE ) != 0;

		s_no_spheres_box.generic.type = MTYPE_SPINCONTROL;
		s_no_spheres_box.generic.x	= 0;
		s_no_spheres_box.generic.y	= y += (MENU_FONT_SIZE+2); //180
		s_no_spheres_box.generic.name	= "remove spheres";
		s_no_spheres_box.generic.callback = DMFlagCallback;
		s_no_spheres_box.itemnames = yes_no_names;
		s_no_spheres_box.curvalue = ( dmflags & DF_NO_SPHERES ) != 0;

	}
//ROGUE
//============
	// Knightmare added
	else if (CTF_menumode())
	{
		s_ctf_forceteam_box.generic.type = MTYPE_SPINCONTROL;
		s_ctf_forceteam_box.generic.x	= 0;
		s_ctf_forceteam_box.generic.y	= y += (MENU_FONT_SIZE+2); //150
		s_ctf_forceteam_box.generic.name	= "force team join";
		s_ctf_forceteam_box.generic.callback = DMFlagCallback;
		s_ctf_forceteam_box.itemnames = yes_no_names;
		s_ctf_forceteam_box.curvalue = ( dmflags & DF_CTF_FORCEJOIN ) != 0;

		s_ctf_armor_protect_box.generic.type = MTYPE_SPINCONTROL;
		s_ctf_armor_protect_box.generic.x	= 0;
		s_ctf_armor_protect_box.generic.y	= y += (MENU_FONT_SIZE+2); //160
		s_ctf_armor_protect_box.generic.name	= "team armor protect";
		s_ctf_armor_protect_box.generic.callback = DMFlagCallback;
		s_ctf_armor_protect_box.itemnames = yes_no_names;
		s_ctf_armor_protect_box.curvalue = ( dmflags & DF_ARMOR_PROTECT ) != 0;

		s_ctf_notechs_box.generic.type = MTYPE_SPINCONTROL;
		s_ctf_notechs_box.generic.x	= 0;
		s_ctf_notechs_box.generic.y	= y += (MENU_FONT_SIZE+2); //170
		s_ctf_notechs_box.generic.name	= "disable techs";
		s_ctf_notechs_box.generic.callback = DMFlagCallback;
		s_ctf_notechs_box.itemnames = yes_no_names;
		s_ctf_notechs_box.curvalue = ( dmflags & DF_CTF_NO_TECH ) != 0;
	}

	Menu_AddItem( &s_dmoptions_menu, &s_falls_box );
	Menu_AddItem( &s_dmoptions_menu, &s_weapons_stay_box );
	Menu_AddItem( &s_dmoptions_menu, &s_instant_powerups_box );
	Menu_AddItem( &s_dmoptions_menu, &s_powerups_box );
	Menu_AddItem( &s_dmoptions_menu, &s_health_box );
	Menu_AddItem( &s_dmoptions_menu, &s_armor_box );
	Menu_AddItem( &s_dmoptions_menu, &s_spawn_farthest_box );
	Menu_AddItem( &s_dmoptions_menu, &s_samelevel_box );
	Menu_AddItem( &s_dmoptions_menu, &s_force_respawn_box );
	Menu_AddItem( &s_dmoptions_menu, &s_teamplay_box );
	Menu_AddItem( &s_dmoptions_menu, &s_allow_exit_box );
	Menu_AddItem( &s_dmoptions_menu, &s_infinite_ammo_box );
	Menu_AddItem( &s_dmoptions_menu, &s_fixed_fov_box );
	Menu_AddItem( &s_dmoptions_menu, &s_quad_drop_box );
	Menu_AddItem( &s_dmoptions_menu, &s_friendlyfire_box );
	// Xatrix
	if (modType("xatrix"))
		Menu_AddItem( &s_dmoptions_menu, &s_quadfire_drop_box );
//=======
//ROGUE
	else if (roguepath())
	{
		Menu_AddItem( &s_dmoptions_menu, &s_no_mines_box );
		Menu_AddItem( &s_dmoptions_menu, &s_no_nukes_box );
		Menu_AddItem( &s_dmoptions_menu, &s_stack_double_box );
		Menu_AddItem( &s_dmoptions_menu, &s_no_spheres_box );
	}
//ROGUE
//=======
	// CTF
	else if (CTF_menumode())
	{
		Menu_AddItem( &s_dmoptions_menu, &s_ctf_forceteam_box );
		Menu_AddItem( &s_dmoptions_menu, &s_ctf_armor_protect_box );
		Menu_AddItem( &s_dmoptions_menu, &s_ctf_notechs_box );
	}

	Menu_Center( &s_dmoptions_menu );

	// set the original dmflags statusbar
	DMFlagCallback( 0 );
	Menu_SetStatusBar( &s_dmoptions_menu, dmoptions_statusbar );
}

void DMOptions_MenuDraw(void)
{
	M_Banner( "m_banner_start_server" ); // Knightmare added
	Menu_Draw( &s_dmoptions_menu );
}

const char *DMOptions_MenuKey( int key )
{
	return Default_MenuKey( &s_dmoptions_menu, key );
}

void M_Menu_DMOptions_f (void)
{
	DMOptions_MenuInit();
	M_PushMenu( DMOptions_MenuDraw, DMOptions_MenuKey );
}

/*
=============================================================================

DOWNLOADOPTIONS BOOK MENU

=============================================================================
*/
static menuframework_s s_downloadoptions_menu;

static menuseparator_s	s_download_title;
static menulist_s	s_allow_download_box;
static menulist_s	s_allow_download_maps_box;
// Knightmare- option to allow downloading 24-bit textures
static menulist_s	s_allow_download_enh_textures_box;

static menulist_s	s_allow_download_models_box;
static menulist_s	s_allow_download_players_box;
static menulist_s	s_allow_download_sounds_box;

static void DownloadCallback( void *self )
{
	menulist_s *f = ( menulist_s * ) self;

	if (f == &s_allow_download_box)
	{
		Cvar_SetValue("allow_download", f->curvalue);
	}

	else if (f == &s_allow_download_maps_box)
	{
		Cvar_SetValue("allow_download_maps", f->curvalue);
	}

	// Knightmare- option to allow downloading 24-bit textures
	else if (f == &s_allow_download_enh_textures_box)
	{
		Cvar_SetValue("allow_download_enh_textures", f->curvalue);
	}

	else if (f == &s_allow_download_models_box)
	{
		Cvar_SetValue("allow_download_models", f->curvalue);
	}

	else if (f == &s_allow_download_players_box)
	{
		Cvar_SetValue("allow_download_players", f->curvalue);
	}

	else if (f == &s_allow_download_sounds_box)
	{
		Cvar_SetValue("allow_download_sounds", f->curvalue);
	}
}

void DownloadOptions_MenuInit( void )
{
	static const char *yes_no_names[] =
	{
		"no", "yes", 0
	};
	int y = 0;

	s_downloadoptions_menu.x = viddef.width * 0.50;
	s_downloadoptions_menu.nitems = 0;

	s_download_title.generic.type = MTYPE_SEPARATOR;
	s_download_title.generic.name = "Download Options";
	s_download_title.generic.x    = 48;
	s_download_title.generic.y	 = y; //0

	s_allow_download_box.generic.type = MTYPE_SPINCONTROL;
	s_allow_download_box.generic.x	= 0;
	s_allow_download_box.generic.y	= y += 2*(MENU_FONT_SIZE+2); //20
	s_allow_download_box.generic.name	= "allow downloading";
	s_allow_download_box.generic.callback = DownloadCallback;
	s_allow_download_box.itemnames = yes_no_names;
	s_allow_download_box.curvalue = (Cvar_VariableValue("allow_download") != 0);

	s_allow_download_maps_box.generic.type = MTYPE_SPINCONTROL;
	s_allow_download_maps_box.generic.x	= 0;
	s_allow_download_maps_box.generic.y	= y += 2*(MENU_FONT_SIZE+2); //40
	s_allow_download_maps_box.generic.name	= "maps/textures";
	s_allow_download_maps_box.generic.callback = DownloadCallback;
	s_allow_download_maps_box.itemnames = yes_no_names;
	s_allow_download_maps_box.curvalue = (Cvar_VariableValue("allow_download_maps") != 0);

	// Knightmare- option to allow downloading 24-bit textures
	s_allow_download_enh_textures_box.generic.type = MTYPE_SPINCONTROL;
	s_allow_download_enh_textures_box.generic.x	= 0;
	s_allow_download_enh_textures_box.generic.y	= y += (MENU_FONT_SIZE+2); //50
	s_allow_download_enh_textures_box.generic.name	= "24-bit textures";
	s_allow_download_enh_textures_box.generic.callback = DownloadCallback;
	s_allow_download_enh_textures_box.generic.statusbar = "enable to allow downloading of JPG and TGA textures";
	s_allow_download_enh_textures_box.itemnames = yes_no_names;
	s_allow_download_enh_textures_box.curvalue = (Cvar_VariableValue("allow_download_enh_textures") != 0);

	s_allow_download_players_box.generic.type = MTYPE_SPINCONTROL;
	s_allow_download_players_box.generic.x	= 0;
	s_allow_download_players_box.generic.y	= y += (MENU_FONT_SIZE+2); //60
	s_allow_download_players_box.generic.name	= "player models/skins";
	s_allow_download_players_box.generic.callback = DownloadCallback;
	s_allow_download_players_box.itemnames = yes_no_names;
	s_allow_download_players_box.curvalue = (Cvar_VariableValue("allow_download_players") != 0);

	s_allow_download_models_box.generic.type = MTYPE_SPINCONTROL;
	s_allow_download_models_box.generic.x	= 0;
	s_allow_download_models_box.generic.y	= y += (MENU_FONT_SIZE+2); //70
	s_allow_download_models_box.generic.name	= "models";
	s_allow_download_models_box.generic.callback = DownloadCallback;
	s_allow_download_models_box.itemnames = yes_no_names;
	s_allow_download_models_box.curvalue = (Cvar_VariableValue("allow_download_models") != 0);

	s_allow_download_sounds_box.generic.type = MTYPE_SPINCONTROL;
	s_allow_download_sounds_box.generic.x	= 0;
	s_allow_download_sounds_box.generic.y	= y += (MENU_FONT_SIZE+2); //80
	s_allow_download_sounds_box.generic.name	= "sounds";
	s_allow_download_sounds_box.generic.callback = DownloadCallback;
	s_allow_download_sounds_box.itemnames = yes_no_names;
	s_allow_download_sounds_box.curvalue = (Cvar_VariableValue("allow_download_sounds") != 0);

	Menu_AddItem( &s_downloadoptions_menu, &s_download_title );
	Menu_AddItem( &s_downloadoptions_menu, &s_allow_download_box );
	Menu_AddItem( &s_downloadoptions_menu, &s_allow_download_maps_box );

	// Knightmare- option to allow downloading 24-bit textures
	Menu_AddItem( &s_downloadoptions_menu, &s_allow_download_enh_textures_box );

	Menu_AddItem( &s_downloadoptions_menu, &s_allow_download_players_box );
	Menu_AddItem( &s_downloadoptions_menu, &s_allow_download_models_box );
	Menu_AddItem( &s_downloadoptions_menu, &s_allow_download_sounds_box );

	Menu_Center( &s_downloadoptions_menu );

	// skip over title
	if (s_downloadoptions_menu.cursor == 0)
		s_downloadoptions_menu.cursor = 1;
}

void DownloadOptions_MenuDraw(void)
{
	Menu_Draw( &s_downloadoptions_menu );
}

const char *DownloadOptions_MenuKey( int key )
{
	return Default_MenuKey( &s_downloadoptions_menu, key );
}

void M_Menu_DownloadOptions_f (void)
{
	DownloadOptions_MenuInit();
	M_PushMenu( DownloadOptions_MenuDraw, DownloadOptions_MenuKey );
}
/*
=============================================================================

ADDRESS BOOK MENU

=============================================================================
*/
#define NUM_ADDRESSBOOK_ENTRIES 9

static menuframework_s	s_addressbook_menu;
static menufield_s		s_addressbook_fields[NUM_ADDRESSBOOK_ENTRIES];

void AddressBook_MenuInit( void )
{
	int i;

	s_addressbook_menu.x = viddef.width / 2 - scaledVideo(MENU_FONT_SIZE*17.75);	// was 142
	s_addressbook_menu.y = viddef.height / 2 - scaledVideo(MENU_FONT_SIZE*7.25);	// was 58
	s_addressbook_menu.nitems = 0;

	for ( i = 0; i < NUM_ADDRESSBOOK_ENTRIES; i++ )
	{
		cvar_t *adr;
		char buffer[20];

		Com_sprintf( buffer, sizeof( buffer ), "adr%d", i );

		adr = Cvar_Get( buffer, "", CVAR_ARCHIVE );

		s_addressbook_fields[i].generic.type = MTYPE_FIELD;
		s_addressbook_fields[i].generic.name = 0;
		s_addressbook_fields[i].generic.callback = 0;
		s_addressbook_fields[i].generic.x		= 0;
		s_addressbook_fields[i].generic.y		= i * 2.25*(MENU_FONT_SIZE+2) + 0; //18
		s_addressbook_fields[i].generic.localdata[0] = i;
		s_addressbook_fields[i].cursor			= 0;
		s_addressbook_fields[i].length			= 60;
		s_addressbook_fields[i].visible_length	= 30;

		strcpy( s_addressbook_fields[i].buffer, adr->string );

		Menu_AddItem( &s_addressbook_menu, &s_addressbook_fields[i] );
	}
}

const char *AddressBook_MenuKey( int key )
{
	if ( key == K_ESCAPE )
	{
		int index;
		char buffer[20];

		for ( index = 0; index < NUM_ADDRESSBOOK_ENTRIES; index++ )
		{
			Com_sprintf( buffer, sizeof( buffer ), "adr%d", index );
			Cvar_Set( buffer, s_addressbook_fields[index].buffer );
		}
	}
	return Default_MenuKey( &s_addressbook_menu, key );
}

void AddressBook_MenuDraw(void)
{
	M_Banner( "m_banner_addressbook" );
	Menu_Draw( &s_addressbook_menu );
}

void M_Menu_AddressBook_f(void)
{
	AddressBook_MenuInit();
	M_PushMenu( AddressBook_MenuDraw, AddressBook_MenuKey );
}

/*
=============================================================================

PLAYER CONFIG MENU

=============================================================================
*/
static menuframework_s	s_player_config_menu;
static menufield_s		s_player_name_field;
static menulist_s		s_player_model_box;
static menulist_s		s_player_skin_box;
static menulist_s		s_player_handedness_box;
static menulist_s		s_player_rate_box;
static menuseparator_s	s_player_skin_title;
static menuseparator_s	s_player_model_title;
static menuseparator_s	s_player_hand_title;
static menuseparator_s	s_player_rate_title;
static menuaction_s		s_player_download_action;

// Knightmare- save skins and models here so as to not have to re-register every frame
struct model_s *playermodel;
struct model_s *weaponmodel;
struct image_s *playerskin;
char *currentweaponmodel;
// end Knightmare

#define MAX_DISPLAYNAME 16
#define MAX_PLAYERMODELS 1024

typedef struct
{
	int		nskins;
	char	**skindisplaynames;
	char	displayname[MAX_DISPLAYNAME];
	char	directory[MAX_QPATH];
} playermodelinfo_s;

static playermodelinfo_s s_pmi[MAX_PLAYERMODELS];
static char *s_pmnames[MAX_PLAYERMODELS];
static int s_numplayermodels;

static int rate_tbl[] = { 2500, 3200, 5000, 10000, 25000, 0 };
static const char *rate_names[] = { "28.8 Modem", "33.6 Modem", "Single ISDN",
	"Dual ISDN/Cable", "T1/LAN", "User defined", 0 };

void DownloadOptionsFunc( void *self )
{
	M_Menu_DownloadOptions_f();
}

static void HandednessCallback( void *unused )
{
	Cvar_SetValue( "hand", s_player_handedness_box.curvalue );
}

static void RateCallback( void *unused )
{
	if (s_player_rate_box.curvalue != sizeof(rate_tbl) / sizeof(*rate_tbl) - 1)
		Cvar_SetValue( "rate", rate_tbl[s_player_rate_box.curvalue] );
}

static void ModelCallback( void *unused )
{
	char scratch[MAX_QPATH];

	s_player_skin_box.itemnames = s_pmi[s_player_model_box.curvalue].skindisplaynames;
	s_player_skin_box.curvalue = 0;

	// Knightmare- only register model and skin on starup or when changed
	Com_sprintf( scratch, sizeof( scratch ), "players/%s/tris.md2", s_pmi[s_player_model_box.curvalue].directory );
	playermodel = re.RegisterModel( scratch );

	Com_sprintf( scratch, sizeof( scratch ), "players/%s/%s.pcx", s_pmi[s_player_model_box.curvalue].directory, s_pmi[s_player_model_box.curvalue].skindisplaynames[s_player_skin_box.curvalue] );
	playerskin = re.RegisterSkin( scratch );

	// show current weapon model (if any)
	if (currentweaponmodel && strlen(currentweaponmodel)) {
		Com_sprintf( scratch, sizeof( scratch ), "players/%s/%s", s_pmi[s_player_model_box.curvalue].directory, currentweaponmodel );
		weaponmodel = re.RegisterModel( scratch );
		if (!weaponmodel) {
			Com_sprintf( scratch, sizeof( scratch ), "players/%s/weapon.md2", s_pmi[s_player_model_box.curvalue].directory );
			weaponmodel = re.RegisterModel( scratch );
		}
	}
	else {
		Com_sprintf( scratch, sizeof( scratch ), "players/%s/weapon.md2", s_pmi[s_player_model_box.curvalue].directory );
		weaponmodel = re.RegisterModel( scratch );
	}
	// end Knightmare
}

// Knightmare- only register skin on starup and when changed
static void SkinCallback( void *unused )
{
	char scratch[MAX_QPATH];

	Com_sprintf( scratch, sizeof( scratch ), "players/%s/%s.pcx", s_pmi[s_player_model_box.curvalue].directory, s_pmi[s_player_model_box.curvalue].skindisplaynames[s_player_skin_box.curvalue] );
	playerskin = re.RegisterSkin( scratch );
}


// Knightmare- Psychospaz's menu support for TGA and JPG skins
//static qboolean IconOfSkinExists( char *skin, char **pcxfiles, int npcxfiles )
static qboolean IconOfSkinExists( char *skin, char **files, int nfiles, char *suffix )
{
	int i;
	char scratch[1024];

	strcpy( scratch, skin );
	*strrchr( scratch, '.' ) = 0;
	//strcat( scratch, "_i.pcx" );
	strcat( scratch, suffix );

	for ( i = 0; i < nfiles; i++ )
	{
		if ( strcmp( files[i], scratch ) == 0 )
			return true;
	}

	return false;
}


// Knightmare- added Psychospaz's menus support for JPG and TGA skins
static qboolean PlayerConfig_ScanDirectories (void)
{
	char findname[1024];
	char scratch[1024];
	int ndirs = 0, npms = 0;
	char **dirnames;
	char *path = NULL;
	int i;

	extern char **FS_ListFiles( char *, int *, unsigned, unsigned );

	s_numplayermodels = 0;

	// Knightmare- loop back to here if there were no valid player models found in the selected path
	do
	{
		/*
		** get a list of directories
		*/
		do 
		{
			path = FS_NextPath( path );
			Com_sprintf( findname, sizeof(findname), "%s/players/*.*", path );

			if ( ( dirnames = FS_ListFiles( findname, &ndirs, SFF_SUBDIR, 0 ) ) != 0)
				break;
		} while ( path );

		if ( !dirnames )
			return false;

		/*
		** go through the subdirectories
		*/
		npms = ndirs;
		if ( npms > MAX_PLAYERMODELS )
			npms = MAX_PLAYERMODELS;

		for ( i = 0; i < npms; i++ )
		{
			int k, s;
			char	*a, *b, *c;
			char	**skinnames;
			char	**imagenames;
			int		nimagefiles;
			int		nskins = 0;

			if ( dirnames[i] == 0 )
				continue;

			// verify the existence of tris.md2
			strcpy( scratch, dirnames[i] );
			strcat( scratch, "/tris.md2" );
			if ( !Sys_FindFirst( scratch, 0, SFF_SUBDIR | SFF_HIDDEN | SFF_SYSTEM ) )
			{
				free( dirnames[i] );
				dirnames[i] = 0;
				Sys_FindClose();
				continue;
			}
			Sys_FindClose();

			// verify the existence of at least one skin
			//strcpy( scratch, dirnames[i] );
			//strcat( scratch, "/*.pcx" );
			strcpy( scratch, va("%s%s", dirnames[i], "/*.*" ));

			imagenames = FS_ListFiles( scratch, &nimagefiles, 0, SFF_SUBDIR | SFF_HIDDEN | SFF_SYSTEM );

			if ( !imagenames )
			{
				free( dirnames[i] );
				dirnames[i] = 0;
				continue;
			}

			// count valid skins, which consist of a skin with a matching "_i" icon
			/*for ( k = 0; k < nimagefiles-1; k++ )
			{
				if ( !strstr( imagenames[k], "_i.pcx" ) )
				{
					if ( IconOfSkinExists( imagenames[k], imagenames, nimagefiles - 1 ) )
					{
						nskins++;
					}
				}
			}*/
			for ( k = 0; k < nimagefiles-1; k++ )
				if (	!strstr( imagenames[k], "_i.pcx")
					||  !strstr( imagenames[k], "_i.jpg" )
					||  !strstr( imagenames[k], "_i.tga") )
					if (	IconOfSkinExists( imagenames[k], imagenames, nimagefiles - 1 , "_i.pcx")
						||	IconOfSkinExists( imagenames[k], imagenames, nimagefiles - 1 , "_i.jpg")
						||	IconOfSkinExists( imagenames[k], imagenames, nimagefiles - 1 , "_i.tga"))
						nskins++;

			if ( !nskins )
				continue;

			skinnames = malloc( sizeof( char * ) * ( nskins + 1 ) );
			memset( skinnames, 0, sizeof( char * ) * ( nskins + 1 ) );

			// copy the valid skins
			if (nimagefiles)
				for ( s = 0, k = 0; k < nimagefiles-1; k++ )
				{
					char *a, *b, *c;

				//	if ( !strstr( imagenames[k], "_i.pcx" ) )
					if (	!strstr( imagenames[k], "_i.pcx")
						||  !strstr( imagenames[k], "_i.jpg" )
						||  !strstr( imagenames[k], "_i.tga") )
					{
					//	if ( IconOfSkinExists( imagenames[k], imagenames, nimagefiles - 1, "_i.pcx"))
						if (	IconOfSkinExists( imagenames[k], imagenames, nimagefiles - 1 , "_i.pcx")
							||	IconOfSkinExists( imagenames[k], imagenames, nimagefiles - 1 , "_i.jpg")
							||	IconOfSkinExists( imagenames[k], imagenames, nimagefiles - 1 , "_i.tga"))
						{
							a = strrchr( imagenames[k], '/' );
							b = strrchr( imagenames[k], '\\' );

							if ( a > b )
								c = a;
							else
								c = b;

							strcpy( scratch, c + 1 );

							if ( strrchr( scratch, '.' ) )
								*strrchr( scratch, '.' ) = 0;

							skinnames[s] = strdup( scratch );
							s++;
						}
					}
				}

			// at this point we have a valid player model
			s_pmi[s_numplayermodels].nskins = nskins;
			s_pmi[s_numplayermodels].skindisplaynames = skinnames;

			// make short name for the model
			a = strrchr( dirnames[i], '/' );
			b = strrchr( dirnames[i], '\\' );

			if ( a > b )
				c = a;
			else
				c = b;

			strncpy( s_pmi[s_numplayermodels].displayname, c + 1, MAX_DISPLAYNAME-1 );
			strcpy( s_pmi[s_numplayermodels].directory, c + 1 );

			FreeFileList( imagenames, nimagefiles );

			s_numplayermodels++;
		}
		
		if ( dirnames )
			FreeFileList( dirnames, ndirs );

	// Knightmare- if no valid player models found in path,
	// try next path, if there is one
	} while (s_numplayermodels == 0 && path);

	return true;	//** DMP warning fix
}

static int pmicmpfnc( const void *_a, const void *_b )
{
	const playermodelinfo_s *a = ( const playermodelinfo_s * ) _a;
	const playermodelinfo_s *b = ( const playermodelinfo_s * ) _b;

	/*
	** sort by male, female, then alphabetical
	*/
	if ( strcmp( a->directory, "male" ) == 0 )
		return -1;
	else if ( strcmp( b->directory, "male" ) == 0 )
		return 1;

	if ( strcmp( a->directory, "female" ) == 0 )
		return -1;
	else if ( strcmp( b->directory, "female" ) == 0 )
		return 1;

	return strcmp( a->directory, b->directory );
}


qboolean PlayerConfig_MenuInit( void )
{
	extern cvar_t *name;
	extern cvar_t *team;
	extern cvar_t *skin;
	char currentdirectory[1024];
	char currentskin[1024];
	char scratch[MAX_QPATH];
	int i = 0;
	int y = 0;

	int currentdirectoryindex = 0;
	int currentskinindex = 0;

	cvar_t *hand = Cvar_Get( "hand", "0", CVAR_USERINFO | CVAR_ARCHIVE );

	static const char *handedness[] = { "right", "left", "center", 0 };

	PlayerConfig_ScanDirectories();

	if (s_numplayermodels == 0)
		return false;

	if ( hand->value < 0 || hand->value > 2 )
		Cvar_SetValue( "hand", 0 );

	strcpy( currentdirectory, skin->string );

	if ( strchr( currentdirectory, '/' ) )
	{
		strcpy( currentskin, strchr( currentdirectory, '/' ) + 1 );
		*strchr( currentdirectory, '/' ) = 0;
	}
	else if ( strchr( currentdirectory, '\\' ) )
	{
		strcpy( currentskin, strchr( currentdirectory, '\\' ) + 1 );
		*strchr( currentdirectory, '\\' ) = 0;
	}
	else
	{
		strcpy( currentdirectory, "male" );
		strcpy( currentskin, "grunt" );
	}

	qsort( s_pmi, s_numplayermodels, sizeof( s_pmi[0] ), pmicmpfnc );

	memset( s_pmnames, 0, sizeof( s_pmnames ) );
	for ( i = 0; i < s_numplayermodels; i++ )
	{
		s_pmnames[i] = s_pmi[i].displayname;
		if ( Q_stricmp( s_pmi[i].directory, currentdirectory ) == 0 )
		{
			int j;

			currentdirectoryindex = i;

			for ( j = 0; j < s_pmi[i].nskins; j++ )
			{
				if ( Q_stricmp( s_pmi[i].skindisplaynames[j], currentskin ) == 0 )
				{
					currentskinindex = j;
					break;
				}
			}
		}
	}

	if (menu_scale->value && viddef.width >= MENU_STATIC_WIDTH )
	//if (videoScale() > 1)
	{
		s_player_config_menu.x = viddef.width  / 2 - scaledVideo(210);
		//s_player_config_menu.x = scaledVideo(150);
		s_player_config_menu.y = viddef.height / 2 - scaledVideo(70.0);
		s_player_config_menu.nitems = 0;

		s_player_name_field.generic.type = MTYPE_FIELD;
		s_player_name_field.generic.flags = QMF_LEFT_JUSTIFY;
		s_player_name_field.generic.name = "name";
		s_player_name_field.generic.callback = 0;
		s_player_name_field.generic.x		= -MENU_FONT_SIZE;
		s_player_name_field.generic.y		= y; //0
		s_player_name_field.length	= 20;
		s_player_name_field.visible_length = 20;
		strcpy( s_player_name_field.buffer, name->string );
		s_player_name_field.cursor = strlen( name->string );

		s_player_model_title.generic.type = MTYPE_SEPARATOR;
		s_player_model_title.generic.flags = QMF_LEFT_JUSTIFY;
		s_player_model_title.generic.name = "model";
		s_player_model_title.generic.x    = -MENU_FONT_SIZE;
		s_player_model_title.generic.y	 = y += 3*(MENU_FONT_SIZE+2); //24

		s_player_model_box.generic.type = MTYPE_SPINCONTROL;
		s_player_model_box.generic.x	= -7*MENU_FONT_SIZE;
		s_player_model_box.generic.y	= y += (MENU_FONT_SIZE+2); //32
		s_player_model_box.generic.callback = ModelCallback;
		s_player_model_box.generic.cursor_offset = -6*MENU_FONT_SIZE;
		s_player_model_box.curvalue = currentdirectoryindex;
		s_player_model_box.itemnames = s_pmnames;

		s_player_skin_title.generic.type = MTYPE_SEPARATOR;
		s_player_skin_title.generic.flags = QMF_LEFT_JUSTIFY;
		s_player_skin_title.generic.name = "skin";
		s_player_skin_title.generic.x    = -2*MENU_FONT_SIZE;
		s_player_skin_title.generic.y	 = y += 2*(MENU_FONT_SIZE+2); //48

		s_player_skin_box.generic.type = MTYPE_SPINCONTROL;
		s_player_skin_box.generic.x	= -7*MENU_FONT_SIZE;
		s_player_skin_box.generic.y	= y += (MENU_FONT_SIZE+2); //56
		s_player_skin_box.generic.name	= 0;
		s_player_skin_box.generic.callback = SkinCallback; // Knightmare added, was 0
		s_player_skin_box.generic.cursor_offset = -6*MENU_FONT_SIZE;
		s_player_skin_box.curvalue = currentskinindex;
		s_player_skin_box.itemnames = s_pmi[currentdirectoryindex].skindisplaynames;

		s_player_hand_title.generic.type = MTYPE_SEPARATOR;
		s_player_hand_title.generic.flags = QMF_LEFT_JUSTIFY;
		s_player_hand_title.generic.name = "handedness";
		s_player_hand_title.generic.x    = 4*MENU_FONT_SIZE;
		s_player_hand_title.generic.y	 = y += 2*(MENU_FONT_SIZE+2); //72

		s_player_handedness_box.generic.type = MTYPE_SPINCONTROL;
		s_player_handedness_box.generic.x	= -7*MENU_FONT_SIZE;
		s_player_handedness_box.generic.y	= y += (MENU_FONT_SIZE+2); //80
		s_player_handedness_box.generic.name	= 0;
		s_player_handedness_box.generic.cursor_offset = -6*MENU_FONT_SIZE;
		s_player_handedness_box.generic.callback = HandednessCallback;
		s_player_handedness_box.curvalue = Cvar_VariableValue( "hand" );
		s_player_handedness_box.itemnames = handedness;

		for (i = 0; i < sizeof(rate_tbl) / sizeof(*rate_tbl) - 1; i++)
			if (Cvar_VariableValue("rate") == rate_tbl[i])
				break;

		s_player_rate_title.generic.type = MTYPE_SEPARATOR;
		s_player_rate_title.generic.flags = QMF_LEFT_JUSTIFY;
		s_player_rate_title.generic.name = "connect speed";
		s_player_rate_title.generic.x    = 7*MENU_FONT_SIZE;
		s_player_rate_title.generic.y	 = y += 2*(MENU_FONT_SIZE+2); //96

		s_player_rate_box.generic.type = MTYPE_SPINCONTROL;
		s_player_rate_box.generic.x	= -7*MENU_FONT_SIZE;
		s_player_rate_box.generic.y	= y += (MENU_FONT_SIZE+2); //104
		s_player_rate_box.generic.name	= 0;
		s_player_rate_box.generic.cursor_offset = -6*MENU_FONT_SIZE;
		s_player_rate_box.generic.callback = RateCallback;
		s_player_rate_box.curvalue = i;
		s_player_rate_box.itemnames = rate_names;

		s_player_download_action.generic.type = MTYPE_ACTION;
		s_player_download_action.generic.name	= "download options";
		s_player_download_action.generic.flags = QMF_LEFT_JUSTIFY;
		s_player_download_action.generic.x	= -5*MENU_FONT_SIZE;
		s_player_download_action.generic.y	= y += 2*(MENU_FONT_SIZE+2); //120
		s_player_download_action.generic.statusbar = NULL;
		s_player_download_action.generic.callback = DownloadOptionsFunc;
	}
	else
	{
		s_player_config_menu.x = viddef.width / 2 - 95; 
		s_player_config_menu.y = viddef.height / 2 - 97;	
		s_player_config_menu.nitems = 0;

		s_player_name_field.generic.type = MTYPE_FIELD;
		s_player_name_field.generic.name = "name";
		s_player_name_field.generic.callback = 0;
		s_player_name_field.generic.x		= 0;
		s_player_name_field.generic.y		= 0;
		s_player_name_field.length	= 20;
		s_player_name_field.visible_length = 20;
		strcpy( s_player_name_field.buffer, name->string );
		s_player_name_field.cursor = strlen( name->string );

		s_player_model_title.generic.type = MTYPE_SEPARATOR;
		s_player_model_title.generic.name = "model";
		s_player_model_title.generic.x    = -MENU_FONT_SIZE;
		s_player_model_title.generic.y	 = 60;

		s_player_model_box.generic.type = MTYPE_SPINCONTROL;
		s_player_model_box.generic.x	= -56;
		s_player_model_box.generic.y	= 70;
		s_player_model_box.generic.callback = ModelCallback;
		s_player_model_box.generic.cursor_offset = -48;
		s_player_model_box.curvalue = currentdirectoryindex;
		s_player_model_box.itemnames = s_pmnames;

		s_player_skin_title.generic.type = MTYPE_SEPARATOR;
		s_player_skin_title.generic.name = "skin";
		s_player_skin_title.generic.x    = -16;
		s_player_skin_title.generic.y	 = 84;

		s_player_skin_box.generic.type = MTYPE_SPINCONTROL;
		s_player_skin_box.generic.x	= -56;
		s_player_skin_box.generic.y	= 94;
		s_player_skin_box.generic.name	= 0;
		s_player_skin_box.generic.callback = SkinCallback; // Knightmare added, was 0
		s_player_skin_box.generic.cursor_offset = -48;
		s_player_skin_box.curvalue = currentskinindex;
		s_player_skin_box.itemnames = s_pmi[currentdirectoryindex].skindisplaynames;

		s_player_hand_title.generic.type = MTYPE_SEPARATOR;
		s_player_hand_title.generic.name = "handedness";
		s_player_hand_title.generic.x    = 32;
		s_player_hand_title.generic.y	 = 108;

		s_player_handedness_box.generic.type = MTYPE_SPINCONTROL;
		s_player_handedness_box.generic.x	= -56;
		s_player_handedness_box.generic.y	= 118;
		s_player_handedness_box.generic.name	= 0;
		s_player_handedness_box.generic.cursor_offset = -48;
		s_player_handedness_box.generic.callback = HandednessCallback;
		s_player_handedness_box.curvalue = Cvar_VariableValue( "hand" );
		s_player_handedness_box.itemnames = handedness;

		for (i = 0; i < sizeof(rate_tbl) / sizeof(*rate_tbl) - 1; i++)
			if (Cvar_VariableValue("rate") == rate_tbl[i])
				break;

		s_player_rate_title.generic.type = MTYPE_SEPARATOR;
		s_player_rate_title.generic.name = "connect speed";
		s_player_rate_title.generic.x    = 56;
		s_player_rate_title.generic.y	 = 156;

		s_player_rate_box.generic.type = MTYPE_SPINCONTROL;
		s_player_rate_box.generic.x	= -56;
		s_player_rate_box.generic.y	= 166;
		s_player_rate_box.generic.name	= 0;
		s_player_rate_box.generic.cursor_offset = -48;
		s_player_rate_box.generic.callback = RateCallback;
		s_player_rate_box.curvalue = i;
		s_player_rate_box.itemnames = rate_names;

		s_player_download_action.generic.type = MTYPE_ACTION;
		s_player_download_action.generic.name	= "download options";
		s_player_download_action.generic.flags= QMF_LEFT_JUSTIFY;
		s_player_download_action.generic.x	= -24;
		s_player_download_action.generic.y	= 186;
		s_player_download_action.generic.statusbar = NULL;
		s_player_download_action.generic.callback = DownloadOptionsFunc;
	}

	// Knightmare- only register model and skin on starup or when changed
	Com_sprintf( scratch, sizeof( scratch ), "players/%s/tris.md2", s_pmi[s_player_model_box.curvalue].directory );
	playermodel = re.RegisterModel( scratch );

	Com_sprintf( scratch, sizeof( scratch ), "players/%s/%s.pcx", s_pmi[s_player_model_box.curvalue].directory, s_pmi[s_player_model_box.curvalue].skindisplaynames[s_player_skin_box.curvalue] );
	playerskin = re.RegisterSkin( scratch );

	// show current weapon model (if any)
	if (currentweaponmodel && strlen(currentweaponmodel)) {
		Com_sprintf( scratch, sizeof( scratch ), "players/%s/%s", s_pmi[s_player_model_box.curvalue].directory, currentweaponmodel );
		weaponmodel = re.RegisterModel( scratch );
		if (!weaponmodel) {
			Com_sprintf( scratch, sizeof( scratch ), "players/%s/weapon.md2", s_pmi[s_player_model_box.curvalue].directory );
			weaponmodel = re.RegisterModel( scratch );
		}
	}
	else
	{
		Com_sprintf( scratch, sizeof( scratch ), "players/%s/weapon.md2", s_pmi[s_player_model_box.curvalue].directory );
		weaponmodel = re.RegisterModel( scratch );
	}
	// end Knightmare

	Menu_AddItem( &s_player_config_menu, &s_player_name_field );
	Menu_AddItem( &s_player_config_menu, &s_player_model_title );
	Menu_AddItem( &s_player_config_menu, &s_player_model_box );
	if ( s_player_skin_box.itemnames )
	{
		Menu_AddItem( &s_player_config_menu, &s_player_skin_title );
		Menu_AddItem( &s_player_config_menu, &s_player_skin_box );
	}
	Menu_AddItem( &s_player_config_menu, &s_player_hand_title );
	Menu_AddItem( &s_player_config_menu, &s_player_handedness_box );
	Menu_AddItem( &s_player_config_menu, &s_player_rate_title );
	Menu_AddItem( &s_player_config_menu, &s_player_rate_box );
	Menu_AddItem( &s_player_config_menu, &s_player_download_action );

	return true;
}

void PlayerConfig_MenuDraw( void )
{
	extern float CalcFov( float fov_x, float w, float h );
	refdef_t refdef;
	char scratch[MAX_QPATH];

	M_Banner( "m_banner_plauer_setup" );

	memset( &refdef, 0, sizeof( refdef ) );

	if (menu_scale->value && viddef.width >= MENU_STATIC_WIDTH )
	//if (videoScale() > 1)
	{
		refdef.width = scaledVideo(250);
		refdef.height = scaledVideo(250);
		refdef.x = viddef.width / 2 + scaledVideo(35);
		refdef.y = viddef.height / 2 - scaledVideo(30);
		//refdef.x = viddef.width - (refdef.width + scaledVideo(75));
		//refdef.y = viddef.height - (refdef.height + scaledVideo(50));
		refdef.fov_x = 50;
	}
	else
	{
		refdef.x = viddef.width / 2;
		refdef.y = viddef.height / 2 - 9*MENU_FONT_SIZE; //was 72
		refdef.width = 144;
		refdef.height = 168;
		refdef.fov_x = 40;
	}
	refdef.fov_y = CalcFov( refdef.fov_x, refdef.width, refdef.height );
	refdef.time = cls.realtime*0.001;
 
	if ( s_pmi[s_player_model_box.curvalue].skindisplaynames )
	{
		static int yaw;
		int maxframe = 29, w, h, bordercolor;
		// Knightmare- Psychopspaz's support for showing weapon model
		vec3_t lightOrg;
		entity_t entity[2], *ent;

		VectorSet(lightOrg, 80, 30, 20);

		refdef.num_entities = 2;

		//Set Up Player Model
		ent = &entity[0];
		{
			memset( &entity[0], 0, sizeof( entity[0] ) );

		// Knightmare- moved registration code to init and change only
		//	Com_sprintf( scratch, sizeof( scratch ), "players/%s/tris.md2", s_pmi[s_player_model_box.curvalue].directory );
		//	ent->model = re.RegisterModel( scratch );
			ent->model = playermodel;
		//	Com_sprintf( scratch, sizeof( scratch ), "players/%s/%s.pcx", s_pmi[s_player_model_box.curvalue].directory, s_pmi[s_player_model_box.curvalue].skindisplaynames[s_player_skin_box.curvalue] );
		//	ent->skin = re.RegisterSkin( scratch );
			ent->skin = playerskin;

			ent->flags = RF_FULLBRIGHT|RF_NOSHADOW;
			if (hand->value == 1)
				ent->flags |= RF_MIRRORMODEL;
			ent->origin[0] = 80;
			ent->origin[1] = 0;
			ent->origin[2] = 0;

			VectorCopy( ent->origin, ent->oldorigin );
			ent->frame = 0;
			ent->oldframe = 0;
			ent->backlerp = 0.0;
			ent->angles[1] = yaw++;
			if ( ++yaw > 360 )
				yaw -= 360;
			
			if (hand->value == 1)
				ent->angles[1] = 360 - ent->angles[1];
		}
		ent = &entity[1];
		{
			memset( &entity[1], 0, sizeof( entity[1] ) );

			// try to find a way to get right vwep model here (if it's set)

		// Knightmare- moved registration code to init and change only
		//	Com_sprintf( scratch, sizeof( scratch ), "players/%s/weapon.md2", s_pmi[s_player_model_box.curvalue].directory );
		//	ent->model = re.RegisterModel( scratch );
			ent->model = weaponmodel;

			if (!ent->model)
				refdef.num_entities = 1;

			ent->skinnum = 0;

			ent->flags = RF_FULLBRIGHT|RF_NOSHADOW;
			if (hand->value == 1)
				ent->flags |= RF_MIRRORMODEL;
			ent->origin[0] = 80;
			ent->origin[1] = 0;
			ent->origin[2] = 0;

			VectorCopy( ent->origin, ent->oldorigin );
			ent->frame = 0;
			ent->oldframe = 0;
			ent->backlerp = 0.0;
			ent->angles[1] = yaw;//++;
			//if ( ++yaw > 360 )
			//	yaw -= 360;
			
			if (hand->value == 1)
				ent->angles[1] = 360 - ent->angles[1];
		}

		refdef.areabits = 0;
		refdef.entities = &entity;
		refdef.lightstyles = 0;
		refdef.rdflags = RDF_NOWORLDMODEL;
		// end Knightmare
		/*
		entity_t entity;

		memset( &entity, 0, sizeof( entity ) );

		Com_sprintf( scratch, sizeof( scratch ), "players/%s/tris.md2", s_pmi[s_player_model_box.curvalue].directory );
		entity.model = re.RegisterModel( scratch );
		Com_sprintf( scratch, sizeof( scratch ), "players/%s/%s.pcx", s_pmi[s_player_model_box.curvalue].directory, s_pmi[s_player_model_box.curvalue].skindisplaynames[s_player_skin_box.curvalue] );
		entity.skin = re.RegisterSkin( scratch );
		entity.flags = RF_FULLBRIGHT;
		entity.origin[0] = 80;
		entity.origin[1] = 0;
		entity.origin[2] = 0;
		VectorCopy( entity.origin, entity.oldorigin );
		entity.frame = 0;
		entity.oldframe = 0;
		entity.backlerp = 0.0;
		entity.angles[1] = yaw++;
		if ( ++yaw > 360 )
			yaw -= 360;

		refdef.areabits = 0;
		refdef.num_entities = 1;
		refdef.entities = &entity;
		refdef.lightstyles = 0;
		refdef.rdflags = RDF_NOWORLDMODEL;
		*/

		Menu_Draw( &s_player_config_menu );

		if (menu_scale->value && viddef.width >= MENU_STATIC_WIDTH )
		//if (videoScale() > 1)
		{
			//MAKE BORDER AROUND MODEL
			bordercolor = 3;

				re.DrawFill (refdef.x-5, refdef.y-5, 
					refdef.width + 10, 5, 
					bordercolor);

				re.DrawFill (refdef.x-5, refdef.y-5, 
					5, refdef.height + 10, 
					bordercolor);

				re.DrawFill (refdef.x+refdef.width, refdef.y-5, 
					5, refdef.height+10, 
					bordercolor);

				re.DrawFill (refdef.x-5, refdef.y+refdef.height, 
					refdef.width+10, 5, 
					bordercolor);
		}
		else
		{
			M_DrawTextBox( (refdef.x) * (320.0F / viddef.width) - MENU_FONT_SIZE, (viddef.height / 2) * ( 240.0F / viddef.height) - 77, refdef.width / 8, refdef.height / 8 );
			refdef.height += 4;
		}
		re.RenderFrame( &refdef );

		Com_sprintf( scratch, sizeof( scratch ), "/players/%s/%s_i.pcx", 
			s_pmi[s_player_model_box.curvalue].directory,
			s_pmi[s_player_model_box.curvalue].skindisplaynames[s_player_skin_box.curvalue] );

		if (menu_scale->value && viddef.width >= MENU_STATIC_WIDTH )
		//if (videoScale() > 1)
		{
			re.DrawGetPicSize ( &w, &h, scratch);
			re.DrawFill (
					refdef.x,
					refdef.y,
					scaledVideo(w)+5,
					scaledVideo(h)+5, 
					bordercolor);
			re.DrawStretchPic (
				refdef.x,
				refdef.y,
				scaledVideo(w), scaledVideo(h), scratch, 1.0);
		}
		else
			re.DrawPic( s_player_config_menu.x - 40, refdef.y, scratch );
	}
}

void PConfigAccept (void)
{
	int i;
	char scratch[1024];

	Cvar_Set( "name", s_player_name_field.buffer );

	Com_sprintf( scratch, sizeof( scratch ), "%s/%s", 
		s_pmi[s_player_model_box.curvalue].directory, 
		s_pmi[s_player_model_box.curvalue].skindisplaynames[s_player_skin_box.curvalue] );

	Cvar_Set( "skin", scratch );

	for ( i = 0; i < s_numplayermodels; i++ )
	{
		int j;

		for ( j = 0; j < s_pmi[i].nskins; j++ )
		{
			if ( s_pmi[i].skindisplaynames[j] )
				free( s_pmi[i].skindisplaynames[j] );
			s_pmi[i].skindisplaynames[j] = 0;
		}
		free( s_pmi[i].skindisplaynames );
		s_pmi[i].skindisplaynames = 0;
		s_pmi[i].nskins = 0;
	}
}

const char *PlayerConfig_MenuKey (int key)
{
	if ( key == K_ESCAPE )
		PConfigAccept();

	return Default_MenuKey( &s_player_config_menu, key );
}


void M_Menu_PlayerConfig_f (void)
{
	if (!PlayerConfig_MenuInit())
	{
		Menu_SetStatusBar( &s_multiplayer_menu, "No valid player models found" );
		return;
	}
	Menu_SetStatusBar( &s_multiplayer_menu, NULL );
	M_PushMenu( PlayerConfig_MenuDraw, PlayerConfig_MenuKey );
}


/*
=======================================================================

GALLERY MENU

=======================================================================
*/
#if 0
void M_Menu_Gallery_f( void )
{
	extern void Gallery_MenuDraw( void );
	extern const char *Gallery_MenuKey( int key );

	M_PushMenu( Gallery_MenuDraw, Gallery_MenuKey );
}
#endif

/*
=======================================================================

QUIT MENU

=======================================================================
*/

const char *M_Quit_Key (int key)
{
	switch (key)
	{
	case K_ESCAPE:
	case 'n':
	case 'N':
		M_PopMenu ();
		break;

	case 'Y':
	case 'y':
		cls.key_dest = key_console;
		CL_Quit_f ();
		break;

	default:
		break;
	}
	return NULL;
}


void M_Quit_Draw (void)
{
	int		w, h;

	re.DrawGetPicSize (&w, &h, "quit");
	re.DrawStretchPic (
		(viddef.width-scaledVideo(w))/2, 
		(viddef.height-scaledVideo(h))/2,
		scaledVideo(w), scaledVideo(h), "quit", 1.0);
	//re.DrawPic ( (viddef.width-w)/2, (viddef.height-h)/2, "quit");
}


void M_Menu_Quit_f (void)
{
	M_PushMenu (M_Quit_Draw, M_Quit_Key);
}



//=============================================================================
/* Menu Subsystem */


/*
=================
M_Init
=================
*/
void M_Init (void)
{
	// Knightmare- init this cvar here so M_Print can use it
	if (!alt_text_color)
		alt_text_color = Cvar_Get ("alt_text_color", "2", CVAR_ARCHIVE);
	
	Cmd_AddCommand ("menu_main", M_Menu_Main_f);
	Cmd_AddCommand ("menu_game", M_Menu_Game_f);
		Cmd_AddCommand ("menu_loadgame", M_Menu_LoadGame_f);
		Cmd_AddCommand ("menu_savegame", M_Menu_SaveGame_f);
		Cmd_AddCommand ("menu_joinserver", M_Menu_JoinServer_f);
			Cmd_AddCommand ("menu_addressbook", M_Menu_AddressBook_f);
		Cmd_AddCommand ("menu_startserver", M_Menu_StartServer_f);
			Cmd_AddCommand ("menu_dmoptions", M_Menu_DMOptions_f);
		Cmd_AddCommand ("menu_playerconfig", M_Menu_PlayerConfig_f);
			Cmd_AddCommand ("menu_downloadoptions", M_Menu_DownloadOptions_f);
		Cmd_AddCommand ("menu_credits", M_Menu_Credits_f );
	Cmd_AddCommand ("menu_multiplayer", M_Menu_Multiplayer_f );
	Cmd_AddCommand ("menu_video", M_Menu_Video_f);
	Cmd_AddCommand ("menu_options", M_Menu_Options_f);
		Cmd_AddCommand ("menu_keys", M_Menu_Keys_f);
	Cmd_AddCommand ("menu_quit", M_Menu_Quit_f);
}


/*
=================================
Menu Mouse Cursor - psychospaz
=================================
*/

void refreshCursorMenu (void)
{
	cursor.menu = NULL;
}
void refreshCursorLink (void)
{
	cursor.menuitem = NULL;
}

int Slider_CursorPositionX ( menuslider_s *s )
{
	float range;

	range = ( s->curvalue - s->minvalue ) / ( float ) ( s->maxvalue - s->minvalue );

	if ( range < 0)
		range = 0;
	if ( range > 1)
		range = 1;

	return ( int )( scaledVideo(MENU_FONT_SIZE) + RCOLUMN_OFFSET + (SLIDER_RANGE)*scaledVideo(MENU_FONT_SIZE) * range );
}

int newSliderValueForX (int x, menuslider_s *s)
{
	float newValue;
	int newValueInt;
	int pos = x - scaledVideo(MENU_FONT_SIZE + RCOLUMN_OFFSET + s->generic.x) - s->generic.parent->x;

	newValue = ((float)pos)/((SLIDER_RANGE-1)*scaledVideo(MENU_FONT_SIZE));
	newValueInt = s->minvalue + newValue * (float)( s->maxvalue - s->minvalue );

	return newValueInt;
}

void Slider_CheckSlide( menuslider_s *s )
{
	if ( s->curvalue > s->maxvalue )
		s->curvalue = s->maxvalue;
	else if ( s->curvalue < s->minvalue )
		s->curvalue = s->minvalue;

	if ( s->generic.callback )
		s->generic.callback( s );
}

void Menu_DragSlideItem (menuframework_s *menu, void *menuitem)
{
	menucommon_s *item = ( menucommon_s * ) menuitem;
	menuslider_s *slider = ( menuslider_s * ) menuitem;

	slider->curvalue = newSliderValueForX(cursor.x, slider);
	Slider_CheckSlide ( slider );
}

void Menu_ClickSlideItem (menuframework_s *menu, void *menuitem)
{
	int min, max;
	menucommon_s *item = ( menucommon_s * ) menuitem;
	menuslider_s *slider = ( menuslider_s * ) menuitem;

	min = menu->x + scaledVideo(item->x + Slider_CursorPositionX(slider) - 4);
	max = menu->x + scaledVideo(item->x + Slider_CursorPositionX(slider) + 4);

	if (cursor.x < min)
		Menu_SlideItem( menu, -1 );
	if (cursor.x > max)
		Menu_SlideItem( menu, 1 );
}

void M_Think_MouseCursor (void)
{
	char * sound = NULL;
	menuframework_s *m = (menuframework_s *)cursor.menu;

	if (m_drawfunc == M_Main_Draw) //have to hack for main menu :p
	{
		CheckMainMenuMouse();
		return;
	}
	if (m_drawfunc == M_Credits_MenuDraw) //have to hack for credits :p
	{
		if (cursor.buttonclicks[MOUSEBUTTON2])
		{
			cursor.buttonused[MOUSEBUTTON2] = true;
			cursor.buttonclicks[MOUSEBUTTON2] = 0;
			cursor.buttonused[MOUSEBUTTON1] = true;
			cursor.buttonclicks[MOUSEBUTTON1] = 0;
			S_StartLocalSound( menu_out_sound );
			if (creditsBuffer)
				FS_FreeFile (creditsBuffer);
			M_PopMenu();
			return;
		}
	}

	if (!m)
		return;

	//Exit with double click 2nd mouse button

	if (cursor.menuitem)
	{
		//MOUSE1
		if (cursor.buttondown[MOUSEBUTTON1])
		{
			if (cursor.menuitemtype == MENUITEM_SLIDER)
			{
				Menu_DragSlideItem(m, cursor.menuitem);
			}
			else if (!cursor.buttonused[MOUSEBUTTON1] && cursor.buttonclicks[MOUSEBUTTON1])
			{
				if (cursor.menuitemtype == MENUITEM_ROTATE)
				{
					if (menu_rotate->value)					
						Menu_SlideItem( m, -1 );
					else			
						Menu_SlideItem( m, 1 );

					sound = menu_move_sound;
					cursor.buttonused[MOUSEBUTTON1] = true;
				}
				else
				{
					cursor.buttonused[MOUSEBUTTON1] = true;
					Menu_MouseSelectItem( cursor.menuitem );
					sound = menu_move_sound;
				}
			}
		}
		//MOUSE2
		if (cursor.buttondown[MOUSEBUTTON2] && cursor.buttonclicks[MOUSEBUTTON2])
		{
			if (cursor.menuitemtype == MENUITEM_SLIDER && !cursor.buttonused[MOUSEBUTTON2])
			{
				Menu_ClickSlideItem(m, cursor.menuitem);
				sound = menu_move_sound;
				cursor.buttonused[MOUSEBUTTON2] = true;
			}
			else if (!cursor.buttonused[MOUSEBUTTON2])
			{
				if (cursor.menuitemtype == MENUITEM_ROTATE)
				{
					if (menu_rotate->value)					
						Menu_SlideItem( m, 1 );
					else			
						Menu_SlideItem( m, -1 );

					sound = menu_move_sound;
					cursor.buttonused[MOUSEBUTTON2] = true;
				}
			}
		}
	}
	else if (!cursor.buttonused[MOUSEBUTTON2] && cursor.buttonclicks[MOUSEBUTTON2]==2 && cursor.buttondown[MOUSEBUTTON2])
	{
		if (m_drawfunc==PlayerConfig_MenuDraw)
			PConfigAccept();

		if (m_drawfunc==Options_MenuDraw && options_menu->value>0)
		{
			Cvar_SetValue( "options_menu", 0 );
			refreshCursorLink();
		}
		else
			M_PopMenu();

		sound = menu_out_sound;
		cursor.buttonused[MOUSEBUTTON2] = true;
		cursor.buttonclicks[MOUSEBUTTON2] = 0;
		cursor.buttonused[MOUSEBUTTON1] = true;
		cursor.buttonclicks[MOUSEBUTTON1] = 0;
	}

	if ( sound )
		S_StartLocalSound( sound );
}

void M_Draw_Cursor (void)
{
	float alpha = 1, scale = 1;
	int w,h;
	char *overlay = NULL;
	char *cur_img = NULL;

	if (m_drawfunc == M_Main_Draw)
	{
		if (MainMenuMouseHover)
		{
			if ((cursor.buttonused[0] && cursor.buttonclicks[0])
				|| (cursor.buttonused[1] && cursor.buttonclicks[1]))
			{
				cur_img = "m_cur_click";
				alpha = 0.85 + 0.15*sin(anglemod(cl.time*0.005));
			}
			else
			{
				cur_img = "m_cur_hover";
				alpha = 0.85 + 0.15*sin(anglemod(cl.time*0.005));
			}
		}
		else
			cur_img = "m_cur_main";
		overlay = "m_cur_over";
	}
	else
	{
		if (cursor.menuitem)
		{
			if (cursor.menuitemtype == MENUITEM_TEXT)
			{
				cur_img = "m_cur_text";
			}
			else
			{
				if ((cursor.buttonused[0] && cursor.buttonclicks[0])
					|| (cursor.buttonused[1] && cursor.buttonclicks[1]))
				{
					cur_img = "m_cur_click";
					alpha = 0.85 + 0.15*sin(anglemod(cl.time*0.005));
				}
				else
				{
					cur_img = "m_cur_hover";
					alpha = 0.85 + 0.15*sin(anglemod(cl.time*0.005));
				}
				overlay = "m_cur_over";
			}
		}
		else
		{
			cur_img = "m_cur_main";
			overlay = "m_cur_over";
		}
	}
	
	if (cur_img)
	{
		re.DrawGetPicSize( &w, &h, cur_img );
		re.DrawScaledPic( cursor.x - w/2, cursor.y - h/2,
			scaledVideo(scale), alpha, cur_img, true, false);

		if (overlay)
		{
			re.DrawGetPicSize( &w, &h, "m_cur_over" );
			re.DrawScaledPic( cursor.x - w/2, cursor.y - h/2,
				scaledVideo(scale), 1, overlay, true, false);
		}
	}
}

/*void M_Draw_Cursor (void)
{
	int w,h;

	//get sizing vars
	re.DrawGetPicSize( &w, &h, "m_mouse_cursor" );
	w = scaledVideo(w)*0.5;
	h = scaledVideo(h)*0.5;
	re.DrawStretchPic (cursor.x-w/2, cursor.y-h/2, w, h, "m_mouse_cursor", 1.0);
}*/

void InitMenuScale (void)
{
	//this is for scaling the menu...
	//this is kinda overkill, but whatever...
	// Knightmare- allow disabling of menu scaling
	// also, don't scale if < 800x600, then it would be smaller
	if (menu_scale->value && viddef.width > MENU_STATIC_WIDTH )
	{
		menuScale.x = viddef.width/MENU_STATIC_WIDTH;
		menuScale.y = viddef.height/MENU_STATIC_HEIGHT;
		menuScale.avg = (menuScale.x + menuScale.y)*0.5f;
	}
	else
	{
		menuScale.x = 1;
		menuScale.y = 1;
		menuScale.avg = 1;
	}
}

/*
=================
M_Draw
=================
*/
void M_Draw (void)
{
	if (cls.key_dest != key_menu)
		return;

	// Knigthmare- added Psychospaz's scaled menu stuff
	InitMenuScale();

	// repaint everything next frame
	SCR_DirtyScreen ();

	// dim everything behind it down
	if (cl.cinematictime > 0)
	{
		if (re.RegisterPic("/gfx/background.pcx")) {
			re.DrawStretchPic (0, 0, viddef.width, viddef.height, "/gfx/background.pcx", 1.0);
			re.DrawFadeScreen ();
		}
		else
			re.DrawFill (0,0,viddef.width, viddef.height, 0);
	}
	else
		re.DrawFadeScreen ();

	// Knigthmare- added Psychospaz's mouse support
	refreshCursorMenu();

	m_drawfunc ();

	// delay playing the enter sound until after the
	// menu has been drawn, to avoid delay while
	// caching images
	if (m_entersound)
	{
		S_StartLocalSound( menu_in_sound );
		m_entersound = false;
	}

	// Knigthmare- added Psychospaz's mouse support
	//menu cursor for mouse usage :)
	M_Draw_Cursor();
}


/*
=================
M_Keydown
=================
*/
void M_Keydown (int key)
{
	const char *s;

	if (m_keyfunc)
		if ( ( s = m_keyfunc( key ) ) != 0 )
			S_StartLocalSound( ( char * ) s );
}


